/*
  Copyright 1991 University of Rhode Island
 
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose is hereby granted without fee, provided that
  the above copyright notice appear in all copies and that both that
  copyright notice and this permission notice appear in supporting
  documentation, and that the name of U.R.I. not be used in advertising or
  publicity pertaining to distribution of the software without specific,
  written prior permission.  U.R.I. makes no representations about the
  suitability of this software for any purpose.  It is provided "as is"
  without express or implied warranty.
 
  U.R.I. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.R.I.
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
  BROWSE handles browsing at a high level of abstraction. An
  instance of BROWSE reads a file name from the object pointed to
  by list, displays the information it recieves on the `device' pointed
  to by display, and writes log information to the object pointed to by
  log.
  
  When init_browse is called, it first the argument `file'. If it
  is non-NULL, then that value is used a the name of a file that
  contains a list of files (on the server) to browse. If `file' is NULL,
  init_browse examines regexp.  if regexp is non-NULL, then
  BROWSE assumes that there is a valid regexp in that parameter
  and will transmit that regexp to the server. Subsequently, the server
  will transmit all images that match that regexp. If regexp is NULL,
  then an error is signaled.
  
  Modifications
  
  7/19/90
  Removed the references to the field navdata. They may be added by defining
  the symbol BROWSE_NAV.
  
  7/25/90
  Modified stats display so it returned bytes and time (for trans)
  
  9/10/90
  Removed 'BROWSE_NAV' lines - navigation is handled differently.
  
  12/10/90
  Ported to unix.

  2/12/91
  Added code to set/reset region_defd_p field of BROWSE. This flag is TRUE
  when the user has defined a region, FALSE otherwise.

  2/13/91
  Added images_txd and cost. They record the total number of images
  transmitted and the total cost. Added levelByteCount which records the
  number fo bytes transmitted for the current level.

  4/3/91
  Modified cost function: new value is $0.04/KByte.

  5/2/91
  Cost is now updated with every packet.

  8/30/91
  fixed up error handling to use dialog boxes, ...
*/

/*
 * $Log: browse.c,v $
 * Revision 3.8  1993/02/02  16:12:32  jimg
 * Added conditional include for debugging malloc header.
 * Added initialization of BROWSE fields `filename' and `path'.
 *
 * Revision 3.7  1992/11/19  21:14:30  jimg
 * ?
 *
 * Revision 3.6  1992/11/05  17:51:52  jimg
 * Removed old cost computation - server now computes cost.
 * Added new function to start timing transmit time.
 *
 * Revision 3.5  1992/11/04  18:21:58  jimg
 * Fixed path and filename bug when using new `query' mode of the image server.
 *
 * Revision 3.5  1992/11/04  18:21:58  jimg
 * Fixed path and filename bug when using new `query' mode of the image server.
 *
 * Revision 3.4  1992/10/22  16:01:02  jimg
 * Added fields for image server info string and logins required flag.
 *
 * Revision 3.3  1992/10/21  17:17:03  jimg
 * Changed image name from char * to char [MAXPATHLEN]
 *
 * Revision 3.2  1992/10/19  14:57:32  jimg
 * Added code to handle `query-based' access to the server.
 *
 * Revision 3.1  1992/10/14  13:19:00  jimg
 * Many minor bug fixes.
 *
 * Revision 3.0  1992/07/28  15:24:44  jimg
 * All sources for xbrowse modified to use the libcomm.a, liberrmsg.a and new
 * dialog boxes.
 *
 * Revision 1.1  1992/03/23  20:39:01  jimg
 * Initial revision
 */

static char rcsid[]="$Id: browse.c,v 3.8 1993/02/02 16:12:32 jimg Exp jimg $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <netdb.h>		/* for struct hostent - to get hostname */
#include <time.h>

#include <sys/types.h>
#include <sys/time.h>
#include <rpc/rpc.h>

#include <netinet/in.h>

#include <X11/X.h>
#include <X11/Intrinsic.h>	/* toolkit */
#include <X11/StringDefs.h>

#ifdef MALLOC_DEBUG
#include <malloc.h>
#endif

#ifndef MAXPATHLEN
#define MAXPATHLEN 256
#endif
#define COST_PER_KBYTE (0.04)	/* in dollars */
#define fix_path(p) if (p[strlen(p) - 1] != '/') strcat(p, "/")

#include "errmsg.h"
#include "comm.h"
#include "browse-list.h"
#include "browse-log.h"
#include "vertex.h"
#include "region.h"
#include "ximage.h"
#include "pyramid.h"
#include "pixmap-mark.h"
#include "xbt-const.h"
#include "xbt-types.h"
#include "xbt.h"
#include "dialog.h"
#include "browse.h"

/*
   This breaks SGIs
   int gettimeofday();
*/

/*
  Create and initialize an instance of BROWSE. Either list xor
  path must be specified, the other will be null. In the former case,
  'list' is used to create an instance of BrowseList, in the later case,
  the path is assigned to the field path.

  Changed IMGFILE (Dsp specific type) to X's Pixmap as part of the
  port to X. jhrg 12/11/90
*/

BROWSE *
init_browse(level, stats, no_question_exit, auto_zoom)
int level;
int stats;
int no_question_exit;
int auto_zoom;
{
    BROWSE *browse = XtNew(BROWSE);

    if (!browse) 
	sys_quit_dialog("Could not allocate struct BROWSE");

    browse->pyramid = init_pyramid(level);
    
    /* get the current user's login name */
    browse->name = XtNewString(cuserid(NULL));

    browse->use_local_list_p = FALSE;
    browse->use_remote_list_p = FALSE;
    browse->use_path_p = FALSE;

    browse->num_xbt_drops = FALSE;
    browse->image_server_info = NULL;
    browse->img_srvr_logins_p = FALSE;
    browse->img_srvr_logged_in = FALSE;
    browse->log_file = NULL;
    browse->current_image = -1;
    browse->image_list = NULL;
    browse->pretty_list = NULL;
    browse->filename = NULL;
    browse->path = NULL;
    browse->image_list_len = 0;
    browse->auth_user_p = FALSE;
    browse->cost_thresh = 0;	/* get real value from server */
    browse->statistics_p = stats;
    browse->no_question_exit_p = no_question_exit;
    browse->auto_zoom_p = auto_zoom;
    browse->auto_region_p = FALSE;
    browse->region_sent_p = FALSE;
    browse->region_defd_p = FALSE;	/* TRUE when user defines a region */
    browse->region_hm_disp_p = FALSE;
    browse->accumulate_cost_p = FALSE;
    browse->bytes_txd = 0;
    browse->images_txd = 0;
    browse->total_cost = 0.0;
    browse->image_cost = 0.0;
    
    return browse;
}

void 
free_browse(browse)
BROWSE *browse;
{
    if (browse) {
	free_pyramid(browse->pyramid);
	
	if (browse->list)
	    close_list(browse->list);
	
	if (browse->log_file) 
	    close_log(browse->log_file);

	if (browse->path)
	    XtFree(browse->path);
	
	XtFree((char *)browse);
    }
}

/*
  Given a pointer to an instance of BROWSE, a buffer and the size of that
  buffer, take information from `data' and add it to the pyramid. Add
  information in units of whole scan lines in the current level of the
  pyramid.  If the last byte of the current level is reached, return (i.e.,
  don't get info from more than one level from the buffer). Return TRUE if
  more rows remain in the current level, FALSE if the current level has been
  completely displayed.

  Note: improve this by replacing browse->pyramid->resolution with a function
  that returns the correct number of bytes given the current region.
*/

void
data_to_pyramid(browse, buf, buf_size)
BROWSE *browse;
char *buf;
long buf_size;
{
    char *orig_buf = buf;
    long num_rows = 0;
    struct {
	short row_num;
	short row_len;
	short start;
	char last_row;
	char optimized;
#ifdef NEVER
	/* rll code - under development */
	char rll;
	char rll_byte;
#endif
    } row_pack;

    do {
	assert((buf - orig_buf) < buf_size);

	memcpy(&row_pack, buf, sizeof(row_pack));

	buf += sizeof(row_pack);
	num_rows++;
	
	if (ntohs(row_pack.row_len)) /* trap null rows */
	    buf += add_sampled_row(browse->pyramid, buf,
				   ntohs(row_pack.row_len),
				   ntohs(row_pack.start),
				   ntohs(row_pack.row_num),
				   row_pack.optimized
#ifdef NEVER
				   /* rll - under development */
				   row_pack.rll,
				   row_pack.rll_byte
#endif
				   );
	
	assert((buf - orig_buf) <= buf_size);
    } while (!row_pack.last_row);

    browse->bytes_txd += (buf - orig_buf) - num_rows * sizeof(row_pack);
    
#ifdef NEVER
    if (browse->accumulate_cost_p)
	browse->image_cost = calculate_cost(browse->bytes_txd);
#endif
}

/*
  Set up browse state variables for the next image.
*/

void
browse_next_image(browse)
BROWSE *browse;
{
    /* erase the hash marks for the current region. No effect first time */
    erase_region_hash_marks(browse);
    browse->region_hm_disp_p = FALSE;

    if (browse->num_xbt_drops) {
	erase_xbt_drops(browse->num_xbt_drops, browse->xbt_drops,
			browse->pyramid, browse->xbt_mark);
	browse->num_xbt_drops = 0;
    }
    
    browse->bytes_txd = 0;
    browse->images_txd++;
    browse->total_cost += browse->image_cost;
    browse->image_cost = 0.0;
    browse->accumulate_cost_p = TRUE;

    if (browse->total_cost > 0.01)
	print_cost(browse);

    print_image_name(browse->pyramid);
}

/*
  Zero the transmission time for an image. This function is called right
  before the message is sent to the server requesting the next image. Thus,
  the time displayed includes time to get the image out of the archive as well
  as transmission times.
*/

void
browse_zero_time(browse)
BROWSE *browse;
{
    if (browse->statistics_p) {
	if (gettimeofday(&browse->start_time, NULL) == -1) {
	    sys_err_dialog("Could not set start time for statistics");
	    browse->statistics_p = FALSE;
	    return;
	}
    }
}

/*
  When the user is done with the current image, record the time. Not called
  if the statistics_p field is FALSE.
*/

void 
stop_browse_image(browse)
BROWSE *browse;
{
    if (browse->statistics_p) 
	if (gettimeofday(&browse->end_time, NULL) == -1)
	    sys_err_dialog("Could not set end time for statistics");
}

/*
  Print statistics about the transmission - time, bytes sent, ... Not called
  if the statistics_p field is FALSE.
*/

void 
print_browse_stats(browse)
BROWSE *browse;
{
    long sec, usec;
    double time;
    double trans_rate;		/* K bytes/sec */
    
    if (browse->statistics_p) {
	sec = browse->end_time.tv_sec - browse->start_time.tv_sec;
	usec = browse->end_time.tv_usec - browse->start_time.tv_usec;
	
	if (usec < 0) {
	    sec--;
	    usec += 1000000;	/* microsec in one sec */
	}
	
	time = (double)sec + (double)usec / 1000000;
	
	trans_rate = (double)browse->bytes_txd / time / 1000;
	
	/* err_print dumps this to stderr - nominally the CLI */
	err_print("Bytes: %d, Time: %5.2lf sec, (%3.1lf Kbytes/sec)", 
		  browse->bytes_txd, time, trans_rate);
    }
}

void 
mark_old_region_hash_marks(browse)
BROWSE *browse;
{
    PYRAMID *pyramid = browse->pyramid;
    VERTICES *vertices = browse->pyramid->region->vertices;
    PIXMAP_MARK *mark = browse->hash_mark;

    if (vertices_p(vertices)) {		/* if not NULL */
	reset_vertices(vertices);
	
	do {
	    long x = image_to_screen_x(pyramid, get_vertex_x(vertices));
	    long y = image_to_screen_y(pyramid, get_vertex_y(vertices));
	    
	    draw_pixmap_mark2(mark, x, y);
	} while(next_vertex(vertices));
    }
}

/*
  Draw region vertices.
*/

void 
draw_region_hash_marks(browse)
BROWSE *browse;
{
    PYRAMID *pyramid = browse->pyramid;
    VERTICES *vertices = browse->pyramid->region->vertices;
    PIXMAP_MARK *mark = browse->hash_mark;
    
    if (vertices_p(vertices)) {	/* if not NULL */
	reset_vertices(vertices);
	
	do {
	    long x = image_to_screen_x(pyramid, get_vertex_x(vertices));
	    long y = image_to_screen_y(pyramid, get_vertex_y(vertices));

	    draw_pixmap_mark(mark, x, y);
	} while(next_vertex(vertices));
    }
}

/*
  Erase region vertices.
*/

void 
erase_region_hash_marks(browse)
BROWSE *browse;
{
    PYRAMID *pyramid = browse->pyramid;
    VERTICES *vertices = pyramid->region->vertices;
    PIXMAP_MARK *mark = browse->hash_mark;
    
    if (vertices_p(vertices)) {
	reset_vertices(vertices);

	do {
	    long x = image_to_screen_x(pyramid, get_vertex_x(vertices));
	    long y = image_to_screen_y(pyramid, get_vertex_y(vertices));

	    erase_pixmap_mark(mark, x, y);
	} while(next_vertex(vertices));
    }
}
