/*
  Copyright 1991, 1992 University of Rhode Island
 
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose is hereby granted without fee, provided that
  the above copyright notice appear in all copies and that both that
  copyright notice and this permission notice appear in supporting
  documentation, and that the name of U.R.I. not be used in advertising or
  publicity pertaining to distribution of the software without specific,
  written prior permission.  U.R.I. makes no representations about the
  suitability of this software for any purpose.  It is provided "as is"
  without express or implied warranty.
 
  U.R.I. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.R.I.
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
  This file conatains a state machine (implemented using callbacks and an
  XtInput or XtWork Proc) that sends commands to the browse daemon based on
  the user's selection of button items and reads image data from the daemon.
  It also calls the neccessary functions to expand the data so that X can
  display it the user's window.  Note that when the state machine is put into
  the wait state, the WorkProc is removed. It is reinstalled when the users
  leaves the wait state.

  Modifications:
  
  1/31/91
  Switched to 4K packets (packets can be smaller if one level will fit in <
  4K of space).

  Modified updates of the X window -- now it is updated every time a packet
  is received.

  4/17/91
  Finished the new zooming code. Fixed bug where sections of the image that
  were visible were not updated.

  5/2/91
  Fixed zooming divide by zero error & added error messages.  Modified states
  so that falling through causes all the buttons except quit to be dimmed and
  the image name to become 'Done'. Thus, the window no longer just
  disappears. This confused people.

  8/7/91
  Fixed multiple zoom bug + saving windows of zoomed images.

  8/26/91
  Added `Use Region' button and user region functions.

  8/30/91
  Reduced to number of states.

  9/3/91
  reduced overhead due to over zealous error checking (and rechecking, and ...)

  9/5/91
  changed from switch to callbacks.

  9/12/91
  Enabled InputProc mode (uses XtInputCallback to read the socket). See
  #define WORKPROC

  10/2/91
  Added code to beep when user tries to define a second region w/vertices
  outside the current region.

  10/21/91
  Added preprocessor symbol to control whether `cost' is printed in the
  xbrowse window.

  1/6/92
  Removed last vestiges of old (VMS version) state machine code.
*/

/*
 * $Log: callbacks.c,v $
 * Revision 3.9  1993/02/02  16:12:32  jimg
 * Added conditional include for debugging malloc header.
 * Fixed malloc error in beautify_name().
 *
 * Revision 3.8  1992/11/19  21:14:30  jimg
 * Modified for grid.
 * Organized functions a bit.
 *
 * Revision 3.7  1992/11/06  18:49:02  jimg
 * Fixed valid_date() so that it no longer required single digit months and
 * days to be prefixed by `0''s.
 *
 * Revision 3.6  1992/11/05  17:51:52  jimg
 * Fixed computation of transmission time so that it includes timerequired
 * to get the image out of the archive.
 *
 * Revision 3.5  1992/11/04  18:34:03  jimg
 * Fixed bug where login cancel was ignored - this caused a crash.
 *
 * Revision 3.4  1992/11/03  19:48:06  jimg
 * Fixed bug in plot_xbt() where a cancel in xbt_connect() was ignored.
 *
 * Revision 3.3  1992/10/21  17:17:03  jimg
 * Fixed nagging communications bug: program would hang when user stopped
 * transmission in the instant between the end of one level and begining
 * of a the next level (of the image being displayed).
 *
 * Acompanying changes were made to the image server (browsed).
 *
 * Revision 3.2  1992/10/19  14:57:32  jimg
 * Added code to handle `query-based' access to the server.
 *
 * Revision 3.1  1992/10/14  13:19:00  jimg
 * Many minor bug fixes.
 *
 * Revision 3.0  1992/07/28  15:24:44  jimg
 * All sources for xbrowse modified to use the libcomm.a, liberrmsg.a and new
 * dialog boxes.
 * A synchronization point was added in use_region() to fix a bug where the
 * client and server would deadlock - each waiting for the other to send
 * information.
 *
 * Revision 1.1  1992/03/23  20:39:01  jimg
 * Initial revision
 */

static char rcsid[]="$Id: callbacks.c,v 3.9 1993/02/02 16:12:32 jimg Exp jimg $";

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/file.h>
#include <rpc/rpc.h>

#include <X11/Intrinsic.h>	/* toolkit */
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>

#ifdef MALLOC_DEBUG
#include <malloc.h>
#endif 

#define MAX_STR_LEN 256
#define SOCKET_SIZE 4096
#define PORT_NO 7227		/* connect to server running on this port */
#define DATA_PORT_NO 7233	/* image server data only */
#define SERVICE_NAME NULL	/* connect to service using this name */
#define MAX_PIXEL_VAL 255	/* ... in an image on the screen */
#undef PRINT_COST_INFO		/* if defined, cost will be displayed -
				   displaying this while access is/was free
				   caused confusion, heartache, ... Make sure
				   to change the label in XBrowse.ad too. */
#define CLIENT_VERSION_MAJOR 3	/* version of this program */
#define CLIENT_VERSION_MINOR 3
#define LOWEST_MAJOR_VERSION 2	/* lowest version server to connect with */
#define LOWEST_MINOR_VERSION 0
#define DATE_LEN 10
#define BASE_YEAR 1900
#define LEAP_YEAR(year) ((int) ((year % 4 == 0 && year % 100 != 0) \
				|| year % 400 == 0) ? 1 : 0) /* K&R p.37 */
#define fix_path(p) if (p[strlen(p) - 1] != '/') strcat(p, "/")

#if XtSpecificationRelease < 5
#if NeedFunctionPrototypes
typedef void* XtPointer;
#else
#define XtPointer char *
#endif /* Need function prototypes */
#endif /* Not R5  */

#include "errmsg.h"
#include "comm.h"
#include "browsed-codes.h"
#include "xbrowse-strings.h"
#include "dialog.h"
#include "xbrowse.h"
#include "browse-list.h"
#include "browse-log.h"
#include "vertex.h"
#include "region.h"
#include "ximage.h"
#include "pyramid.h"
#include "browsed-funcs.h"	/* for browsed <-> xbrowse struct trans */
#include "pixmap-mark.h"
#include "xbt-const.h"
#include "xbt-types.h"
#include "xbt.h"
#include "browse.h"
#include "grid.h"
#include "callbacks.h"
#include "make-pathname.h"

#include "vertex.bitmap"
#include "xbt.bitmap"

static void browse_remote_list();
static void browse_remote_path();
static void browse_local_list();
static void do_zoom();
static long do_log_entry();
static void do_next_level();
static void start_row_tx();
static void do_exit();
static long define_region();
static void use_region();
static void write_region_file();
static long read_region_file();
static int do_save();
static int do_info();
static int do_init();
static int do_version();
static int do_login();

static long read_vertices();
static int read_next_packet();
static void send_region();
static void do_file_path();
static int read_image_params();
static void pt_outside_region_message();
static Region vertices_to_x_region();
static void interrupt_handler();
static void set_buttons_done();
static int write_pgm();
static int valid_date();
static int browse_next_level();
static char *beautify_name();
static char *get_dsp_pretty_name();


/* globals */
BROWSE *browse;			/* information used by the state machine */
XDR_TCP *server;		/* image server: action/resps only ( */
XDR_TCP *xbt_server;		/* xbt server socket */
int rec_data;
XtWorkProcId read_proc_id;

/* from xbrowse.c */
extern Widget toplevel;		/* ancestor of all widgets */
extern Widget the_image;	/* holdws image data on screen */
extern Widget image_name;	/* displays image name */
extern Widget projection, satellite_type;
extern Widget max_res, current_res;
extern Widget cost, cursor;
extern Widget quit, pause_tx, save, next; /* button widgets */
extern Widget palette, log_entry, zoom;
extern Widget region_use, region_define;
#ifdef NEVER
extern Widget region_write, region_read;
#endif
extern Widget auto_zoom;
extern Widget xbt_display;
extern Widget auto_region;

extern long window_pixmap_x;	/* size of the image pixmap */
extern long window_pixmap_y;
extern static_visual;
extern unsigned char image_pal[];
extern APPDATA a_data;		/* cmd line switches/resources */
extern XtAppContext a_context;	/* for portability */
extern Display *display;
extern Colormap colormap;


/* WorkProcs */

/*
  Connect with the image server and initialze the browse structure (along
  with its substructures). This is run only once.
*/

Boolean
init_state(toplevel)
XtPointer toplevel;		/* unused */
{
    int sd /*, sd_data*/;
    Cursor watch;
    Cursor arrow;
    Cursor cross_hair;

    xdr_tcp_set_struct(IMAGE_HEADER_CODE, sizeof(struct image_header),
		       xdr_image_header);

    /* show watch while initializing */
    watch = XCreateFontCursor(display, XC_watch);
    XDefineCursor(display, XtWindow(toplevel), watch);

    if (!a_data.server[0]) { /* not on cmd line or in resource db */
	char *temp_server = get_x11_string("Server machine name:",
					   "zeno.gso.uri.edu");
	if (!temp_server)
	    exit(1);		/* users hits cancel */
	a_data.server = XtNewString(temp_server);
    }

    sd = client_connect(a_data.server, SERVICE_NAME, PORT_NO, sys_err_dialog);
    
    /* if named sever cannot be accessed, then prompt user for differ name */
    
    while (sd == -1) {
	char *temp_server = get_x11_string("Server machine name:",
					   "zeno.gso.uri.edu");
	if (!temp_server)
	    exit(1);		/* users hits cancel */
	a_data.server = XtNewString(temp_server);
	sd = client_connect(a_data.server, SERVICE_NAME, PORT_NO,
			    sys_err_dialog); 
    }

    server = xdr_tcp_bind(sd, SOCKET_SIZE, sys_quit_dialog);

    /* `xbt_server' is bound with the first request. see get_xbt...() */
    xbt_server = NULL;

    /* initialize BROWSE structure once connected to an image server */
    browse = init_browse(a_data.level, a_data.statistics,
			 a_data.no_question_exit, a_data.auto_zoom);

    /* allocate & initialize the region and xbt marks */
    browse->hash_mark = create_pixmap_mark(vertex_bits, vertex_width,
					   vertex_height, vertex_x_hot,
					   vertex_y_hot, a_data.hashmark,
					   a_data.old_hashmark);

    browse->xbt_mark = create_pixmap_mark(xbt_bits, xbt_width,
					  xbt_height, xbt_x_hot,
					  xbt_y_hot, "black", "red");
    
    (void) signal(SIGINT, interrupt_handler); /* trap ctrl-c */
    
    if (!do_init(server, browse)) {
	err_dialog(LEVEL_INIT_ERR_MSG, browse->pyramid->max_pyramid_level,
		   browse->pyramid->init_level);
	exit(1);
    }
    
    if (!do_version(server, browse)) {
	err_dialog(VERSION_INIT_ERR_MSG, CLIENT_VERSION_MAJOR,
		   CLIENT_VERSION_MINOR);
	exit(1);
    }

    /* get the server's info string - these are used by browse_remote_list()
       and do_save() */

    browse->img_srvr_logins_p = do_info(server, &browse->image_server_info);
    
    /* if the user specifies a list or a path, set `browse' to use those
       modes, otherwise, prompt for a range of dates and use the remote list
       mode. */
    if (a_data.list[0])
	browse_local_list(browse, a_data.list);
    else if (a_data.path[0])
	browse_remote_path(server, browse, a_data.path);
    else {
	browse_remote_list(server, browse);
	XFlush(display);	/* pop down the dialog right away */
    }
	
    /* if the using a list (local or remote) show scollable list */
    if (browse->use_remote_list_p || browse->use_local_list_p)
	create_image_list_window(server, browse);
    
    /* set toplevel's cursor to arrow and the_image to crosshair */

    arrow = XCreateFontCursor(display, XC_top_left_arrow);
    XDefineCursor(display, XtWindow(toplevel), arrow);

    cross_hair = XCreateFontCursor(display, XC_crosshair);
    XDefineCursor(display, XtWindow(the_image), cross_hair);

    if (browse->auto_zoom_p)
	XtVaSetValues(auto_zoom, XtNlabel, "No Auto Zoom", NULL);
    else
	XtVaSetValues(auto_zoom, XtNlabel, "Auto Zoom", NULL);

    if (a_data.grid)
	init_image_grid();
    
    do_next_image(server, browse); /* get the ball rolling */
    
    return TRUE;		/* don't run any more */
}

/*
  Action procedure that reads the next packet from the socket and adds the
  data to the image pyramid. If the EOD marker is read, the rec_data and
  was_rec_data globals are set to false and read_image_data removes itself
  from the action proc queue. Otherwise rec_data is TRUE, was_rec_data is
  false and read_image_data remains in the queue.
*/

Boolean
read_image_data(toplevel)
XtPointer toplevel;
{
    if (!rec_data)
	return TRUE;
    if (read_next_packet(server, browse)) /* more data ? */
	return FALSE;		/* stay in queue */
    else if (browse_next_level(server, browse)) {
	do_next_level(server, browse);
	return FALSE;		/* stay in queue */
    }
    else {			/* must be done w/this image */
	stop_browse_image(browse);
	XtSetSensitive(pause_tx, False);
	rec_data = FALSE;
	return TRUE;		/* return TRUE removes the workproc */
    }
}

/* Button Callbacks */

/*
  Functions run as callbacks from the set of buttons.
*/  

void
quit_button(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    int was_rec_data;
    
    if (!XtIsSensitive(quit))
	return;
    
    if (rec_data)
	was_rec_data = TRUE;
    else
	was_rec_data = FALSE;

    browse->accumulate_cost_p = FALSE;
    do_stop_image_tx(server, browse);

    if (browse->no_question_exit_p || question_dialog("Really Quit ?")) {
	stop_browse_image(browse);		
	print_browse_stats(browse);
	browse->images_txd++;
	browse->total_cost += browse->image_cost;
	if (xbt_server)
	    xbt_close(xbt_server);
	do_exit(server);
	free_browse(browse);
	exit(0);
    }
    else if (was_rec_data) {	/* user decides not to quit - should we
				   restart?  */
	browse->accumulate_cost_p = TRUE;
	start_row_tx(server);
    }
}

void
pause_button(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    if (!XtIsSensitive(pause_tx))
	return;

    if (rec_data) {
	do_stop_image_tx(server, browse);
	XtVaSetValues(pause_tx, XtNlabel, "Resume", NULL);
    }
    else {
	start_row_tx(server);
	XtVaSetValues(pause_tx, XtNlabel, "Pause", NULL);
    }	
}

void
next_button(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{ 
    if (!XtIsSensitive(next))
	return;

    browse->accumulate_cost_p = FALSE;
    do_stop_image_tx(server, browse);

    stop_browse_image(browse);		
    print_browse_stats(browse);
    XtSetSensitive(pause_tx, True);
    XtVaSetValues(pause_tx, XtNlabel, "Pause", NULL);
    if (browse->region_defd_p)
	XtSetSensitive(region_use, TRUE);
    do_next_image(server, browse);
}

void
log_button(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{ 
    int was_rec_data = FALSE;
    
    if (!XtIsSensitive(log_entry))
	return;
    
    if (rec_data)
	was_rec_data = TRUE;
    
    browse->accumulate_cost_p = FALSE;
    do_stop_image_tx(server, browse);

    stop_browse_image(browse);		
    print_browse_stats(browse);
    
    if (do_log_entry(browse)) {
	if (was_rec_data)
	    XtVaSetValues(pause_tx, XtNlabel, "Resume", NULL);
	else
	    XtVaSetValues(pause_tx, XtNlabel, "Pause", NULL);	    
	XtSetSensitive(pause_tx, True);
    }
    else if (was_rec_data) {	/* user decides not to log - should we
				   restart?  */
	browse->accumulate_cost_p = TRUE;
	start_row_tx(server);
    }
}

void
save_button(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    int was_rec_data = FALSE;
    
    if (!XtIsSensitive(save))
	return;
    
    if (rec_data)
	was_rec_data = TRUE;

    browse->accumulate_cost_p = FALSE;
    do_stop_image_tx(server, browse);

    stop_browse_image(browse);		
    print_browse_stats(browse);

    if (do_save(browse)) {
	if (was_rec_data)
	    XtVaSetValues(pause_tx, XtNlabel, "Resume", NULL);
	else
	    XtVaSetValues(pause_tx, XtNlabel, "Pause", NULL);	    
	XtSetSensitive(pause_tx, True);
    }
    else if (was_rec_data) {	/* user decides not to save - should we
				   restart?  */
	browse->accumulate_cost_p = TRUE;
	start_row_tx(server);
    }
}

void
palette_button(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{ 
    if (!XtIsSensitive(palette))
	return;
    
    do_palette(browse, "", FALSE);	/* "" causes do_palette to prompt */
    update_display(browse->pyramid);
    redraw_picture(the_image, NULL);
}

void
use_region_button(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{ 
    int was_rec_data = FALSE;
    
    if (!XtIsSensitive(region_use))
	return;
    
    if (rec_data)
	was_rec_data = TRUE;

    do_stop_image_tx(server, browse);
    
    use_region(server, browse);

    if (browse->auto_zoom_p) { /* auto zoom */
	do_zoom(server, browse);
	XtSetSensitive(zoom, False);
	XtSetSensitive(pause_tx, True);
	/* if the user lets the display fill up and then zooms, transmission
	   will have stopped. To restart the pause button must be hit. Thus,
	   the pause button label must be `resume' (the display is currently
	   paused). */
	if (!was_rec_data)
	    XtVaSetValues(pause_tx, XtNlabel, "Resume", NULL);
    }
    
    if (was_rec_data)
	start_row_tx(server);
}

void
def_region_button(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{ 
    int was_rec_data = FALSE;
    
    if (!XtIsSensitive(region_define))
	return;
    
    if (rec_data)
	was_rec_data = TRUE;

    do_stop_image_tx(server, browse);
    
    XtVaSetValues(region_define, XtNlabel, "Exit: Button 2", NULL);

    if (define_region(browse)) {
	use_region(server, browse);

	if (browse->auto_zoom_p) { /* auto zoom */
	    do_zoom(server, browse);
	    XtSetSensitive(zoom, False);
	    XtSetSensitive(pause_tx, True);

	    if (!was_rec_data)	/* hack - due to automatic pause `feature' */
		start_row_tx(server);
	}
    }

    XtVaSetValues(region_define, XtNlabel, "Define Region", NULL);

    if (was_rec_data)
	start_row_tx(server);
}

void
write_region_button(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{ 
    int was_rec_data = FALSE;
    
#ifdef NEVER
    if (!XtIsSensitive(region_write))
	return;
#endif
    
    if (rec_data) {
	was_rec_data = TRUE;
	do_stop_image_tx(server, browse);
    }
    
    write_region_file(browse);
    
    if (was_rec_data)
	start_row_tx(server);
}

void
read_region_button(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{ 
    int was_rec_data = FALSE;
    
#ifdef NEVER
    if (!XtIsSensitive(region_read))
	return;
#endif
    
    if (rec_data) {
	was_rec_data = TRUE;
	do_stop_image_tx(server, browse);
    }
    
    if (read_region_file(browse)) {
#ifdef NEVER
	XtSetSensitive(region_write, True);
#endif
	use_region(server, browse);
    }

    if (was_rec_data)
	start_row_tx(server);
}

void
zoom_button(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    if (!XtIsSensitive(zoom))
	return;
    
    do_stop_image_tx(server, browse);
    
    do_zoom(server, browse);
    XtSetSensitive(pause_tx, True);

    if (browse->pyramid->max_image_level == browse->pyramid->max_pyramid_level)
	XtSetSensitive(zoom, False);

    start_row_tx(server);
}

void
auto_zoom_button(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    if (!XtIsSensitive(auto_zoom))
	return;

    browse->auto_zoom_p = !browse->auto_zoom_p;

    if (browse->auto_zoom_p)
	XtVaSetValues(auto_zoom, XtNlabel, "No Auto Zoom", NULL);
    else
	XtVaSetValues(auto_zoom, XtNlabel, "Auto Zoom", NULL);
}

void
get_xbt_button(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    float box[4];
    char image_name[10];	/* temp hack */
    long image_time;
    
    if (!xbt_server) {	/* static global; executed first time only */
	xbt_server = xbt_connect(a_data.xbt_server);
	if (!xbt_server)	/* user hits cancel in xbt_connect() */
	    return;
    }

    box[0] = browse->pyramid->header->top_left_lon;
    box[1] = browse->pyramid->header->top_left_lat;
    box[2] = browse->pyramid->header->bot_right_lon;
    box[3] = browse->pyramid->header->bot_right_lat;

    /* assume we can get the date from the image file name - HACK */
    /* filename format <s><yy><ddd><hh><mm><ss> *or* <s><yy><ddd><hh><?> */
    
    if (browse->path)
	strncpy(image_name, &browse->filename[1], 9);
    else if (strrchr(browse->filename, '/'))
	strncpy(image_name, strrchr(browse->filename, '/') + 2, 7);
    else if (strrchr(browse->filename, ']'))
	strncpy(image_name, strrchr(browse->filename, ']') + 2, 7);

    image_time = atoi(image_name);
    image_time *= 100;		/* effectively zeros minutes (see `7' above) */

    browse->num_xbt_drops = get_xbt_drops(xbt_server, box, image_time, 12,
					  browse->xbt_drops); 

    if (!browse->num_xbt_drops)
	err_dialog("No xbt coverage corresponding to this image");
    else
	draw_xbt_drops(browse->num_xbt_drops, browse->xbt_drops,
		       browse->pyramid, browse->xbt_mark);
}

void
auto_region_button(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    if (!XtIsSensitive(auto_region))
	return;

    browse->auto_region_p = !browse->auto_region_p;

    if (browse->auto_region_p)
	XtVaSetValues(auto_region, XtNlabel, "No Auto Region", NULL);
    else
	XtVaSetValues(auto_region, XtNlabel, "Auto Region", NULL);
}

void
get_server_cwd(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    int was_rec_data = FALSE;
    
#ifdef NEVER
    if (!XtIsSensitive(region_define))
	return;
#endif
    
    if (rec_data) {
	was_rec_data = TRUE;
	do_stop_image_tx(server, browse);
    }
    
    send_action(server, A_GET_CWD, TRUE);

    switch (read_type_code(server, TRUE)) {
    case action_tc: {
	u_int action = read_action(server);
	if (action == A_ERR)
	    err_dialog(CANT_GET_CWD_MSG);
	else
	    sys_err_dialog(UNKNOWN_ACTION_MSG);
	break;
    }

    case string_tc: {
	char path[MAXPATHLEN];
	read_string(server, path, MAXPATHLEN);
	err_print("server cwd: `%s'", path);
	break;
    }

    default:
	sys_err_dialog(UNKNOWN_TYPE_CODE_MSG);
    }

    if (was_rec_data) {
	start_row_tx(server);
    }
}

void
set_server_cwd(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    int was_rec_data = FALSE;
    char *path;
    
#ifdef NEVER
    if (!XtIsSensitive(region_define))
	return;
#endif
    
    if (rec_data) {
	was_rec_data = TRUE;
	do_stop_image_tx(server, browse);
    }

    send_action(server, A_CHANGE_CWD, TRUE);

    path = get_x11_string("New directory spec:", "");

    if (path)
	send_string(server, path, strlen(path) , TRUE);
    else
	send_string(server, "", strlen("") , TRUE);
    
    switch(get_action(server, TRUE)) {
    case A_ACK:
	break;
	
    case A_ERR:
	err_dialog("Could not change to directory `%s'", path);
	break;
    }

    if (path)
	free(path);
    
    if (was_rec_data) {
	start_row_tx(server);
    }
}

void
get_server_dir_listing(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    int was_rec_data = FALSE;
#ifdef NEVER
    u_int ack;
#endif
    
    if (rec_data) {
	was_rec_data = TRUE;
	do_stop_image_tx(server, browse);
    }

    (void) send_action(server, A_GET_DIR_LISTING, TRUE);

    switch (read_type_code(server, TRUE)) {
    case action_tc: {
	u_int action = read_action(server);
	if (action == A_ERR)
	    err_dialog(CANT_GET_DIR_MSG);
	else
	    sys_err_dialog(UNKNOWN_ACTION_MSG);
	break;
    }

    case long_tc: {
	long i, n;
	char files[100][MAXPATHLEN];

	(void) read_long(server, &n);
	if (n <= 0)
	    break;		/* -1 incidactes err */
	
	for (i = 0; i < n; i++) {
	    (void) get_string(server, files[i], MAXPATHLEN, TRUE);
	    err_print("recieved `%s'", files[i]);
	}

	break;
    }

    default:
	sys_err_dialog(UNKNOWN_TYPE_CODE_MSG);
    }
    
    if (was_rec_data) {
	start_row_tx(server);
    }
}

/*
  Functions that communicate with the image server.
*/

/*
  Stop transmission of the image. First send the stop transmission message,
  then read the image information cought in the lag time between the time the
  stop message was sent and when the server received it.
*/

void
do_stop_image_tx(server, browse)
XDR_TCP *server;
BROWSE *browse;
{
    send_action(server, A_STOP_IMAGE_TX, TRUE); /* stop trans */

    /* must read data caught in time lag */
    while (rec_data && read_next_packet(server, browse))
	;
}

/*
  do_zoom - just what is says. Let the user zoom in on a region of the image.
  Currently, the region must be defined before entering this function. If the
  image and the display are the same size, then it is not possible to zoom.
  
  6/20/91
  Changed so that this function no longer checks to see if the image
  can or cannot be zoomed. Rather, the state machine sets as sensitive or
  insensitive the zoom button depending on whether the current image can be
  zoomed.

  8/31/91
  Removed `error' codes. Changed function type to void.
*/

static void
do_zoom(server, browse)
XDR_TCP *server;
BROWSE *browse;
{
    long zoom_factor;
    long x, y;
    XImage *new_image;
    
    if (!browse->region_defd_p) {
	err_dialog("You must define a region before zooming");
	return;
    }

    /* We can increase the size of the current region within the current
       display window by zoom_factor times. This will typically be 2 or 4. */
    
    zoom_factor = pyramid_zoom_factor(browse->pyramid);

    if (!zoom_factor) {
	err_dialog(REGION_TOO_LARGE_MSG);
	return;
    }

    /* In order to zoom the current region zoom_factor times, magnify the
       contents of the rectangle enclosed by the points x,y and x + window
       size x, y + window size y by zoom_factor. */
    
    calc_new_origin(browse->pyramid->region, zoom_factor,
		    browse->pyramid->scale_factor, &x, &y);

    new_image = zoom_ximage(browse->pyramid->raw_image, zoom_factor, x, y);
    if (!new_image)
	return;

    XDestroyImage(browse->pyramid->raw_image);
    browse->pyramid->raw_image = new_image;

    /* image_origin_{x,y} is the point in the image file that will be
       displayed at the origin of the current window after the current region
       is zoomed by zoom_factor. */
    /* when auto zoom is on, marking the `old' hash marks (which have just
       been set by the user) looks dumb */
    
    if (!browse->auto_zoom_p || browse->region_hm_disp_p)
	mark_old_region_hash_marks(browse);

    if (browse->num_xbt_drops) {
	erase_xbt_drops(browse->num_xbt_drops, browse->xbt_drops,
			browse->pyramid, browse->xbt_mark);
    }

    browse->pyramid->image_origin_x += x * browse->pyramid->level_ratio;
    browse->pyramid->image_origin_y += y * browse->pyramid->level_ratio;

    zoom_pyramid(browse->pyramid, zoom_factor, y);

    /* show grid must go before update_display() since update_display()
       copies from `the_image' to `window_pixmap' using `image_gc'. Hence
       `image_gc' must have its clipmask reset. The pixmap marks are erased
       before this segment of code. */
       
    if (a_data.grid)
	show_image_grid(browse, TRUE);
    
    update_display(browse->pyramid);

    redraw_picture(the_image, NULL);

    draw_region_hash_marks(browse);
    browse->region_hm_disp_p = TRUE;

    if (browse->num_xbt_drops) {
	draw_xbt_drops(browse->num_xbt_drops, browse->xbt_drops,
		       browse->pyramid, browse->xbt_mark);
    }
}

/*
  If the log file is already open, append the current image file pathname to
  it. If the file is not open then check the resource entries for the log
  file name and path and use those to open the file if possible. If those
  resources have not been defined, prompt the user for the name of a log file.

  9/23/91
  The log file should not be opened until the user decides to put a filename
  in it (when a file is opened, its contents are lost; see open and TRUNC).

  9/30/91
  return TRUE if the image name was logged, FALSE otherwise.

  10/6/92
  If non NULL, use a_data.log as the initial filename in a new_filename()
  dialog box.
*/

static long
do_log_entry(browse)
BROWSE *browse;
{
    /* if the log file is not open, prompt for a filename */
    
    if (!browse->log_file) {	/* else, keep trying until success or cancel */
	char *loc_filename;

	if (a_data.log_path) {
	    char *loc_pathname = make_pathname(a_data.log_path, a_data.log);

	    loc_filename = new_filename("Log Filename:", loc_pathname);
	    XtFree(loc_pathname);
	}
	else
	    loc_filename = new_filename("Log Filename:", a_data.log);

	if (!loc_filename)	/* user cancels */
	    return FALSE;

	browse->log_file = open_log(loc_filename);

	XtFree(loc_filename);	/* free mem alloc'd by new_filename()*/
    }

    /* if the log file is open, save data */
    
    if (browse->log_file)
	add_log_entry(browse->log_file,
		      browse->image_list[browse->current_image]); 

    return TRUE;
}

static void
do_next_level(server, browse)
XDR_TCP *server;
BROWSE *browse;
{
    send_action(server, A_NEXT_LEVEL, TRUE);

    start_row_tx(server);
    
    print_level(browse->pyramid, TRUE);
}

static void
start_row_tx(server)
XDR_TCP *server;
{
    send_action(server, A_TX_ROWS, TRUE);	/* start transmission */

    read_proc_id = XtAppAddWorkProc(a_context, read_image_data, toplevel);

    rec_data = TRUE;
}

/*
  If the images are being specified by a regexp, then the first time
  this function is called it must tell the server what that regexp is.
  Subsequently, it must ask for the next image and test if the server
  says there are no more. Otherwise, it tells the server to open the
  image named by browse->list->name. This type of logic should be in the
  state machine proper, but in this instance that approach seems to entail a
  lot of work.

  Dropped regex code - path option looks at every image in the given
  directory. 
*/

long 
do_next_image(server, browse)
XDR_TCP *server;
BROWSE *browse;
{
    if (browse->use_path_p) {	/* using path previously sent to server */
	browse_zero_time(browse);
	send_action(server, A_NEXT_IMAGE, TRUE);
	get_string(server, browse->filename, 256, TRUE);
    }
    else if (browse->use_remote_list_p
	     || browse->use_local_list_p) { /* get filename from remote list */
	char *full_path, *filename;
	
	/* go to the next image - init'd to -1 in init_browse() */
	browse->current_image++;

	/* if done ... */
	if (browse->current_image >= browse->image_list_len) {
	    browse->statistics_p = FALSE;
	    browse->no_question_exit_p = TRUE;
	    set_buttons_done();
	    XBell(display, 50);
	    return FALSE;
	}

	/* otherwise, get name of next image to browse */
	full_path = browse->image_list[browse->current_image];

	highlight_current_image(browse);
	
	browse_zero_time(browse);
	send_action(server, A_NEW_IMAGE, TRUE);
	send_string(server, full_path, MAX_STR_LEN, TRUE);

	/* this must extract the file name from the full pathname for both
	   unix and vms style pathnames. */

	/* first look for vms file spec, else assume unix */
	if (strchr(full_path, ':') && strchr(full_path, '['))
	    filename = strrchr(full_path, ']');
	else
	    filename = strrchr(full_path, '/');

	filename++;		/* move past ']' or '/' */

	if (browse->filename)
	    XtFree(browse->filename);
	browse->filename = XtNewString(filename);

	if (browse->path)
	    XtFree(browse->path);
	browse->path = XtNewString(full_path);
    }
    else {
	err_dialog("You must choose which images to browse");
	browse_remote_list(server, browse);
	return TRUE;
    }
	     
    /* determine if image was opened by the server  - an action is sent only
       when the request to open an image fails, else a sequence of ints are
       sent. */

    switch(read_type_code(server, TRUE)) {
    case action_tc:		/* server couldn't open image */
	switch(read_action(server)) {
	case A_IMAGE_NOT_FOUND:
	    return do_next_image(server, browse);
	    
	case A_NO_MORE_IMAGES:
	    browse->statistics_p = FALSE;
	    browse->no_question_exit_p = TRUE;
	    set_buttons_done();
	    return FALSE;
	    
	default:
	    err_dialog(COMM_1_ERR_MSG);
	    exit(1);		/* call `do_quit'? */
	}
	break;
	
    case long_tc: {		    /* server opened the image */
	long image_size_x, image_size_y;

	read_long(server, &image_size_x);
	get_long(server, &image_size_y, TRUE);
	
	if (!read_image_params(server, browse))
	    return FALSE;
	
	browse_next_image(browse); /* set internal fields of BROWSE */

	if (!reset_pyramid(browse->pyramid, image_size_x, image_size_y))
	    return FALSE;

	print_level(browse->pyramid, TRUE);

	/* no image can be zoomed until a region is defined/used */
	XtSetSensitive(zoom, False); 
					
	/* allow auto zoom only for images that can actually be zoomed ! */
	if (can_zoom(browse->pyramid))
	    XtSetSensitive(auto_zoom, True);
	else
	    XtSetSensitive(auto_zoom, False);
	
	start_row_tx(server);

	if (a_data.grid)
	    show_image_grid(browse, FALSE);
	
	return TRUE;
    }
	
    default:
	err_dialog(COMM_2_ERR_MSG);
	exit(1);
    }

    return FALSE;		/* should not get here */
}

static void
do_exit(server)
XDR_TCP *server;
{
    send_action(server, A_EXIT, TRUE);
}

/*
  read_next_packet serves a dual function; it is used to read successive rows
  of image data AND sense the end of data marker that the server transmits
  when a level has been completely transmitted. Note that image information
  has a type code of T_IMAGE and the A_EOD marker has a type code T_ACTION.
  The function returns TRUE if image information was read and A_EOD if the
  A_EOD marker was read. If an unexpected EOD or an unknown type code is
  read, read_next_packet returns FALSE.
*/

static int
read_next_packet(server, browse)
XDR_TCP *server;
BROWSE *browse;
{
    switch (read_type_code(server, TRUE)) {
	case opaque_tc: {
	    static char packet[SOCKET_SIZE];
	    int f_row, l_row; /* update only new rows */
	    long len;
	    XExposeEvent e;

	    (void) read_opaque(server, packet, &len);
	    f_row = image_to_screen_y(browse->pyramid,
				      browse->pyramid->current_row);
	    data_to_pyramid(browse, packet, SOCKET_SIZE);
	    l_row = image_to_screen_y(browse->pyramid,
				      browse->pyramid->current_row);

	    e.x = 0; e.width = window_pixmap_x;
	    e.y = f_row; e.height = l_row - f_row;

	    redraw_picture(the_image, &e); /* screen update */
	    return TRUE;
	}
	    
	case action_tc: {
	    long action;

	    action = read_action(server);
	    XtRemoveWorkProc(read_proc_id);
	    rec_data = FALSE;

	    if (action == A_EOD)
		return FALSE;
	    else {
		err_dialog(COMM_2_ERR_MSG);
		return FALSE;	/* try to recover from unknown type */
	    }
	}

	default:
	    err_dialog(COMM_2_ERR_MSG);
	    return FALSE;	/* try to recover from unknown type */
    }
}

static void
send_region(server, region)
XDR_TCP *server;
REGION *region;
{
    long length = 0;
    VERTICES *vertices = region->vertices;
    
    send_action(server, A_SET_REGION, TRUE);

    reset_vertices(vertices);

    do {
	length++;
    } while(next_vertex(vertices));
    
    send_long(server, length, TRUE);

    reset_vertices(vertices);

    do {
	send_long(server, get_vertex_x(vertices), TRUE);
	send_long(server, get_vertex_y(vertices), TRUE);
    } while(next_vertex(vertices));
}

/*
  Save the curent window to a ppm file. Put the header
  information in as quotes.

  When the server recieves the A_SAVE_IMAGE action code, it responds by
  sending an authorization flag that represents the maximum resolution at
  which the image can be saved (level number) and the associated cost. Once
  the image is saved, xbrowse sends a flag (TRUE) indication that the save
  was successful. If the save was not successful, xbrowse sends FALSE.

  NB: a max save level of 0 indicates that the user is not authorized to save
  anything - cost in this case is meaningless.
*/
    
static int
do_save(browse)
BROWSE *browse;
{
    int image_file;
    char *filename, temp_name[MAXPATHLEN];
    char save_label[MAX_STR_LEN];
    long save_level, dollars, cents;
    PYRAMID *pyramid = browse->pyramid;
    
    if (browse->img_srvr_logins_p && !browse->img_srvr_logged_in) {
	if (do_login(server, ANON_LOGINS, browse))
	    browse->img_srvr_logged_in = TRUE;
	else
	    return FALSE;	/* logins req'd but login unsuccessful */
    }
	
    (void) send_action(server, A_SAVE_IMAGE, TRUE);
    (void) get_long(server, &save_level, TRUE);
    (void) get_long(server, &dollars, TRUE);
    (void) get_long(server, &cents, TRUE);

    /* if user is not allowed to save imagery, leave */
    if (!save_level) {
	err_dialog(NOT_ALLOWED_TO_SAVE_MSG);
	(void) send_long(server, FALSE, TRUE);
	return FALSE;
    }

    /* strip off directory components */

    if (strrchr(browse->filename, ']'))
	strncpy(temp_name, strrchr(browse->filename, ']') + 1L, MAXPATHLEN);
    else if (strrchr(browse->filename, '/'))
	strncpy(temp_name, strrchr(browse->filename, '/') + 1L, MAXPATHLEN);
    else
	strncpy(temp_name, browse->filename, MAXPATHLEN);

    /* remove _Z or .Z */
    
    if (strstr(temp_name, "_Z"))
	strcpy(strstr(temp_name, "_Z"), "");
    else if (strstr(temp_name, ".Z"))
	strcpy(strstr(temp_name, ".Z"), "");
    else if (strstr(temp_name, "_z"))
	strcpy(strstr(temp_name, "_z"), "");

    /* build `This image will cost $??.?? to save.\nSave image to:' */
    sprintf(save_label,
	    "This image will cost $%d.%d to save.\nSave image to:", dollars,
	    cents);
    
    /* get a filename from the user */
    if (strcmp(a_data.save_path, "") != 0) {
	char *loc_pathname = make_pathname(a_data.save_path, temp_name);

	filename = new_filename(save_label, loc_pathname);
	XtFree(loc_pathname);
    }
    else
	filename = new_filename(save_label, temp_name);

    if (!filename) {		/* Cancel button returns NULL */
	(void) send_long(server, FALSE, TRUE);
	return FALSE;
    }

    /* get a file descriptor for the file */
    image_file = open(filename, O_WRONLY | O_CREAT, 0664);
    if (image_file == -1) {
	sys_err_dialog(CANT_CREATE_FILE_MSG);
	XtFree(filename);
	(void) send_long(server, FALSE, TRUE);
	return FALSE;
    }

    XtFree(filename);

    /* just dump what's in the window - not image_size_y by image_size_x */
    if (!write_pgm(image_file, pyramid, save_level)) {
	close(image_file);
	err_dialog(CANT_WRITE_FILE_MSG);
	(void) send_long(server, FALSE, TRUE);
	return FALSE;
    }
	
    close(image_file);
    (void) send_long(server, TRUE, TRUE);

    /* once suscessful, update image cost */
    browse->image_cost = (double)dollars + (double)cents/100.0;
    print_cost(browse);
    
    return TRUE;
}

static void
do_file_path(server, browse)
XDR_TCP *server;
BROWSE *browse;
{
    if (browse->path) {
	send_action(server, A_FILE_PATH, TRUE);
	send_string(server, browse->path, strlen(browse->path) + 1, TRUE);
    }

    switch(get_action(server, TRUE)) {
    case A_ERR: {
	char *tmp_pathname;
	
	err_dialog(CANT_OPEN_PATH_MSG, browse->path);
	tmp_pathname = get_x11_string("Image path (on server)", browse->path);
	if (!tmp_pathname) {
	    do_exit(server);	/* shutdown server cleanly */
	    exit(1);
	}

	fix_path(tmp_pathname);

	/* storage for tmp_pathanme allocated by get_x11_string() using
	   XtNewString() */
	
	browse->path = tmp_pathname; 
	
	do_file_path(server, browse); /* try again ... */
    }

    case A_ACK:
	return;
    }
}

/*
  Given that a new image has been opend by the server, and that the server
  has sent the image parameters over the net, read those params.
*/

static int
read_image_params(server, browse)
XDR_TCP *server;
BROWSE *browse;
{
    unsigned int type;
    caddr_t status;

    /* status == NULL indicates error */
    status = get_struct(server, (caddr_t)browse->pyramid->header, &type, TRUE);

    return status && (type == IMAGE_HEADER_CODE);    
}

/*
  Exchange version numbers with the image server. Set any flags in the BROWSE
  structure that depend on specific server versions. Return TRUE if the
  server version numbers are non-zero, FALSE otherwise.
*/

static int
do_version(server, browse)
XDR_TCP *server;
BROWSE *browse;
{
    long server_version_major, server_version_minor;

    (void) send_action(server, A_VERSION, TRUE);
    
    (void) send_long(server, CLIENT_VERSION_MAJOR, TRUE);
    (void) send_long(server, CLIENT_VERSION_MINOR, TRUE);

    (void) get_long(server, &server_version_major, TRUE);
    (void) get_long(server, &server_version_minor, TRUE);

    /* Set any flags in BROWSE that depend on the server version number */

    if ((server_version_major >= LOWEST_MAJOR_VERSION)
	&& (server_version_minor >= LOWEST_MINOR_VERSION))
	return TRUE;
    else
	return FALSE;
}

/*
  Return TRUE if the server allows anonymous logins (although some
  restrictions may apply), FALSE if not. In addition, assing the info pointer
  to reference a string (statically allocated) that describes the server.
*/

static int
do_info(server, info)
XDR_TCP *server;
char **info;
{
    long logins_required_p;
    static char info_string[4096];

    (void) send_action(server, A_INFO, TRUE);
    
    (void) get_long(server, &logins_required_p, TRUE);
    (void) get_string(server, info_string, 4096, TRUE);

    *info = &info_string[0];
    
    return logins_required_p;
}

static int
do_login(server, info, browse)
XDR_TCP *server;
char *info;
BROWSE *browse;
{
    char *username, *password = NULL;
    char banner[256];
    long accepted_p;
    
    sprintf(banner, "Access billing account on image server at %s",
	    a_data.server);
    username = XtNewString(browse->name);

    if (!login(banner, info, &username, &password))
	return FALSE;
    
    (void) send_action(server, A_LOGIN, TRUE);

    (void) send_string(server, username, strlen(username) + 1, TRUE);
    (void) send_string(server, password, strlen(password) + 1, TRUE);

    (void) get_long(server, &accepted_p, TRUE);

    if (accepted_p)
	return TRUE;
    else
	return (question_dialog("Login failed. Try again?") 
		&& do_login(server, info, browse));
}

/*
  Send displayable level along with init level. Return TRUE if all goes well,
  otherwise return FALSE.
*/

static int
do_init(server, browse)
XDR_TCP *server;
BROWSE *browse;
{
    long ack;

    (void) send_action(server, A_INIT, TRUE);
    
    (void) send_long(server, browse->pyramid->init_level, TRUE);
    (void) send_long(server, browse->pyramid->max_pyramid_level, TRUE);

    (void) get_long(server, &ack, TRUE);
    if (ack)
	return TRUE;
    else
	return FALSE;
}
   

/* Support Functions for WorkProcs, button callbacks and server comm */

/*
  Set the BROWSE list variable. Calling this function configures a browsing
  session to use a local (to the client) text file as a list of images that
  are to be browsed.
*/

static void
browse_local_list(browse, list)
BROWSE *browse;
char *list;
{
    char *filename;
    FILE *fp;
    
    browse->use_local_list_p = FALSE;
    browse->use_remote_list_p = FALSE;
    browse->use_path_p = FALSE;

    if (*list) {
	int i;
	fp = open_list(list);
	if (!fp)
	    return;
	browse->use_local_list_p = TRUE;

	/* count the length of the local list */
	while ((filename = next_filename(fp)))
	    browse->image_list_len++;
	rewind(fp);
	
	browse->image_list = (char **)malloc(browse->image_list_len *
					     sizeof(char *));
	browse->pretty_list = (char **)malloc(browse->image_list_len *
					      sizeof(char *));

	for (i = 0; (filename = next_filename(fp)); i++) {
	    browse->image_list[i] = malloc(strlen(filename) + 1);
	    strcpy(browse->image_list[i], filename);
	    browse->pretty_list[i] = beautify_name(browse->image_list[i]);
	}
	close_list(fp);

	/* This code was removed because the READ_IMAGE_NAMES action is no
	   longer supported by the server. That was an old action used when
	   the server was used to browse on directory at a time. When the
	   list of images to browse was moved over to the client (for display
	   in the `images to browse' window) this message became a useless
	   `feature'. Now the list is stored on the client-side and an image
	   name sent to the server when that particular image is to be
	   browsed. The change was made to the server in 1993, but never made
	   it to this part of the client. jhrg 4/26/96 */
#ifdef NEVER
	send_action(server, A_READ_IMAGE_NAMES, TRUE);
	send_long(server, browse->image_list_len, TRUE);
	for (i = 0; i < browse->image_list_len; i++)
	    send_string(server, browse->image_list[i],
			strlen(browse->image_list[i]), FALSE);
	xdr_tcp_flush_record(server);
#endif
    }
    else
	err_dialog("List must be the pathname of a (local) text file");
}

/*
  Calling this function prompts the user for the date range of images that
  interest them, transmits the request to the image server and configures
  `browse' to use the resulting `remote list' (assuming that the request
  succeeds).
*/

static void
browse_remote_list(server, browse)
XDR_TCP *server;
BROWSE *browse;
{
    char *start = NULL, *end = NULL;
    int valid_p, a, i;
    u_int type;
    
    browse->use_local_list_p = FALSE;
    browse->use_remote_list_p = FALSE;
    browse->use_path_p = FALSE;

    do {
	if (!get_image_criteria(browse->image_server_info, ENTER_BROWSE_DATES,
				&start, &end))
	    return;		/* user hits cancel */
	valid_p = valid_date(start) && valid_date(end);
	if (!valid_p)
	    err_dialog("Start and End dates must be of the form mm/dd/yy");
    } while(!valid_p);

    send_action(server, A_GET_IMAGES_DATES, TRUE);
    send_string(server, start, strlen(start), TRUE);
    send_string(server, end, strlen(end), TRUE);

    a = get_action(server, TRUE);
    switch(a){
    case A_ACK:
	browse->use_remote_list_p = TRUE;
	(void) get_long(server, &browse->image_list_len, TRUE);
	xdr_tcp_flush_record(server);
	browse->image_list = (char **)malloc(browse->image_list_len *
					     sizeof(char *));
	browse->pretty_list = (char **)malloc(browse->image_list_len *
					      sizeof(char *));
	for (i = 0; i < browse->image_list_len; i++) {
	    browse->image_list[i] = get_string(server, NULL,
					       MAX_STR_LEN, FALSE);
	    browse->pretty_list[i] = beautify_name(browse->image_list[i]);
	}
	break;

    case A_ERR: {
	char error_msg[MAX_STR_LEN];
	
	get_string(server, error_msg, MAX_STR_LEN, TRUE);
	err_dialog(error_msg);
	browse_remote_list(server, browse);
	break;
    }
	
    default:
	err_dialog("Unknown return code from image server");
	break;
    }
}

/*
  Set the BROWSE path variable. Calling this function configures a browsing
  session that looks at every image in a directory (named b `path') on the 
  server machine.
*/

static void
browse_remote_path(server, browse, path)
XDR_TCP *server;
BROWSE *browse;
char *path;
{
    browse->use_local_list_p = FALSE;
    browse->use_remote_list_p = FALSE;
    browse->use_path_p = FALSE;
    
    if (*path) {
	if (browse->path)
	    XtFree(browse->path);
	browse->path = XtNewString(path);
	fix_path(browse->path);
	browse->path_sent_p = FALSE; /* mark it so that it will be sent */
	browse->use_path_p = TRUE;
	browse->list = NULL;	/* `list' is a FILE * */
	do_file_path(server, browse);
    }
    else
	err_dialog("Path must be a directory on the image server");
}

/*
  define a region using the mouse.
*/

static long
define_region(browse)
BROWSE *browse;
{
#ifdef NEVER
    FILE *region_file;
#endif
    REGION *tmp_region = init_region(browse->pyramid->region->max_size_x,
				     browse->pyramid->region->max_size_y);
    PIXMAP_MARK *hash_mark = browse->hash_mark;

    if (!tmp_region)
	return FALSE;
    
    if (browse->region_defd_p)
	mark_old_region_hash_marks(browse);
    
    if (read_vertices(browse, tmp_region) < 3) {
	PYRAMID *pyramid = browse->pyramid;
	VERTICES *vertices = tmp_region->vertices;
	
	err_dialog("Regions must have at least three vertices\n");

	if (vertices_p(vertices)) {
	    reset_vertices(vertices);

	    do {
		long x = image_to_screen_x(pyramid, get_vertex_x(vertices));
		long y = image_to_screen_y(pyramid, get_vertex_y(vertices));
		erase_pixmap_mark(hash_mark, x, y);
	    } while(next_vertex(vertices));
	}
	draw_region_hash_marks(browse);
	free_region(tmp_region);
	return FALSE;
    }

    if (a_data.sort_vertices)
	sort_region_vertices(tmp_region);

    free_region(browse->pyramid->user_region);
    browse->pyramid->user_region = tmp_region;
    browse->region_hm_disp_p = TRUE;
    browse->region_sent_p = FALSE; /* newly defined region not yet sent */
    browse->region_defd_p = TRUE;

    return TRUE;
}

static void
use_region(server, browse)
XDR_TCP *server;
BROWSE *browse;
{
    long ack;
    PYRAMID *pyramid = browse->pyramid;
    
    /* other parts of xbrowse access the current region using ->region; set
       it here. When a new image is displayed, ->region is set to
       image_region */

    if (!copy_region(pyramid->region, pyramid->user_region))
	return;

    region_bounding_box(pyramid);

    if (!browse->auto_zoom_p) {	/* don't draw if we're going to zoom
				   imeadiately */
	draw_region_hash_marks(browse);	/* draw new vertices */
	browse->region_hm_disp_p = TRUE;
    }

    /* if the region has just been defined by the user (or if they are
       redefining the user region) then it must be sent to the server.
       However, if the user region has already been sent to the server, just
       tell the server to use what it has stored in its `user_region'
       structure. (the A_SET_REGION action code telss the server to store the
       new data in its user_region structure and the A_USE_REGION action
       tells the server to start using that data */
    
    if (!browse->region_sent_p) {
	send_region(server, pyramid->region);
	ack = get_action(server, TRUE);
	if (ack != A_ACK) {
	    err_dialog(REGION_ACK_ERR_MSG);
	    return ;
	}
	browse->region_sent_p = TRUE;
    }
    
    send_action(server, A_USE_REGION, TRUE);
    ack = get_action(server, TRUE);
    if (ack != A_ACK) {
	err_dialog(USE_REGION_ERR_MSG);
	return ;
    }
    
    if (can_zoom(browse->pyramid)) {
	XtSetSensitive(zoom, True);
    }
    else {
	XtSetSensitive(zoom, False);
    }
    
    XtSetSensitive(region_use, False); /* already used ! */
}

static void
write_region_file(browse)
BROWSE *browse;
{
    static char *filename = NULL;
    char *loc_filename;
    FILE *region_file = NULL;
    
    if (filename)
	loc_filename = get_filename("Region Filename:", filename);
    else
	loc_filename = get_filename("Region Filename:", "");
    
    if (!loc_filename) /* user cancels */
	return;
    
    /* save user's last input for next prompt */
    if (filename)
	XtFree(filename);
    filename = XtNewString(loc_filename);

    loc_filename = make_pathname(a_data.region_path, filename);
    
    if ((region_file = fopen(loc_filename, "w"))) {
	write_region(browse->pyramid->user_region, region_file);
	fclose(region_file);
	XtFree(loc_filename);
    }
    else {
	sys_err_dialog(OPEN_REGION_ERR_MSG);
	XtFree(loc_filename);
	write_region_file(browse); /* keep trying until success or cancel */
    }
}

static long
read_region_file(browse)
BROWSE *browse;
{
    static char *filename = NULL;
    char *loc_filename;
    FILE *region_file;    
    REGION *tmp_region;
    PIXMAP_MARK *hash_mark = browse->hash_mark;

    if (filename)		/* FILENAME is remembered across calls */
	loc_filename = get_filename("Region Filename:", filename);
    else
	loc_filename = get_filename("Region Filename:", "");
    
    if (!loc_filename) /* user cancels */
	return FALSE;

    /* save user's newest entry */
    if (filename)
	XtFree(filename);
    filename = XtNewString(loc_filename);

    loc_filename = make_pathname(a_data.region_path, filename);
    
    if ((region_file = fopen(loc_filename, "r"))) {
	tmp_region =
	    init_region(browse->pyramid->region->max_size_x,
			browse->pyramid->region->max_size_y);
	if (!tmp_region) {
	    sys_err_dialog(REGION_INIT_ERR_MSG);
	    XtFree(loc_filename);
	    return FALSE;	/* quit; no retry */
	}
	
	if (browse->region_defd_p)
	    mark_old_region_hash_marks(browse);
	
	if (read_region(tmp_region, region_file) < 3) {
	    PYRAMID *pyramid = browse->pyramid;
	    VERTICES *vertices = tmp_region->vertices;
	
	    err_dialog("Regions must have at least three vertices\n");

	    if (vertices_p(vertices)) {
		reset_vertices(vertices);

		do {
		    long x=image_to_screen_x(pyramid, get_vertex_x(vertices));
		    long y=image_to_screen_y(pyramid, get_vertex_y(vertices));
		    erase_pixmap_mark(hash_mark, x, y);
		} while(next_vertex(vertices));
	    }
	    draw_region_hash_marks(browse);
	    free_region(tmp_region);
	    XtFree(loc_filename);
	    return read_region_file(browse); /* try again */
	}
	fclose(region_file);

	/* Make sure that the new region is within the current region.
	   There must be no area in the new (user) region not in the current
	   region. This is a hack - regions should be redefined to take
	   advantage of X11 Regions ... */
    	{
	    Region current_region
		= vertices_to_x_region(browse->pyramid,
				       browse->pyramid->region); 
	    Region user_region
		= vertices_to_x_region(browse->pyramid, tmp_region);
	    Region user_but_not_current = XCreateRegion();
	    
	    XSubtractRegion(user_region, current_region, user_but_not_current);
	    if (!XEmptyRegion(user_but_not_current)) {
		err_dialog(REGION_INCLUSIVE_MSG);
		draw_region_hash_marks(browse);
		free_region(tmp_region);
		XtFree(loc_filename);
		XDestroyRegion(user_but_not_current);
		XDestroyRegion(user_region);
		XDestroyRegion(current_region);
		return read_region_file(browse); /* try again */
	    }

	    XDestroyRegion(user_but_not_current);
	    XDestroyRegion(user_region);
	    XDestroyRegion(current_region);
	}
    }
    else {
	sys_err_dialog(OPEN_REGION_ERR_MSG);
	XtFree(loc_filename);
	return read_region_file(browse);
    }

    if (a_data.sort_vertices)
	sort_region_vertices(tmp_region);

    free_region(browse->pyramid->user_region);
    browse->pyramid->user_region = tmp_region;
    browse->region_hm_disp_p = TRUE;
    browse->region_sent_p = FALSE; /* newly defined region not yet sent */
    browse->region_defd_p = TRUE;

    return TRUE;
}


/*
  Read points from the screen (entered via mouse). Button 2 (or any key) ends
  entry. The points are stored in image coordinates.
*/

static long 
read_vertices(browse, region)
BROWSE *browse;
REGION *region;
{
    long done = FALSE;
    long num_vert = 0;
    XEvent e;
    PYRAMID *pyramid = browse->pyramid;
    VERTICES *v = region->vertices;
    PIXMAP_MARK *mark = browse->hash_mark;
    Region current_region;	/* Region is an X11 data type */

    /* erase the current user defined region vertices */

    free_vertices(v);
    v = init_vertices();
    if (!v)
	return FALSE;

    /* current_region is used to ensur that the new region lies within any
       existing region. the server cannot currently handle the case where a
       new region does not fall completely within all the old regions. */

    current_region = vertices_to_x_region(pyramid, pyramid->region);
    
    while (!done) {
	XtAppNextEvent(a_context, &e);

	switch (e.type) {

	case ButtonPress:

	    switch (e.xbutton.button) {
	    case 1:		/* buttons 2 & 3 end input */
		if (e.xbutton.window != XtWindow(the_image))
		    pt_outside_region_message();
		else if (!browse->region_defd_p
		    && (e.xbutton.x < 0
			|| e.xbutton.x >= window_pixmap_x
			|| e.xbutton.y < 0
			|| e.xbutton.y >= window_pixmap_y))
		    pt_outside_region_message();
		else if (browse->region_defd_p && current_region
			 && !XPointInRegion(current_region, e.xbutton.x,
					    e.xbutton.y))
		    pt_outside_region_message();
		else {
		    num_vert++;
		    add_vertex(v, screen_to_image_x(pyramid, e.xbutton.x),
			       screen_to_image_y(pyramid, e.xbutton.y));
		    draw_pixmap_mark(mark, e.xbutton.x, e.xbutton.y);
		    print_location(the_image, &e.xbutton);
		}
		break;

	    case 2:
	    case 3:
		done = TRUE;
		break;
	    }
	    break;

	case KeyPress:
	    done = TRUE;
	    break;

	default:
	    XtDispatchEvent(&e);
	    break;
	}
    }

    if (current_region)
	XDestroyRegion(current_region);
    
    return num_vert;
}

/* A quick little hack function to pop up a message the first time a user
   `colors outside the lines'. There after just beep. */
   
static void
pt_outside_region_message()
{
    static long first_error = TRUE;
    
    if (first_error) {
	err_dialog(REGION_IN_IMAGE_MSG);
	first_error = FALSE;
    }
    else
	XBell(display, 50);
}

void
do_palette(browse, pal_name)
BROWSE *browse;
char *pal_name;
{
    char *loc_filename;
    FILE *palette_fp;

    /* first , look at `static_visual'. If TRUE, return w/o doing anything */
    if (static_visual) {
	err_dialog(CANT_DEF_PALETTE_MSG);
	return;
    }
    
    /* Get the filename of the user's palette file and open it. If the
       palette_path resource is non-NULL, use it. Otherwise use the users
       CWD.  */
    if (a_data.palette_path) {
	char *loc_pathname = make_pathname(a_data.palette_path, pal_name);

	loc_filename = get_filename("Palette Filename:", loc_pathname);
	XtFree(loc_pathname);	/* free memory alloc'd by make_pathname() */
    }
    else
	loc_filename = get_filename("Palette Filename:", pal_name);
    
    if (!loc_filename) /* user cancels */
	return;
    
    palette_fp = fopen(loc_filename, "r");
    if (palette_fp) {
	add_palette_to_colormap(palette_fp, image_pal, FALSE);
	(void) fclose(palette_fp);
    }
    else
	sys_err_dialog(CANT_OPEN_PAL_FILE_MSG, loc_filename);

    XtFree(loc_filename);
}

/*
  Write the image data to a pgm file. The `header' information is written as
  comments at the begining of the file. This assumes thatxbrowse is built on
  a machine that has libpgm.a on it. If an error is detected, return FALSE,
  otherwise return TRUE.

  NB: The return value is useless in this function, but other functions to
  save images in other file formats will certainly (?) have a meaningfull
  return code. I included one here so that all the `save to X type file '
  functions can have the same interface.
*/

static int
write_pgm(image_file, pyramid, max_level)
int image_file;
PYRAMID *pyramid;
long max_level;
{
    int i;
    double tl_lat, tl_lon, br_lat, br_lon;
    FILE *f = fdopen(image_file, "w");
    int argc = 0;
    char *argv[2];
    float res = pyramid->header->resolution * (1 << pyramid->max_image_level
						- pyramid->level);

    argv[0] = ""; argv[1] ="";

    pgm_init(&argc, argv);	/* libpgm claims this is necessary */

    /* in order for some programs to read files with comments in them, you
       must use the ascii file format - hence `TRUE' in this call */
    pgm_writepgminit(f, window_pixmap_x, window_pixmap_y, MAX_PIXEL_VAL,
		     TRUE);
    
    get_lat_lon(pyramid, 0, 0, &tl_lat, &tl_lon);
    get_lat_lon(pyramid, window_pixmap_x - 1, window_pixmap_y - 1,
		&br_lat, &br_lon);
    
    fprintf(f, "# File name: %s\n", browse->filename);
    fprintf(f, "# Image Time: %s\n", pyramid->header->image_name);
    fprintf(f, "# Projection: %s\n", pyramid->header->projection);
    fprintf(f, "# Projection Equations: %s %s\n", pyramid->header->lat_to_y,
	    pyramid->header->lon_to_x);
    fprintf(f, "# Projection Equations: %s %s\n", pyramid->header->y_to_lat,
	    pyramid->header->x_to_lon);
    fprintf(f, "# Resolution: %5.2f\n", res);
    fprintf(f, "# File size (pixels - x, pixels - y): (%d, %d)\n",
	    window_pixmap_x, window_pixmap_y);
    fprintf(f, "# Top left pixel (lon, lat): (%6.3f, %6.3f)\n",
	    tl_lon, tl_lat);
    fprintf(f, "# Bottom right pixel (lon, lat): (%6.3f, %6.3f)\n",
	    br_lon, br_lat);
    fprintf(f,"# Data Mapping: %s %s\n", pyramid->header->data_units,
	    pyramid->header->data_eqn);

    /* allow anybpdy to save up to and including `max_level' resolution data,
       after that they must have a valid account (for billing). If someone
       wants to save and has no account, then reduce the image t a level they
       can save */
    if (pyramid->level > max_level) {
	char *row, *red_row = XtMalloc(window_pixmap_x);
	int j, k, l, samp;

	/* ceil_level() was designed for the 2D case; 2nd arg is redundant */
	samp = ((pyramid->level > ceil_level(window_pixmap_x, window_pixmap_x))
		? 1 << (pyramid->level - max_level)
		: 1 << (ceil_level(window_pixmap_x,window_pixmap_x)-max_level));
	
	for (i = 0; i < window_pixmap_y; i += samp) {
	    row = get_ximage_row(pyramid->raw_image, i);
	    for (j = 0; j < window_pixmap_x; j += samp)
		for (k = 0; k < samp; k++)
		    *(red_row + j + k) = *(row + j);
	    for (l = 0; l < samp; l++)
		pgm_writepgmrow(f, red_row, window_pixmap_x, MAX_PIXEL_VAL,
				TRUE); 
	}
	
	XtFree(red_row);
    }
    else
	for (i = 0; i < window_pixmap_y; i++)
	    pgm_writepgmrow(f, get_ximage_row(pyramid->raw_image, i),
			    window_pixmap_x, MAX_PIXEL_VAL, TRUE);

    fclose(f);
    
    return TRUE;
}

/*
  I pass pyramid and region separately so that I can make an X11 Region for
  different `region''s that are stored in pyramid (e.g., the current region,
  user_region, ...).
*/

static Region
vertices_to_x_region(pyramid, region)
PYRAMID *pyramid;
REGION *region;
{
    long num_vert = 0;
    XPoint *points;
    VERTICES *vertices = region->vertices;
    
    if (vertices_p(vertices)) {
	reset_vertices(vertices);

	do {
	    num_vert++;
	} while (next_vertex(vertices));
	
	points = (XPoint *)XtMalloc(num_vert * sizeof(XPoint));

	num_vert = 0;
	reset_vertices(vertices);
	do {
	    (points + num_vert)->x = image_to_screen_x(pyramid,
						    get_vertex_x(vertices));
	    (points + num_vert)->y = image_to_screen_y(pyramid,
						    get_vertex_y(vertices));
	    num_vert++;
	} while (next_vertex(vertices));
	
	return XPolygonRegion(points, num_vert, WindingRule);
    }
    else
	return (Region)NULL;
}

static void
interrupt_handler(sig)
long sig;
{
    browse->no_question_exit_p = TRUE;
    quit_button(toplevel, NULL, NULL);
}

void
print_image_name(pyramid)
PYRAMID *pyramid;
{
    char value[256];

    sprintf(value, "Time: %s", pyramid->header->image_name); 
    
    XtVaSetValues(image_name, XtNlabel, value, NULL);

    sprintf(value, "Projection: %s", pyramid->header->projection);
    
    XtVaSetValues(projection, XtNlabel, value, NULL);

    sprintf(value, "Platform: %s", pyramid->header->sat_type);
    
    XtVaSetValues(satellite_type, XtNlabel, value, NULL);

    sprintf(value, "Resolution: %5.2f km/pixel",
	    pyramid->header->resolution);
    
    XtVaSetValues(max_res, XtNlabel, value, NULL);
    
}

/*
  These two functions let the usr specify their own text for the widgets and
  avoid nasty bugs due to overwriting storage space. This could be handled
  better by a moving all strings to the resource file.
*/

void 
print_level(pyramid, adding_p)
PYRAMID *pyramid;
long adding_p;			/* still adding data to current level ? */
{
    char value[64];
    float res = pyramid->header->resolution * (1 << pyramid->max_image_level
						- pyramid->level);
    if (adding_p)
	sprintf(value, "Displaying Resolution: %5.2f km/pixel", res);
    else
	sprintf(value, "Displayed Resolution: %5.2f km/pixel", res);

    XtVaSetValues(current_res, XtNlabel, value, NULL);
    XFlush(display);
}

/*
  Display the cost of all images saved so far.
*/

void
print_cost(browse)
BROWSE *browse;
{
    char value[64];
    
    sprintf(value, "Session Cost: $%5.2f",
	    browse->total_cost + browse->image_cost); 
    XtVaSetValues(cost, XtNlabel, value, NULL);
}

static void
set_buttons_done()
{
    XtVaSetValues(image_name, XtNlabel, "Done", NULL);
    XtVaSetValues(projection, XtNlabel, "", NULL);
    XtVaSetValues(satellite_type, XtNlabel, "", NULL);
    XtVaSetValues(max_res, XtNlabel, "", NULL);
    XtVaSetValues(current_res, XtNlabel, "", NULL);
    XtVaSetValues(cursor, XtNlabel, "", NULL);	
    
    XtSetSensitive(pause_tx, False);
    XtSetSensitive(save, False);
    XtSetSensitive(next, False);
    XtSetSensitive(log_entry, False);
    XtSetSensitive(zoom, False);
    XtSetSensitive(auto_zoom, False);	    
    XtSetSensitive(xbt_display, False);
    XtSetSensitive(auto_region, False);
    XtSetSensitive(region_use, False);
    XtSetSensitive(region_define, False);
#ifdef NEVER
    XtSetSensitive(region_write, False);
    XtSetSensitive(region_read, False);	
#endif
    XtSetSensitive(palette, False);
}


/*
  A date is valid if it has the following format: mm/dd/yy. mm must be
  between 1 and 12, dd must be between 1 and 31, 1 and 30 or 1 and 28/29
  depending. yy can be any number.
*/

static int
valid_date(d)
char *d;
{
    int mm, dd, yy;

    /* months[0] is `normal' year, months[1] is leap year */
    static int months[2][12] =
         {{31, 28, 31, 30, 31, 30, 31, 30, 31, 30, 30 ,30},
	  {31, 29, 31, 30, 31, 30, 31, 30, 31, 30, 30 ,30}};

    sscanf(d, "%d/%d/%d", &mm, &dd, &yy);
    
    mm--;
    dd--;
    yy += BASE_YEAR;
    
    if (mm < 0 || mm > 11)
	return FALSE;

    if (dd < 0 || dd > months[LEAP_YEAR(yy)][mm])
	return FALSE;

    return TRUE;
}

static int
browse_next_level(server, browse)
XDR_TCP *server;
BROWSE *browse;
{
    if ((browse->pyramid->level == browse->pyramid->init_level)
	&& browse->auto_region_p && browse->region_defd_p) {
	
	use_region(server, browse);
	
	if (browse->auto_zoom_p) { /* auto zoom */
	    do_zoom(server, browse);
	    XtSetSensitive(zoom, False);
	    XtSetSensitive(pause_tx, True);
	}
    }
    
    return (pyramid_next_level(browse->pyramid));
}

/*
  Given a name like: `LDB1:[1989.4]I89096074335.PVU_Z;1' extract the various
  date and time components and create a string that includes both the
  original name and the extrated bits.

  NB: the caller must handle deallocation of memory.
*/

static char *
beautify_name(name)
char *name;
{
    char *new_name;
    char *tmp, dy[6], tm[6];
    float imgdy, imgtm;

    new_name = XtMalloc((strlen(name) + 1 + 32) * sizeof(char));

    /* I added the code for UNIX below since it seemed like a useful and
       simple addition. jhrg 4/26/96 */
    tmp = strchr(name, ']');	/* try VMS file names first (nostalgia) */
    if (!tmp)
	tmp = strrchr(name, '/'); /* try UNIX file names */

    if (tmp) {
	tmp += 2;		/* advance past `]' of `/' and letter */
	memcpy(dy, tmp, 5);
	dy[5] = '\0';
	memcpy(tm, tmp + 5, 6);
	if (tm[2] == 'F')	/* hack for broken image file names */
	    memcpy(tm + 2, "0000", 5);
	else
	    tm[6] = '\0';

	imgdy = atof(dy);
	imgtm = atof(tm);
    
	sprintf(new_name, "%s %s", name, get_dsp_pretty_name((float)imgdy,
							     (float)imgtm));
    }
    else 
	sprintf(new_name, "%s", name);
    
    return new_name;
}

static char *months[]={"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
			      "Aug", "Sep", "Oct", "Nov", "Dec"};

static char *
get_dsp_pretty_name(imgdy, imgtm)
float imgdy, imgtm;
{
    static char name[32];
    char date[32];
    int hours, min, secs, month, day, year;
    double jdate, fsecs;
    
    /* imgdy is yyddd. to find the julian date, find the julian date of
       01/01/yy and add (ddd - 1) */
    
    sprintf(date, "01/01/%02d", (int)imgdy / 1000);
    jdate = ((double)julian_day(date)
	     + (double)((int)imgdy - ((int)imgdy / 1000 * 1000)) - 1);

    /* only care about month, day and year - get h:m:s from imgtm */
    
    calcdate(jdate, &month, &day, &year, &hours, &min, &fsecs);
	
    hours = imgtm / 10000;
    min = (imgtm - hours * 10000) / 100;
    secs = (int)(imgtm - hours * 10000 - min * 100);
    sprintf(name, "%d-%s-%d %02d:%02d:%02d GMT", day, months[--month], year,
	    hours, min, secs);

    return name;
}
