/*
  Copyright 1992 University of Rhode Island
 
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose is hereby granted without fee, provided that
  the above copyright notice appear in all copies and that both that
  copyright notice and this permission notice appear in supporting
  documentation, and that the name of U.R.I. not be used in advertising or
  publicity pertaining to distribution of the software without specific,
  written prior permission.  U.R.I. makes no representations about the
  suitability of this software for any purpose.  It is provided "as is"
  without express or implied warranty.
 
  U.R.I. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.R.I.
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
  Perform the steps necessary for a client to open a socket connection
  to a TCP/IP based server.

  Originally based on code from Stevens, ``UNIX Network Programming'', pp.
  397-99. 

  jhrg - 11/2/90
*/

/*
 * Change log:
 * $Log: client-connect.c,v $
 * Revision 2.2  1992/11/10  17:44:22  jimg
 * Fixed prototype bugs.
 * Fixed compiler symbols (unix, ...) so that they work with gcc -ansi
 *
 * Revision 2.1  1992/10/30  18:41:18  jimg
 * Removed arg list formal parameter list dfinitions.
 *
 * Revision 2.0  1992/09/28  14:02:33  jimg
 * Release version.
 *
 * Revision 2.0  1992/09/28  14:02:33  jimg
 * Release version.
 *
 * Revision 1.3  1992/09/23  19:36:41  jimg
 * Added copyright notice.
 *
 * Revision 1.2  1992/07/30  21:14:39  jimg
 * Fixed bug where VMS Channel descriptor was stored in an
 * int rrather than a short.
 *
 * Revision 1.1  1992/03/10  16:12:04  jimg
 * Initial revision
 */

#if !(defined(unix) || defined(__unix)) && !defined(vms)
error "Either vms or unix must be defined"
#elif (defined(unix) || defined(__unix)) && defined(vms)
error "Only one of vms or unix can be defined"
#endif

#include <stdio.h>

#if (defined(unix) || defined(__unix)) 

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#elif defined(vms)

#include <stdlib.h>
#include <descrip>
#include <ssdef>
#include <iodef>
#include "multinet_root:[multinet.include.sys]types.h"
#include "multinet_root:[multinet.include.sys]socket.h"
#include "multinet_root:[multinet.include.vms]inetiodef.h"
#include "multinet_root:[multinet.include.netinet]in.h"
#include "multinet_root:[multinet.include]netdb.h"

#endif

#ifndef INADDR_NONE
#define INADDR_NONE	0xffffffff /* should be in <netinet/in.h */
#endif

#include "errmsg.h"

/*
  The following globals are available to the caller, if desired.
*/

struct sockaddr_in dest_srv_addr; /* server's Internet socket address */
struct sockaddr_in clt_addr;	/* this clients addr */

/*
  Return socket descriptor if OK, else -1. This version prints its error
  messages to stderr using function defined in err-sys.c.

  host: name or dotted decimal address of other system.
  service: name of service being requested; can be null iff port > 0
  port: if == 0, nothing special - use port # of service; if < 0, bind a local
        reserved port; if > 0, it's the port # of the server
	(host-byte-order).
  err_report: error reporting function.
*/

#if defined(unix) || defined(__unix)
int
#elif defined(vms)
short
#endif
client_connect(host, service, port, err_report)
char *host;
char *service;
int port;
void (*err_report)();
{
#if defined(unix) || defined(__unix)
    int fd;
#elif defined(vms)
    short fd;
#endif
    int resvport;
    unsigned long inaddr;
    struct servent *sp;
    struct hostent *hp;
    
    /* Initialize the server's Internet address structure. Store the
       actual 4-byte Internet address and the 2 byte port # below */
    
    memset((char *)&dest_srv_addr, '\0', sizeof(dest_srv_addr));
    dest_srv_addr.sin_family = AF_INET;
    
    if (service != NULL) {
	if ((sp=getservbyname(service, "tcp")) == NULL) {
	    (*err_report)("service `%s' not in /etc/services database",
			  service); 
	    return -1;
	}
	
	if (port > 0)
	    dest_srv_addr.sin_port = htons(port); /* caller's value */
	else
	    dest_srv_addr.sin_port = sp->s_port; /* service's value */
    }
    else if (port < 0) {
	(*err_report)("Cannot use port %d", port);
	return -1;
    }
    else
	dest_srv_addr.sin_port = htons(port);
    
    /* First try to convert the host name as a dotted decimal number.
       Only if that fails do we call gethostbyname(). */
    
    if ((inaddr=inet_addr(host)) != INADDR_NONE) {	/* it's dotted dec. */
	memcpy((char *) &dest_srv_addr.sin_addr, (char *)&inaddr,
	      sizeof(inaddr));
    } else {
	if ((hp=gethostbyname(host)) == NULL) {
	    (*err_report)("Could not resolve host name: `%s'", host);
	    return(-1);
	}
	
	memcpy((char *)&dest_srv_addr.sin_addr, hp->h_addr, hp->h_length);
    }
    
    if (port >= 0) {
	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	    (*err_report)("Cannot create tcp socket");
	    return -1;
	}
    }
    else {
	resvport = IPPORT_RESERVED -1;
	if ((fd = rresvport(&resvport)) < 0) {
	    (*err_report)("Cannot get a reserved tcp port");
	    return -1;
	}
    }
    
    /* Connect to the server. */
    
    if (connect(fd, (struct sockaddr *) &dest_srv_addr,
		sizeof(dest_srv_addr)) < 0) {
	(*err_report)("Cannot connect to server at `%s'", host);
	close(fd);
	return (-1);
    }
    
    return(fd);			/* all OK */
}
