/*
  Copyright 1991, 1992 University of Rhode Island
 
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose is hereby granted without fee, provided that
  the above copyright notice appear in all copies and that both that
  copyright notice and this permission notice appear in supporting
  documentation, and that the name of U.R.I. not be used in advertising or
  publicity pertaining to distribution of the software without specific,
  written prior permission.  U.R.I. makes no representations about the
  suitability of this software for any purpose.  It is provided "as is"
  without express or implied warranty.
 
  U.R.I. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.R.I.
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
  This file contains the functions used with the dialog boxes. Each dialog
  box has a function that creates it (creates the popup shell and the dialog
  widget), and shows the dialog. In addition, each type dialog also has
  functions that handle the button(s) in that dialog and ensure that any
  information entered into the dialog's value field gets passed back to the
  caller.
  
  Each dialog box is destroyed after each use. This solves the problem where an
  `old' dialog box would be used with a static text item that did not fit
  with the dialog's borders. It looked stupid; this is slow, but only by a
  computer's standards, not a person's.

  NB:
  Fix the above by making boxes dynamically resize - this will also let text
  entry fields resize (and their enclosing widgets) when long text items are
  entered.
*/

/*
 * $Log: dialog.c,v $
 * Revision 3.6  1993/02/02  16:12:32  jimg
 * Added conditional include for debugging malloc header.
 *
 * Revision 3.5  1992/11/19  21:14:30  jimg
 * Added accelerator code.
 * Keyboard focus in text widgets.
 * Keyboard focus switch with mouse click.
 *
 * Revision 3.4  1992/11/04  18:34:03  jimg
 * Fixed accelerators for login and criteria dialog boxes. Dialog class
 * boxes are still partially broken
 *
 * Revision 3.3  1992/10/30  18:16:13  jimg
 * Removed prototypes (used unprotoize -c -I/usr/local/include dialog.c)
 * after glenn@pimms.mit.edu reported compilation failure on a sun due to
 * const. I assume that the compiler does not recognize prototypes as well.
 *
 * Revision 3.2  1992/10/19  14:57:32  jimg
 * Added code to handle `query-based' access to the server.
 *
 * Revision 3.2  1992/10/19  14:57:32  jimg
 * Added code to handle `query-based' access to the server.
 *
 * Revision 3.1  1992/10/14  13:19:00  jimg
 * Fixed use of accelerators. Many minor bug fixes.
 *
 * Revision 3.0  1992/07/28  15:24:44  jimg
 * All sources for xbrowse modified to use the libcomm.a and liberrmsg.s
 * a
 *
 * Revision 1.3  1992/07/23  17:31:31  jimg
 * Merged changes from xchooser 1.0 during creation of xbrowse 3.0
 *
 * Revision 1.2.1.2  1992/07/22  17:41:30  jimg
 * Removed XsraSelFile dialog and added FWF FileCompletion type dialogs
 * for both opening and saving files.
 *
 * Revision 1.2.1.1  1992/07/09  21:40:47  jimg
 * Modified callback functions so that globals are not used - client_data
 * is now used.
 * Fixed err_dialog, ... bug when compiled using ansi C.
 * Use accelerators (not actions) when implementing keyboard shortcuts.
 *
 * Revision 1.2  1992/06/19  20:48:12  jimg
 * Modified dialogs so that they  use fewer globals.
 * Added login dialog box.
 * Removed `get filename' dialog box - use selFilename library code.
 * Expanded msg_dialog to question_dialog. Use err_dialog for `ok dialog'.
 * Fixed up the error dialogs.
 *
 * Revision 1.1  1992/03/23  20:39:01  jimg
 * Initial revision
 */

static char rcsid[]="$Id: dialog.c,v 3.6 1993/02/02 16:12:32 jimg Exp jimg $";

#include <stdio.h>
#include <errno.h>
#include <unistd.h>

#if defined(__STDC__)
#include <stdarg.h>
#else
#include <varargs.h>
#endif /* __STDC__ */

#include <sys/types.h>
#include <sys/stat.h>

#include <X11/Intrinsic.h>	/* toolkit */
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>

#include "FileComp.h"

#include <X11/Shell.h>

#include <X11/Xaw/Command.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Text.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Form.h>

#ifdef MALLOC_DEBUG
#include <malloc.h>
#endif 

#ifdef NEVER
#include "FileComp.h"
#endif
#include "dialog.h"
#include "errmsg.h"

#define Y_OFFSET 15		/* locate box so pointer is Y_OFFSET pixels
				   inside box (from the bottom) */
#define X_OFFSET (0)		/* place pointer off center */

#ifndef MAXPATHLEN
#define MAXPATHLEN 256
#endif

#if XtSpecificationRelease < 5

#if NeedFunctionPrototypes
typedef void* XtPointer;
#else
#define XtPointer char *
#endif /* Need function prototypes */

#endif /* Not R5  */

/* structs used to bundle params passed to callbacks */

struct criteria_bundle {
    Widget start_text, end_text, dialog_form;
    int done, bpe;
};

struct login_bundle {
    Widget username_text, passwd_text, dialog_form;
    int done, bpe;
};

/* Globals */

/* errno.h doesn't declare this ! */
extern char *sys_errlist[];

/* variables declared in xbrowse.c */
extern Widget toplevel;		/* ancestor of all widgets */
extern XtAppContext a_context;	/* for portability */
extern Display *display;
extern long display_h;
extern long display_w;

static void popup_err_dialog();
static void position_dialog();
static void dialog_cb(); 
static void criteria_ok_dialog_cb();
static void login_dialog_cb();

/*
  Popup a dialog box that prompts for a user name and password. The two
  parameters `username', and `password are pointers to strings (char **). If
  the pointer `username' is non-null, then its value is displayed as a
  default by the dialog box. `password' is *not* returned in encrypted
  format. If the button OK is chosen the function returns TRUE, otherwise it
  returns FALSE.  The string text1 is displayed above the username/password
  labels while the string text2 is displayed below them. text1 is useful for
  messages like `login to ...' while text2 is a good place to display an info
  message.

  NB: This function allocates storage for the strings username and password.
  If those pointers are non-null, then existing storage will first be
  deallocated using XtFree().
*/

int
login(text1, text2, username, password)
     char *text1;
     char *text2;
     char **username;
     char **password;
{
    int login_cancel;
    XEvent e;
    Cursor dot;
    Dimension ok_x, cancel_x, dialog_x;
    Widget login_pshell;	/* popup shell for the server login dialog */
    Widget login_form;		/* form for the labels, text & buttons */
    Widget label1, label2;	/* labels to displat `text1' & `text2' */
    Widget username_label, passwd_label; /* more text labels */
    Widget username_text, passwd_text; /* asciiText widgets */
    Widget ok, cancel;		/* buttons */
    struct login_bundle login_ok;
    
    /* create dialog box's widgets */

    login_pshell = XtVaCreatePopupShell
	("loginShell", transientShellWidgetClass, toplevel, NULL);
    
    login_form = XtVaCreateManagedWidget
	("loginForm", formWidgetClass, login_pshell, NULL); 
    
    label1 = XtVaCreateManagedWidget
	("label1", labelWidgetClass, login_form, NULL);

    username_label = XtVaCreateManagedWidget
	("usernameLabel", labelWidgetClass, login_form, NULL);

    username_text = XtVaCreateManagedWidget
	("usernameText", asciiTextWidgetClass, login_form,
	 XtNstring, *username ? *username : "", XtNeditType, XawtextEdit,
	 NULL); 
    
    passwd_label = XtVaCreateManagedWidget
	("passwdLabel", labelWidgetClass, login_form, NULL);

    passwd_text = XtVaCreateManagedWidget
	("passwdText", asciiTextWidgetClass, login_form,
	 XtNstring, "", XtNeditType, XawtextEdit, XtNecho, FALSE, NULL);

    label2 = XtVaCreateManagedWidget
	("label2", labelWidgetClass, login_form, NULL);

    ok = XtVaCreateManagedWidget
	("Ok", commandWidgetClass, login_form, NULL);
    
    XtAddCallback(ok, XtNcallback, login_dialog_cb, &login_ok);
    
    cancel = XtVaCreateManagedWidget
	("Cancel", commandWidgetClass, login_form, NULL);
    
    XtAddCallback(cancel, XtNcallback, dialog_cb, &login_cancel);

    /* The Unrealize, set values, realize sequence ensures that the error
       dialog will be the proper size given TEXT. Without this, the dialog
       would not resize for each new string */
     
    XtUnrealizeWidget(login_pshell);
    XtVaSetValues(label1, XtNlabel, text1, NULL);
    XtVaSetValues(label2, XtNlabel, text2, NULL);
    XtRealizeWidget(login_pshell);
    
    XtVaGetValues(ok, XtNwidth, &ok_x, NULL);
    XtVaGetValues(cancel, XtNwidth, &cancel_x, NULL);
    XtVaGetValues(login_form, XtNwidth, &dialog_x, NULL);

    XtUnrealizeWidget(login_pshell);
    XtVaSetValues(ok, XtNhorizDistance,
		  (long)((dialog_x - ok_x - cancel_x - 4) / 2), NULL);
    XtRealizeWidget(login_pshell);

    position_dialog(login_pshell, (Position)(dialog_x / 2) + X_OFFSET,
		    (Position)Y_OFFSET);

    XtPopup(login_pshell, XtGrabExclusive);

    /* to get accelerators to work you *must* use the widget name (first
       argumet to XtVaCreat...()) in the resource file and the variable name
       here. In addition, the text widget(s) must also have the accelerators
       explicitly installed (and those accelerators *must* be specified with
       #override). See also the resource file. */

    XtInstallAccelerators(username_text, ok);
    XtInstallAccelerators(username_text, cancel);
    XtInstallAccelerators(passwd_text, ok);
    XtInstallAccelerators(passwd_text, cancel);
    XtInstallAccelerators(login_form, ok);
    XtInstallAccelerators(login_form, cancel);
    
    /* bind the dot cursor */
    dot = XCreateFontCursor(display, XC_dot);
    XDefineCursor(display, XtWindow(login_pshell), dot);
    XFreeCursor(display, dot);

    /* flags set by ok & cancel button callbacks */
    login_ok.bpe = FALSE;
    login_ok.done = FALSE;
    login_ok.username_text = username_text;
    login_ok.passwd_text = passwd_text;
    login_ok.dialog_form = login_form;
    login_cancel = FALSE;

    /* keyboard focus to start_text */
    XtSetKeyboardFocus(login_form, username_text);

    /* display caret only in widget with focus */
    XawTextDisplayCaret(username_text, TRUE);
    XawTextDisplayCaret(passwd_text, FALSE);
    
    /* event processing loop - check callback flags */
    while (!(login_ok.done || login_cancel)) {
	XtAppNextEvent(a_context, &e);

	switch (e.type) {
	    
	case ButtonPress:
	    if (e.xbutton.window == XtWindow(ok))
		login_ok.bpe = TRUE;
	    else if (e.xbutton.window == XtWindow(username_text)) {
		XtSetKeyboardFocus(login_form, username_text);
		XawTextDisplayCaret(username_text, TRUE);
		XawTextDisplayCaret(passwd_text, FALSE);
	    }
	    else if (e.xbutton.window == XtWindow(passwd_text)) {
		XtSetKeyboardFocus(login_form, passwd_text);
		XawTextDisplayCaret(username_text, FALSE);
		XawTextDisplayCaret(passwd_text, TRUE);
	    }
	    break;
	}
	
	XtDispatchEvent(&e);
    }

    /* deallocate space used by username or password */
    if (*username)
	XtFree(*username);
    if (*password)
	XtFree(*password);
    
    /* When the user hits OK button, set username and password to the values
       in the corresponding text edit boxes. If the user hits cancel, set the
       strings to null */
    if (login_ok.done) {
	char *value;
	
	XtVaGetValues(username_text, XtNstring, &value, NULL);
	*username = XtNewString(value);
	XtVaGetValues(passwd_text, XtNstring, &value, NULL);
	*password = XtNewString(value);
    }
    else {
	*username = XtNewString("");
	*password = XtNewString("");
    }

    /* remove the dialog */
    XtPopdown(login_pshell);
    XtDestroyWidget(login_pshell);

    /* return true when the user hits OK button */
    return login_ok.done;
}

/*
  Dialog box used to prompt for dates of iamges to browse. Returns FALSE if
  an error is detected of if the user hits `Cancel', otherwise returns TRUE.
  The start and end dates are returned using the char *'sstart and end.
*/

int
get_image_criteria(text, text2, start, end)
     char *text;
     char *text2;
     char **start;
     char **end;
{
    int criteria_cancel;
    XEvent e;
    Cursor dot;
    Dimension ok_x, cancel_x, dialog_x;
    Widget criteria_pshell;	/* popup shell for the searchcriteria dialog */
    Widget criteria_form;    	/* form for the labels, text & buttons */
    Widget label1, label2;	/* label to displat `text' & `text2'*/
    Widget start_label, end_label; /* more text labels */
    Widget start_text, end_text; /* asciiText widgets */
    Widget ok, cancel;		/* buttons */
    struct criteria_bundle criteria_ok;
    
    /* create dialog box's widgets */

    criteria_pshell = XtVaCreatePopupShell
	("criteriaShell", transientShellWidgetClass, toplevel, NULL);
    
    criteria_form = XtVaCreateManagedWidget
	("criteriaForm", formWidgetClass, criteria_pshell, NULL); 
    
    label1 = XtVaCreateManagedWidget
	("label1", labelWidgetClass, criteria_form, NULL);

    label2 = XtVaCreateManagedWidget
	("label2", labelWidgetClass, criteria_form, NULL);

    start_label = XtVaCreateManagedWidget
	("startLabel", labelWidgetClass, criteria_form, NULL);

    start_text = XtVaCreateManagedWidget
	("startText", asciiTextWidgetClass, criteria_form,
	 XtNstring, *start ? *start : "", XtNeditType, XawtextEdit, NULL); 
    
    end_label = XtVaCreateManagedWidget
	("endLabel", labelWidgetClass, criteria_form, NULL);

    end_text = XtVaCreateManagedWidget
	("endText", asciiTextWidgetClass, criteria_form,
	 XtNstring, *end ? *end : "", XtNeditType, XawtextEdit, NULL);

    ok = XtVaCreateManagedWidget
	("Ok", commandWidgetClass, criteria_form, NULL);

    XtAddCallback(ok, XtNcallback, criteria_ok_dialog_cb, &criteria_ok);
    
    cancel = XtVaCreateManagedWidget
	("Cancel", commandWidgetClass, criteria_form, NULL);
    
    XtAddCallback(cancel, XtNcallback, dialog_cb, &criteria_cancel);
    
    /* The Unrealize, set values, realize sequence ensures that the error
       dialog will be the proper size given TEXT. Without this, the dialog
       would not resize for each new string */
     
    XtUnrealizeWidget(criteria_pshell);
    XtVaSetValues(label1, XtNlabel, text, NULL);
    XtVaSetValues(label2, XtNlabel, text2, NULL);
    XtRealizeWidget(criteria_pshell);
    
    XtVaGetValues(ok, XtNwidth, &ok_x, NULL);
    XtVaGetValues(cancel, XtNwidth, &cancel_x, NULL);
    XtVaGetValues(criteria_form, XtNwidth, &dialog_x, NULL);

    XtUnrealizeWidget(criteria_pshell);
    XtVaSetValues(ok, XtNhorizDistance,
		  (long)((dialog_x - ok_x - cancel_x - 4) / 2), NULL);
    XtRealizeWidget(criteria_pshell);

    position_dialog(criteria_pshell, (Position)(dialog_x / 2) + X_OFFSET,
		    (Position)Y_OFFSET);

    XtPopup(criteria_pshell, XtGrabExclusive);

    XtInstallAccelerators(start_text, ok);
    XtInstallAccelerators(start_text, cancel);
    XtInstallAccelerators(end_text, ok);
    XtInstallAccelerators(end_text, cancel);    
    XtInstallAccelerators(criteria_form, ok);
    XtInstallAccelerators(criteria_form, cancel);
    
    /* bind the dot cursor */
    dot = XCreateFontCursor(display, XC_dot);
    XDefineCursor(display, XtWindow(criteria_pshell), dot);
    XFreeCursor(display, dot);

    /* flags set by ok & cancel button callbacks */
    criteria_ok.bpe = FALSE;	/* button press event */
    criteria_ok.done = FALSE;
    criteria_ok.start_text = start_text;
    criteria_ok.end_text = end_text;
    criteria_ok.dialog_form = criteria_form;
    criteria_cancel = FALSE;

    /* keyboard focus to start_text */
    XtSetKeyboardFocus(criteria_form, start_text);

    /* display caret only in widget with focus */
    XawTextDisplayCaret(start_text, TRUE);
    XawTextDisplayCaret(end_text, FALSE);
    
    /* event processing loop - check callback flags */
    /* since keyboard focus is set to one of the text widgets, button press
       events are caught here so that users can click in another text widget
       to direct input there. Ideally this code would use EnteryNotify
       events, but those seem to be caught by some other code (maybe the text
       widget's accept_focus method?). */
    
    while (!(criteria_ok.done || criteria_cancel)) {
	XtAppNextEvent(a_context, &e);

	switch (e.type) {

	case ButtonPress:
	    if (e.xbutton.window == XtWindow(ok))
		criteria_ok.bpe = TRUE;
	    else if (e.xbutton.window == XtWindow(start_text)) {
		XtSetKeyboardFocus(criteria_form, start_text);
		XawTextDisplayCaret(start_text, TRUE);
		XawTextDisplayCaret(end_text, FALSE);
	    }
	    else if (e.xbutton.window == XtWindow(end_text)) {
		XtSetKeyboardFocus(criteria_form, end_text);
		XawTextDisplayCaret(start_text, FALSE);
		XawTextDisplayCaret(end_text, TRUE);
	    }
	    break;
	}
	
	XtDispatchEvent(&e);
    }

    /* deallocate space used by username or password */
    if (*start)
	XtFree(*start);
    if (*end)
	XtFree(*end);
    
    /* When the user hits OK button, set username and password to the values
       in the corresponding text edit boxes. If the user hits cancel, set the
       strings to null */
    if (criteria_ok.done) {
	char *value;
	
	XtVaGetValues(start_text, XtNstring, &value, NULL);
	*start = XtNewString(value);
	XtVaGetValues(end_text, XtNstring, &value, NULL);
	*end = XtNewString(value);
    }
    else {
	*start = XtNewString("");
	*end = XtNewString("");
    }

    /* remove the dialog */
    XtPopdown(criteria_pshell);
    XtDestroyWidget(criteria_pshell);
    
    /* return true when the user hits OK button */
    return criteria_ok.done;
}

/*
  Popup a dialog so that the user can enter a string. `text' is a string to
  display in the dialog box above the text entry area. `def' is a default
  string to put in the text entry area. get_string() returns either a pointer
  to the user's string (allocated using XtMalloc()) or NULL depending on
  whether the user clicked OK or CANCEL.
*/

char * 
get_x11_string(text, def)
     char *text;
     char *def;
{
    XEvent e;
    Cursor dot;
    char *user_string;
    Widget get_string_pshell, get_string_dialog;
    Widget label, string, ok, cancel;
    Dimension ok_x, cancel_x, dialog_x;
    int get_string_ok, get_string_cancel;
    
    get_string_pshell = XtVaCreatePopupShell
	("Get String", transientShellWidgetClass, toplevel, NULL);
    
    get_string_dialog = XtVaCreateManagedWidget
	("getString", formWidgetClass, get_string_pshell, NULL); 
    
    label = XtVaCreateManagedWidget
	("label", labelWidgetClass, get_string_dialog, NULL);

    string = XtVaCreateManagedWidget
	("String", asciiTextWidgetClass, get_string_dialog,
	 XtNstring, def ? def : "", XtNeditType, XawtextEdit, NULL); 

    ok = XtVaCreateManagedWidget
	("Ok", commandWidgetClass, get_string_dialog, NULL);
    
    XtAddCallback(ok, XtNcallback, dialog_cb, &get_string_ok);
    
    cancel = XtVaCreateManagedWidget
	("Cancel", commandWidgetClass, get_string_dialog, NULL);
    
    XtAddCallback(cancel, XtNcallback, dialog_cb, &get_string_cancel);

    /* The Unrealize, set values, realize sequence ensures that the error
       dialog will be the proper size given TEXT. Without this, the dialog
       would not resize for each new string */
     
    XtUnrealizeWidget(get_string_pshell);
    XtVaSetValues(label, XtNlabel, text, NULL);
    XtRealizeWidget(get_string_pshell);

    XtVaGetValues(ok, XtNwidth, &ok_x, NULL);
    XtVaGetValues(cancel, XtNwidth, &cancel_x, NULL);
    XtVaGetValues(get_string_dialog, XtNwidth, &dialog_x, NULL);

    XtUnrealizeWidget(get_string_pshell);
    XtVaSetValues(ok, XtNhorizDistance,
		  (long)((dialog_x - ok_x - cancel_x - 4) / 2), NULL);
    XtRealizeWidget(get_string_pshell);
    
    position_dialog(get_string_pshell, (Position)(dialog_x / 2) + X_OFFSET,
		    (Position)Y_OFFSET);

    XtPopup(get_string_pshell, XtGrabExclusive);

    XtInstallAccelerators(string, ok);
    XtInstallAccelerators(string, cancel);    
    XtInstallAccelerators(get_string_dialog, ok);
    XtInstallAccelerators(get_string_dialog, cancel);
    
    dot = XCreateFontCursor(display, XC_dot);
    XDefineCursor(display, XtWindow(get_string_pshell), dot);
    XFreeCursor(display, dot);

    /* return calue when one of these is TRUE */
    get_string_ok = FALSE; get_string_cancel = FALSE;

    while (!get_string_ok && !get_string_cancel) {
	XtAppNextEvent(a_context, &e);
	XtDispatchEvent(&e);
    }

    if (get_string_ok) {
	char *value;
	
	XtVaGetValues(string, XtNstring, &value, NULL);
	user_string = XtNewString(value);
    }
    else
	user_string = NULL;
    
    XtPopdown(get_string_pshell);
    XtDestroyWidget(get_string_pshell);
    
    return user_string;
}

/*
  A question dialog. When called, the string TEXT is displayed and the the
  user can choose to hit either the Yes or No button. If Yes is chosen, then
  question_dialog returns TRUE, otherwise it returns FALSE.
*/

int
question_dialog(text)
     char *text;
{
    XEvent e;
    Cursor dot;
    Widget question_dialog_pshell;
    Widget yes, no, question_dialog_widget;
    Dimension yes_x, no_x, dialog_x;
    int question_dialog_yes, question_dialog_no;

    question_dialog_pshell = XtVaCreatePopupShell
	("Question Dialog", transientShellWidgetClass, toplevel, NULL);
    
    question_dialog_widget = XtVaCreateManagedWidget
	("questDialog", dialogWidgetClass, question_dialog_pshell,
	 XtNresizable, TRUE, NULL); 
    
    yes = XtVaCreateManagedWidget
	("Yes", commandWidgetClass, question_dialog_widget, NULL);
    XtAddCallback(yes, XtNcallback, dialog_cb, &question_dialog_yes); 
    
    no = XtVaCreateManagedWidget
	("No", commandWidgetClass, question_dialog_widget, NULL);
    XtAddCallback(no, XtNcallback, dialog_cb, &question_dialog_no); 
    
    /* The Unrealize, set values, realize sequence ensures that the error
       dialog will be the proper size given TEXT. Without this, the dialog
       would not resize for each new string */
     
    XtUnrealizeWidget(question_dialog_pshell);
    XtVaSetValues(question_dialog_widget, XtNlabel, text, NULL);
    XtRealizeWidget(question_dialog_pshell);

    XtVaGetValues(yes, XtNwidth, &yes_x, NULL);
    XtVaGetValues(no, XtNwidth, &no_x, NULL);    
    XtVaGetValues(question_dialog_widget, XtNwidth, &dialog_x, NULL);

    XtUnrealizeWidget(question_dialog_pshell);
    XtVaSetValues(yes, XtNhorizDistance,
		  (long)((dialog_x - yes_x - no_x - 4) / 2), NULL);
    XtRealizeWidget(question_dialog_pshell);
    
    position_dialog(question_dialog_pshell, (Position)(dialog_x / 2) +
		    X_OFFSET, (Position)Y_OFFSET); 

    
    XtPopup(question_dialog_pshell, XtGrabExclusive);

    dot = XCreateFontCursor(display, XC_dot);
    XDefineCursor(display, XtWindow(question_dialog_pshell), dot);
    XFreeCursor(display, dot);

    XtInstallAccelerators(yes, yes);
    XtInstallAccelerators(no, no);
    
    /* return when one of these is TRUE */
    question_dialog_yes = FALSE; question_dialog_no = FALSE;
    
    while (!question_dialog_yes && !question_dialog_no) {
	XtAppNextEvent(a_context, &e);
	XtDispatchEvent(&e);
    }

    XtPopdown(question_dialog_pshell);
    XtDestroyWidget(question_dialog_pshell);
    
    return question_dialog_yes;
}

/*
  An error dialog. This dialog displays error text and provides an single
  button labeld 'OK'. The user can dismiss the dialog by hitting the  'OK'
  button (or return).
*/

#if defined(__STDC__)
void
err_dialog(const char *fmt, ...)
{
#else
void
err_dialog(va_alist)
va_dcl
{
    char *fmt;
#endif /* __STDC__ */    
    va_list args;
    char text[256];

#if defined(__STDC__)
    va_start(args, fmt);
#else
    va_start(args);
#endif /* __STDC__ */

#if !defined( __STDC__)
    fmt = va_arg(args, char *);
#endif
    vsprintf(text, fmt, args);
    va_end(args);
    
    popup_err_dialog(text, "Ok");
}

/*
  sys_err_dialog is just like err_dialog, with one exception; it takes the
  given error text and appends a \n to it. It then appends the string from
  sys_errlist[errno] to it.
*/

#if defined(__STDC__)
void
sys_err_dialog(const char *fmt, ...)
{
#else
void
sys_err_dialog(va_alist)
va_dcl
{
    char *fmt;
#endif /* __STDC__ */    
    va_list args;
    char text[384];		/* why not */

#if defined(__STDC__)
    va_start(args, fmt);
#else
    va_start(args);
#endif /* __STDC__ */

#if !defined( __STDC__)
    fmt = va_arg(args, char *);
#endif
    vsprintf(text, fmt, args);
    va_end(args);
    
    if (errno) {
	strcat(text, ":\n");
	strcat(text, sys_errlist[errno]);
    }
    
    popup_err_dialog(text, "Ok");
}

/*
  Like sys_err_dialog, except that in addition to displaying the text message
  that sys_err_dialog displays, sys_quit_dialog calls either exit(1) or
  abort() after the user hits the Ok button. If NDEBUG is defined, then exit
  will be called, otherwise, abort will be called.
*/

#if defined(__STDC__)
void
sys_quit_dialog(const char *fmt, ...)
{
#else
void
sys_quit_dialog(va_alist)
va_dcl
{
    char *fmt;
#endif /* __STDC__ */    
    va_list args;
    char text[384];		/* why not */

#if defined(__STDC__)
    va_start(args, fmt);
#else
    va_start(args);
#endif /* __STDC__ */
    
#if !defined( __STDC__)
    fmt = va_arg(args, char *);
#endif
    vsprintf(text, fmt, args);
    va_end(args);
    
    if (errno) {
	strcat(text, ":\n");
	strcat(text, sys_errlist[errno]);
    }
    
#if defined (NDEBUG)
    popup_err_dialog(text, "Exit");
    exit(1);
#else
    popup_err_dialog(text, "Abort");
    abort();
#endif /* NDEBUG */    
}

static void 
popup_err_dialog(text, button_name)
     char *text;
     char *button_name;
{
    XEvent e;
    Cursor dot;
    Widget ok;
    Widget err_dialog_pshell;
    Widget err_dialog_widget;
    Dimension ok_x, dialog_x;
    int err_dialog_ok;
    
    err_dialog_pshell = XtVaCreatePopupShell
	("Error Dialog", transientShellWidgetClass, toplevel, NULL);
    
    err_dialog_widget = XtVaCreateManagedWidget
	("errDialog", dialogWidgetClass, err_dialog_pshell, NULL); 
    
    ok = XtVaCreateManagedWidget
	(button_name, commandWidgetClass, err_dialog_widget, NULL);
    
    XtAddCallback(ok, XtNcallback, dialog_cb, &err_dialog_ok);
    
    /* The Unrealize, set values, realize sequence ensures that the error
       dialog will be the proper size given TEXT. Without this, the dialog
       would not resize for each new string */
     
    XtUnrealizeWidget(err_dialog_pshell);
    XtVaSetValues(err_dialog_widget, XtNlabel, text, NULL);
    XtRealizeWidget(err_dialog_pshell);

    XtVaGetValues(ok, XtNwidth, &ok_x, NULL);
    XtVaGetValues(err_dialog_widget, XtNwidth, &dialog_x, NULL);

    XtUnrealizeWidget(err_dialog_pshell);
    XtVaSetValues(ok, XtNhorizDistance, (long)(dialog_x - ok_x) / 2, NULL);
    XtRealizeWidget(err_dialog_pshell);
    
    position_dialog(err_dialog_pshell, (Position)(dialog_x / 2) + X_OFFSET,
		    (Position)Y_OFFSET);

    XtPopup(err_dialog_pshell, XtGrabExclusive);

    XtInstallAccelerators(err_dialog_widget, ok);
    XtInstallAccelerators(ok, ok);
    
    dot = XCreateFontCursor(display, XC_dot);
    XDefineCursor(display, XtWindow(err_dialog_pshell), dot);
    XFreeCursor(display, dot);
    
    err_dialog_ok = FALSE;	/* test this for completion (== TRUE) */

    while (!err_dialog_ok) {
	XtAppNextEvent(a_context, &e);
	XtDispatchEvent(&e);
    }

    XtPopdown(err_dialog_pshell);
    XtDestroyWidget(err_dialog_pshell);
}

/*
  Popup the FWF SelFile dialog prompting for a file to read. If the user
  clicks Ok get_filename() returns a pointer to the string (allocated using
  XtMalloc) otherwise it returns NULL.

  I made the cfs_pshell widget static so that once created it live for
  the lifetime of the program. This way it never needs to be
  destroyed. On the alpha, XDestroyWidget was causing a core dump;
  this hack effectively fixes the bug. 

  This bug showes up only in the dialog boxes that are used to prompt
  for file names (and which come from the FileComp widget from te
  FWF). The other dialogs don't seem to have this problem (of crashing
  the program). jhrg 11/25/94

  See also the comments for new_filename().
*/

char *
get_filename(prompt, init_path)
     char *prompt;
     char *init_path;
{
    int cfs_dialog_ok, cfs_dialog_cancel;
    char path[MAXPATHLEN];
    Dimension dialog_x;
    XEvent e;
    Widget form;
    static Widget cfs_pshell, title, cfs;

    /* create the widgets on the first run...*/

    if (!cfs_pshell) {
	cfs_pshell = XtVaCreatePopupShell
	    ("openDialog", transientShellWidgetClass, toplevel, NULL);

	form = XtVaCreateManagedWidget
	    ("form", formWidgetClass, cfs_pshell, NULL);

	title = XtVaCreateManagedWidget
	    ("title", labelWidgetClass, form, NULL);

	cfs = XtVaCreateManagedWidget
	    ("cfs", xfwffileCompleteWidgetClass, form, NULL);

	XtAddCallback(cfs, XtNcallback, dialog_cb, &cfs_dialog_ok);
	XtAddCallback(cfs, XtNcancelCallback, dialog_cb, &cfs_dialog_cancel);

	XtRealizeWidget(cfs_pshell);
	XtVaGetValues(form, XtNwidth, &dialog_x, NULL);

	position_dialog(cfs_pshell, (Position)(dialog_x / 2) + X_OFFSET,
			(Position)Y_OFFSET);
    }

    /* set or reset the values of the widgets on subsequent runs... */

    /* check that the initial path is valid. If it is not, then a subsequent
       user action to complete the file path may cause a core dump. If
       `init_path' is not valid, set it to the CWD. */

    if (access(init_path, F_OK) == -1)
	getcwd(path, MAXPATHLEN);
    else
	strcpy(path, init_path ? init_path : "");

    XtVaSetValues(title, XtNlabel, prompt, NULL);
    XtVaSetValues(cfs, XtNfilename, path, NULL);

    XtPopup(cfs_pshell, XtGrabExclusive);

    cfs_dialog_ok = cfs_dialog_cancel = FALSE;

    while (!(cfs_dialog_ok || cfs_dialog_cancel)) {
	XtAppNextEvent(a_context, &e);
	XtDispatchEvent(&e);
    }

    if (cfs_dialog_ok) {
	char *file;

	XtVaGetValues(cfs, XtNfilename, &file, NULL); 

	XtPopdown(cfs_pshell);
	return XtNewString(file);
    }
    else {
	XtPopdown(cfs_pshell);
	return NULL;
    }
}

/*
  Get a new filename. Used to prompt the user for the name of a file to save
  data to. Since the contents of an existing file will be over written (by
  UNIX) this dialog will prompt users (asking `Do you really want to replace
  ...') if they supply the name of an existing file. 

  If the user clicks `Cancel' the function returns NULL. Otherwise the
  function returns a string allocated with XtMalloc().  
*/

char *
new_filename(prompt, init_path)
     char *prompt;
     char *init_path;
{
    static int cfs_save_ok, cfs_save_cancel;
    char path[MAXPATHLEN];
    Dimension dialog_x;
    XEvent e;
    Widget form;
    static Widget cfs_pshell, title, cfs;

    /* If running new_filename() for the first time, build the widgets. Some */
    /* are static variables so that they can be set/reset on subsequent */
    /* entries into this function. The dialog boxes that use the FWF file */
    /* completion widgets had to be implemented this way to avoid a bug when */
    /* instances of those widgets were destroyed. */

    if (!cfs_pshell) {
	cfs_pshell = XtVaCreatePopupShell
	    ("saveDialog", transientShellWidgetClass, toplevel, NULL);

	form = XtVaCreateManagedWidget
	    ("form", formWidgetClass, cfs_pshell, NULL);

	title = XtVaCreateManagedWidget
	    ("title", labelWidgetClass, form, NULL);
    
	cfs = XtVaCreateManagedWidget
	    ("cfs", xfwffileCompleteWidgetClass, form, NULL);

	XtAddCallback(cfs, XtNcallback, dialog_cb, (XtPointer)&cfs_save_ok);
	XtAddCallback(cfs, XtNcancelCallback, dialog_cb,
		      (XtPointer)&cfs_save_cancel); 

	XtRealizeWidget(cfs_pshell);
	XtVaGetValues(form, XtNwidth, &dialog_x, NULL);

	position_dialog(cfs_pshell, (Position)(dialog_x / 2) + X_OFFSET,
			(Position)Y_OFFSET);
    }

    /* Once the widgets are built, set their values. After the first time */
    /* this function is run, the title and cfs widget are accessed through */
    /* static variables. */

    /* check that the initial path is valid. If it is not, then a
       subsequent user action to complete the file path may cause a core
       dump. If `init_path' is not valid, set it to the CWD. */

    /* if the INIT_PATH does not check out *and* does not start with '/',
       then assume it is a filename - tack that onto the end of the cwd
       copied into PATH. */

    if (access(init_path, F_OK) == -1) {
	getcwd(path, MAXPATHLEN);
	if (init_path[0] != '/' 
	    && (strlen(path) + strlen(init_path) + 2) < MAXPATHLEN) {
	    strcat(path, "/");
	    strcat(path, init_path);
	}
    }
    else
	strcpy(path, init_path ? init_path : "");

    XtVaSetValues(title, XtNlabel, prompt, NULL);
    XtVaSetValues(cfs, XtNfilename, path, NULL);


    XtPopup(cfs_pshell, XtGrabExclusive);

    cfs_save_ok = cfs_save_cancel = FALSE;

    while (!(cfs_save_ok || cfs_save_cancel)) {
	XtAppNextEvent(a_context, &e);
	XtDispatchEvent(&e);
    }

    if (cfs_save_ok) {
	char *file;

	XtVaGetValues(cfs, XtNfilename, &file, NULL);

	/* check if `file' already exists. If so, offer the user a chance to
	   give another filename. */
	
	if (fopen(file, "r")) {	/* non-NULL if the file exists */
	    char text[MAXPATHLEN];

	    sprintf(text, "Do you want to erase `%s'", file);
	    
	    if (!question_dialog(text)) {
		XtPopdown(cfs_pshell);
		XtDestroyWidget(cfs_pshell);
		return new_filename(prompt, init_path);
	    }
	}
	    
	XtPopdown(cfs_pshell);
	return XtNewString(file);
    }
    else {
	XtPopdown(cfs_pshell);
	return NULL;
    }
}

/*
  Move the dialog widget `w' so that when it XtPopup is called, then dialog
  appears under the pointer. Ideally, the pointer should be within the value
  field of the dialog so that the user can enter text w/o leaveing the
  keyboard. (x_offset, y_offset) is the offset from the top left of the
  dialog to the point in the dialog where the pointer should be.
*/

static void 
position_dialog(w, x_offset, y_offset)
     Widget w;
     long int x_offset;
     long int y_offset;
{
    Window root, child;
    int dummy;
    int x, y;
    int x_root, y_root;
    Dimension dialog_h, dialog_w;
    unsigned int mask;

    XQueryPointer(display, XtWindow(w), &root, &child, &x_root, &y_root,
		  &dummy, &dummy, &mask);

    XtVaGetValues(w, XtNwidth, &dialog_w, XtNheight, &dialog_h, NULL);

    x = x_root - x_offset;
    y = y_root - dialog_h + y_offset;

    x = (x < 0) ? 0 : x;
    x = (x + dialog_w > display_w) ? display_w - dialog_w : x;

    y = (y < 0) ? 0 : y;
    y = (y + dialog_h > display_h) ? display_h - dialog_h : y;

    XtMoveWidget(w, x, y);
}

/*
  This callback is used by all the dialog boxes to set flag variables when
  one of their buttons is clicked. The variable pointed to be client_data is
  specified in the the XtAddCallbacks call associated with each button widget.
  Each time this function is registered as a callback, a different flag can
  be passed in allowing this one function to set different flags depending on
  which button the user clicked.
*/

static void 
dialog_cb(w, client_data, call_data)
     Widget w;
     XtPointer client_data;
     XtPointer call_data;
{
    *(int *)client_data = TRUE;
}


static void
criteria_ok_dialog_cb(w, client_data, call_data)
     Widget w;
     XtPointer client_data;
     XtPointer call_data;
{
    char *value;
    struct criteria_bundle *c_b;
    
    c_b = (struct criteria_bundle *)client_data;

    XtVaGetValues(c_b->start_text, XtNstring, &value, NULL);
    if (! *value) {
	if (c_b->bpe) {
	    err_dialog("You must enter both a starting and an ending date");
	    c_b->bpe = FALSE;
	}
	XtSetKeyboardFocus(c_b->dialog_form, c_b->start_text);
	XawTextDisplayCaret(c_b->start_text, TRUE);
	XawTextDisplayCaret(c_b->end_text, FALSE);
	return;
    }

    XtVaGetValues(c_b->end_text, XtNstring, &value, NULL);
    if (! *value) {
	if (c_b->bpe) {
	    err_dialog("You must enter both a starting and an ending date");
	    c_b->bpe = FALSE;
	}
	XtSetKeyboardFocus(c_b->dialog_form, c_b->end_text);
	XawTextDisplayCaret(c_b->start_text, FALSE);
	XawTextDisplayCaret(c_b->end_text, TRUE);
	return;
    }
    
   c_b->done = TRUE;
}

static void
login_dialog_cb(w, client_data, call_data)
     Widget w;
     XtPointer client_data;
     XtPointer call_data;
{
    char *value, *pvalue;
    struct login_bundle *l_b;
    
    l_b = (struct login_bundle *)client_data;

    XtVaGetValues(l_b->username_text, XtNstring, &value, NULL);
    if (! *value) {
	if (l_b->bpe) {
	    err_dialog("You must enter a username and a password");
	    l_b->bpe = FALSE;
	}
	XtSetKeyboardFocus(l_b->dialog_form, l_b->username_text);
	XawTextDisplayCaret(l_b->username_text, TRUE);
	XawTextDisplayCaret(l_b->passwd_text, FALSE);
	return;
    }

    XtVaGetValues(l_b->passwd_text, XtNstring, &pvalue, NULL);
    /* if passwd is null and username is not anonymous of browse ... */
    if (! *pvalue && ! (strcmp(value, "anonymous") == 0
			|| strcmp(value, "browse") == 0)) {
	if (l_b->bpe) {
	    err_dialog("You must enter a username and a password");
	    l_b->bpe = FALSE;
	}
	XtSetKeyboardFocus(l_b->dialog_form, l_b->passwd_text);
	XawTextDisplayCaret(l_b->username_text, FALSE);
	XawTextDisplayCaret(l_b->passwd_text, TRUE);
	return;
    }
    
   l_b->done = TRUE;
}

    
