/*
  Copyright 1992 University of Rhode Island

  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose is hereby granted without fee, provided that
  the above copyright notice appear in all copies and that both that
  copyright notice and this permission notice appear in supporting
  documentation, and that the name of U.R.I. not be used in advertising or
  publicity pertaining to distribution of the software without specific,
  written prior permission.  U.R.I. makes no representations about the
  suitability of this software for any purpose.  It is provided "as is"
  without express or implied warranty.

  U.R.I. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.R.I.
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS ACTION, ARISING OUT OF OR IN
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
  Error reporting functions. These functions make up a simple error reporting
  mechanism. Error messages are printed on stderr, syslog (if available) or a
  log file. System error messages may be printed as well as user supplied
  text. The message is specified as is a message printed using printf.
  Additionally, certain functions can be configured to exit or dump core
  depending on the values passed to an initialization function. The
  initialization function is used to set three state variables: 1) `pname', a
  string containing the program name, 2) `report_mode', an int specifying
  where to report errors (one of STDERR, SYSLOG or FILELOG), and 
  3) `exit_mode', an int that specifies whether the program is to exit
  (FATAL_EXIT) on fatal errors or dump core (FATAL_ABORT). The initialization
  function can be called at any time during the program's execution. If the
  function is not called, the library uses default values for the three
  state variables (pname == "program", report_mode == STDERR, exit_mode ==
  FATAL_EXIT).

  This file can be compiled under UNIX and VMS. If it is compiled under UNIX,
  the compile-time constant `unix' must be defined. If compiled under VMS,
  `vms' must be defined.

  The compile-time constant HAS_T_ERROR can be used to enable the system V
  `t_errno' error codes. These will be printed along with the regular error
  messages by those functions that print system error message text.

  This library was inspired by a library in Stevens' ``UNIX Network
  Programming'', Prentice Hall, 1990.

  jhrg
*/

/*
 * $Log: errmsg.c,v $
 * Revision 1.10  1992/11/10  17:45:46  jimg
 * Fixed prototype bugs.
 * Fixed compiler symbols (unix, ...) so that they work with gcc -ansi
 *
 * Revision 1.9  1992/10/13  19:40:17  jimg
 * Fixed bug where (under vms) err_init() returned -1. This happened when
 * FILELOG mode was specified and the log file was already open. Under vms
 * each call to err_init() creates a new version of the log file.
 *
 * Revision 1.8  1992/09/25  20:06:32  jimg
 * Fixed unix version so that file names in the log file do not contain
 * the entire pathname of the executable. This make the log files more
 * readable.
 *
 * Revision 1.7  1992/09/24  15:55:04  jimg
 * Added copyright notice.
 * corrected comments, vms compile-time problems.
 *
 * Revision 1.6  1992/09/23  20:16:10  jimg
 * Fixed VMS filename/pahtname mess. Now, under VMS only the program name
 * appears in the logfile - not the complete pahtname.
 *
 * Revision 1.5  1992/05/11  19:08:11  jimg
 * Merged in changes from vms port.
 *
 * Revision 1.3  1992/04/14  19:42:07  jimg
 * Noted bug when used under UNIX and FILELOG mode in a daemon.
 *
 * Revision 1.2  1992/04/14  18:14:11  jimg
 * Changed FILELOG mode so that it no longer uses buffered I/O. This
 * should aid in debugging daemons.
 *
 * Revision 1.1  1992/03/23  19:24:43  jimg
 * Initial revision
 *
 */
    
static char rcsid[]="$Id: errmsg.c,v 1.10 1992/11/10 17:45:46 jimg Exp jimg $";

#if !(defined(unix) || defined(__unix)) && !defined(vms)
error "Either vms or unix must be defined"
#elif (defined(unix) || defined(__unix)) && defined(vms)
error "Only one of vms or unix can be defined"
#elif defined(vms) && defined(HAS_T_ERROR)
error "Only one of vms or HAS_T_ERROR can be defined"
#endif /* unix : vms : HAS_T_ERROR */

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <unistd.h>

#ifdef __STDC__
#include <stdarg.h>
#else    
#include <varargs.h>
#endif /* __STDC__ */

#if defined(unix) || defined(__unix)
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <syslog.h>
#elif defined(vms)
#include <stddef.h>
#include <stdlib.h>
#include <errno.h>
#include <stat.h>
#include <unixio.h>
#include <file.h>
#endif /* unix : vms */

#ifndef NULL
#define NULL ((void *)0)
#endif

#ifndef EXIT_FAILURE
#define EXIT_FAILURE 1
#endif

#define TRUE (1)
#define FALSE (0)

#include "errmsg.h"

static char pname[256] = {""};	/* program name */
static int report_mode = STDERR; /* either STDERR, SYSLOG or FILELOG */
static int exit_mode = FATAL_EXIT; /* either FATAL_EXIT or FATAL_ABORT */
static char emesgstr[256]={0};	/* error message storage for syslog/file */
static int err_file;		/* file pointer for FILELOG report mode */

#if defined(unix) || defined(__unix)
extern int errno;		/* Unix error number */
extern int sys_nerr;		/* # of strings in err msg table */
extern char *sys_errlist[];	/* system err msg table */
#ifdef HAS_T_ERROR
extern int t_errno;
extern int t_nerr;
extern char *t_errlist[1];
#endif /* HAS_T_ERROR */
#endif /* unix */

#ifdef __STDC__
static char *sys_err_str(void);	/* necessary until strerror is everywhere */
static char *get_preamble(void);
#else
static char *sys_err_str();
static char *get_preamble();
#endif

/*
  Initialize the error reporting mechanism. If this function is not called at
  the begining of a program, then the error library assumes the default
  values for report mode (stderr) and exit mode (fatal exit) and the string
  pname is empty. Otherwise, the user supplies values for these parameters by
  calling this function. The values for pname, report_mode and exit_mode may
  be changed at any time.

  NB: Since VMS does not have syslog, calling err_init with report_mode =
  SYSLOG *when vms is defined* is the same as calling it with report_mode =
  FILELOG.
*/

#ifdef __STDC__
int
err_init(const char *ident, int r_mode, int e_mode)
#else
int
err_init(ident, r_mode, e_mode)
char *ident;
int r_mode;
int e_mode;
#endif
{
    strcpy(pname, ident);
    exit_mode = e_mode;

    switch (r_mode) {
    case STDERR: {
	/*
	  If a problem is detected with stderr, then retrun FALSE. If the
	  calling process is a daemon, then fstat(stderr,...) will return
	  (-1). If the calling process was allowed to write to stderr, the
	  daemon would mysteriously die after a while.
	*/
	struct stat buf;

	if (fstat(2, &buf) == -1)
	    return FALSE;
	report_mode = r_mode;
	break;
    }

    case SYSLOG:
	/*
	  If openlog() fails and we can detect it, then try FILELOG mode.
	  Note:
	  1) if the compile-time constant `vms' is defined, then SYSLOG is
	  the same as FILELOG. 
	  2) openlog() is different on many systems; rather than test for
	  each system, throwaway the return value if one exists. Hardly
	  optimal. 
	 */
#if defined(ultrix) || defined(__ultrix)
	if (openlog(ident, LOG_PID) == 0) {
	    report_mode = r_mode;
	    break;
	}
#elif !defined(vms)
	(void) openlog(ident, (LOG_PID | LOG_CONS), LOG_DAEMON);
	report_mode = r_mode;
	break;
#endif /* ultrix : unix */

    case FILELOG: {
	char filename[256];

	if (strcmp(pname, "") == 0)
	    strcpy(pname, "program");

	sprintf(filename, "%s.err", pname);
#if !defined(vms)
	err_file = open(filename, O_WRONLY|O_APPEND|O_CREAT, 0644);
#else /* vms */
	err_file = creat(filename, 0);
#endif /* unix : vms */

	if (err_file == -1)
	    return FALSE;
	report_mode = r_mode;
	break;
    }
    }

    /* under VMS, `ident' must be the full pathname since many programs
       (e.g., network servers) have the CWD as the root of some disk.
       However, this makes for messy logfiles, so I strip the pathname here
       after the file is created. jhrg 9/23/92 */
#ifdef vms
    if (strrchr(pname, ']'))
	strcpy(pname, strrchr(pname, ']') + 1);
#else
    if (strrchr(pname, '/'))
	strcpy(pname, strrchr(pname, '/') + 1);
#endif    
    
    return TRUE;
}

/*
  Print the user supplied error message and return. The error message is
  composed of a format string and variable number of arguments as per printf. 
*/

#ifdef __STDC__
void
err_print(const char *fmt, ...)
{
    va_list args;
    
    va_start(args, fmt);
#else
void
err_print(va_alist)
va_dcl
{
    va_list args;
    char *fmt;
    
    va_start(args);
    fmt = va_arg(args, char *);
#endif /* __STDC__ */

    switch (report_mode) {
    case STDERR:
	if (strcmp(pname, "") != 0)
	    fprintf(stderr, "%s: ", pname);
	vfprintf(stderr, fmt, args);
	fputc('\n', stderr);
	fflush(stderr);
	break;

    case SYSLOG:
#ifndef vms
	vsprintf(emesgstr, fmt, args);
	syslog(LOG_ERR, emesgstr);
	break;
#endif /* ! vms */
	
    case FILELOG: {
	char *preamble = get_preamble();
	write(err_file, preamble, strlen(preamble));

	vsprintf(emesgstr, fmt, args);
	strcat(emesgstr, "\n");
	write(err_file, emesgstr, strlen(emesgstr));
	break;
    }
    }
    
    va_end(args);
    
    return;
}

/*
  Print the user supplied error message and exit (or abort if exit_mode ==
  FATAL_ABORT). The error message is composed of a format string and
  variable number of arguments as per printf.
*/

#ifdef __STDC__
void
err_quit(const char *fmt, ...)
{
    va_list args;
    
    va_start(args, fmt);
#else
void
err_quit(va_alist)
va_dcl
{
    va_list args;
    char *fmt;
    
    va_start(args);
    fmt = va_arg(args, char *);
#endif /* __STDC__ */

    switch (report_mode) {
    case STDERR:
	if (strcmp(pname, "") != 0)
	    fprintf(stderr, "%s: ", pname);
	vfprintf(stderr, fmt, args);
	fputc('\n', stderr);
	fflush(stderr);
	break;

    case SYSLOG:
#ifndef vms
	vsprintf(emesgstr, fmt, args);
	syslog(LOG_ERR, emesgstr);
	closelog();		/* this function either exits or aborts */
	break;
#endif /* ! vms */
	
    case FILELOG: {
	char *preamble = get_preamble();
	write(err_file, preamble, strlen(preamble));

	vsprintf(emesgstr, fmt, args);
	strcat(emesgstr, "\n");
	write(err_file, emesgstr, strlen(emesgstr));
	break;
    }
    }
    
    va_end(args);

    if (exit_mode == FATAL_ABORT)
	abort();
    else
	exit(EXIT_FAILURE);
}

/*
  Print the user supplied error message, append the system error text and
  return. The error message is composed of a format string and variable
  number of arguments as per printf.
*/

#ifdef __STDC__
void
err_ret(const char *fmt, ...)
{
    va_list args;
    
    va_start(args, fmt);
#else
void
err_ret(va_alist)
va_dcl
{
    va_list args;
    char *fmt;
    
    va_start(args);
    fmt = va_arg(args, char *);
#endif /* __STDC__ */

    switch (report_mode) {
    case STDERR:
	if (strcmp(pname, "") != 0)
	    fprintf(stderr, "%s: ", pname);
	vfprintf(stderr, fmt, args);
	fprintf(stderr, " (%s)\n", sys_err_str());
	fflush(stderr);
	break;

    case SYSLOG:
#ifndef vms
	vsprintf(emesgstr, fmt, args);
	sprintf(emesgstr + strlen(emesgstr), " (%s)", sys_err_str());
	syslog(LOG_ERR, emesgstr);
	break;
#endif /* ! vms */
	
    case FILELOG: {
	char *preamble = get_preamble();
	write(err_file, preamble, strlen(preamble));

	vsprintf(emesgstr, fmt, args);
	sprintf(emesgstr + strlen(emesgstr), " (%s)\n", sys_err_str());
	write(err_file, emesgstr, strlen(emesgstr));
	break;
    }
    }
    
    va_end(args);

    return;
}

/*
  Print the user supplied error message, append the system error text and exit
  (or abort if exit_mode == FATAL_ABORT). The error message is composed of a
  format string and variable number of arguments as per printf.
*/

#ifdef __STDC__
void
err_sys(const char *fmt, ...)
{
    va_list args;
    
    va_start(args, fmt);
#else
void
err_sys(va_alist)
va_dcl
{
    va_list args;
    char *fmt;
    
    va_start(args);
    fmt = va_arg(args, char *);
#endif /* __STDC__ */

    switch (report_mode) {
    case STDERR:
	if (strcmp(pname, "") != 0)
	    fprintf(stderr, "%s: ", pname);
	vfprintf(stderr, fmt, args);
	fprintf(stderr, " (%s)\n", sys_err_str());
	fflush(stderr);
	break;

    case SYSLOG:
#ifndef vms
	vsprintf(emesgstr, fmt, args);
	sprintf(emesgstr + strlen(emesgstr), " (%s)", sys_err_str());
	syslog(LOG_ERR, emesgstr);
	closelog();		/* since this function exits or aborts */
	break;
#endif /* ! vms */
	
    case FILELOG: {
	char *preamble = get_preamble();
	write(err_file, preamble, strlen(preamble));

	vsprintf(emesgstr, fmt, args);
	sprintf(emesgstr + strlen(emesgstr), " (%s)\n", sys_err_str());
	write(err_file, emesgstr, strlen(emesgstr));
	break;
    }
    }
    
    va_end(args);

    if (exit_mode == FATAL_ABORT) 
	abort();
    else
	exit(EXIT_FAILURE);
}
    
/*
  Return a string containing some additional OS dependent information.
  Note that different versions of unix assign different meanings to the same
  value of errno.
*/

static char *
sys_err_str()
{
    static char msgstr[256];
    
    if (errno != 0) {
#ifdef vms
	strcpy(msgstr, strerror(errno));
#else
	if (errno > 0 && errno < sys_nerr)
	    sprintf(msgstr, "%s", sys_errlist[errno]);
	else
	    sprintf(msgstr, "errno = %d", errno);
#endif
    }
    else {
	msgstr[0] = '\0';
    }

#ifdef HAS_T_ERROR
    if (t_errno != 0) {
	char tmsgstr[100];
	
	if (t_errno > 0 && t_errno < sys_nerr)
	    sprintf(tmsgstr, ", %s", t_errlist[t_errno]);
	else
	    sprintf(tmsgstr, ", t_errno = %d", t_errno);
	strcat(msgstr, tmsgstr);
    }
#endif
    
    return(msgstr);
}

static char *
get_preamble()
{
    static char preamble[256];
    char time_str[64];	/* it could be 64 chars ... :-) */
    int time_str_len;
    time_t now = time(NULL);
    
    strcpy(time_str, ctime(&now));
    time_str_len = strlen(time_str) - 1; /* strip `\n' from end */
    strncpy(preamble, time_str, time_str_len);
    sprintf(preamble + time_str_len , " %s[%d]: ", pname, getpid());

    return preamble;
}
