/*
  Copyright 1992 University of Rhode Island
 
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose is hereby granted without fee, provided that
  the above copyright notice appear in all copies and that both that
  copyright notice and this permission notice appear in supporting
  documentation, and that the name of U.R.I. not be used in advertising or
  publicity pertaining to distribution of the software without specific,
  written prior permission.  U.R.I. makes no representations about the
  suitability of this software for any purpose.  It is provided "as is"
  without express or implied warranty.
 
  U.R.I. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.R.I.
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
 * $Log: grid.c,v $
 * Revision 1.2  1993/02/02  16:12:32  jimg
 * Added conditional include for debugging malloc header.
 *
 * Revision 1.1  1992/11/19  21:14:30  jimg
 * Initial revision
 */

static char rcsid[]="$Id: grid.c,v 1.2 1993/02/02 16:12:32 jimg Exp jimg $";

#include <stdio.h>

#include <sys/time.h>
#include <rpc/rpc.h>

#include <X11/Intrinsic.h>	/* toolkit */

#ifdef MALLOC_DEBUG
#include <malloc.h>
#endif 

#include "xbrowse.h"
#include "errmsg.h"
#include "comm.h"
#include "vertex.h"
#include "region.h"
#include "ximage.h"
#include "pyramid.h"
#include "pixmap-mark.h"
#include "xbt-const.h"
#include "xbt-types.h"
#include "xbt.h"
#include "dialog.h"
#include "browse.h"

#define FUDGE 10		/* extra space around labels */

/* Static Globals */

static Pixmap grid;		/* a bitmap for the grid */
static GC grid_gc;
static double grid_tllat, grid_tllon, grid_brlon, grid_brlat;
static XFontStruct *grid_font;
static unsigned char grid_color;
static int grid_line_width;

/* Extern Globals */

extern APPDATA a_data;
extern Display *display;
extern Widget the_image;
extern Pixmap window_pixmap;	/* holds image offscreen - server side */
extern Pixmap clip_mask;	/* clip mask so that hash marks show */
extern GC image_gc;		/* used to transfer window_pixmap to screen */
extern GC clip_gc;		/* used to update clip mask */
extern Window root_window;
extern int window_pixmap_x, window_pixmap_y;
extern Widget the_image;

/*
  Draw and label latitude and longitude lines on a bitmap. Only works for a
  linear projection.
*/

static void
draw_grid_lines(pyramid, tllat, tllon, brlat, brlon, x_size, y_size, gs, pm,
		gc, font)
PYRAMID *pyramid;
double tllat, tllon, brlat, brlon;
int x_size, y_size;		/* used in MAP_* macros above */
int gs;				/* grid spacing */
Pixmap pm;
GC gc;
XFontStruct *font;
{
    long i_lat, i_lon, width, i, j, len, wrap_lon;
    int ascent = font->ascent, descent = font->descent;
    char num[5];

    /* first draw the center lat and lon */

    /* draw lines for +/- lat/lon separately */
    
    for (i = (int)tllat / gs * gs; i > (int)brlat; i -= gs) {
	get_x_y(pyramid, 0.0, (double)i, &i_lon, &i_lat);
	if (i_lat - (ascent + descent) - FUDGE < 0
	    || i_lat + (ascent + descent) + FUDGE > y_size)
	    continue;		/* don't draw lines too near edges */
	XDrawLine(display, pm, gc, 0, i_lat, x_size, i_lat);
	if (i > 0) {
	    sprintf(num, "%d", i);
	    strcat(num, "\260N"); len = strlen(num);
	    width = XTextWidth(font, num, len);
	    XDrawString(display, pm, gc, 2, i_lat - descent - 1, num, len); 
	    XDrawString(display, pm, gc, x_size - width - 2,
			i_lat - descent - 1,num, len);
	}
	else {
	    sprintf(num, "%d", -i);
	    strcat(num, "\260S"); len = strlen(num);
	    width = XTextWidth(font, num, len);
	    XDrawString(display, pm, gc, 2, i_lat + ascent + 1, num, len); 
	    XDrawString(display, pm, gc, x_size - width - 2, i_lat + ascent +
			1, num, len);
	}
    }

    tllon += 180.0;		/* transform lon endpoints */
    brlon += 180.0;

    /* wrap_lon lets one loop cover a range of values that cross the
       international date line (340 --> 20 degrees longitude). */
    if (tllon > brlon)		
	wrap_lon = 360;
    else
	wrap_lon = 0;
    
    for (i = (int)tllon / gs * gs; i < (int)brlon + wrap_lon; i += gs) {
	if (i > brlon)		/* crossed over date line */
	    j = i - 180 - wrap_lon;
	else
	    j = i - 180;
	get_x_y(pyramid, (double)j, 0.0, &i_lon, &i_lat);
	if (i_lon - (ascent + descent) - FUDGE < 0 /* should use width */
	    || i_lon + (ascent + descent) + FUDGE > x_size)
	    continue;		/* don't draw lines too near edges */
	XDrawLine(display, pm, gc, i_lon, 0, i_lon, y_size);
	if (j  < 0) {
	    sprintf(num, "%d", -j);
	    strcat(num, "\260W"); len = strlen(num);
	    width = XTextWidth(font, num, len);
	    XDrawString(display, pm, gc, i_lon - width - 2, 1 + ascent,
			num, len);  
	    XDrawString(display, pm, gc, i_lon - width - 2, y_size -
			descent - 1, num, len); 
	}
	else {
	    sprintf(num, "%d", j);
	    strcat(num, "\260E"); len = strlen(num);
	    width = XTextWidth(font, num, len);
	    XDrawString(display, pm, gc, i_lon + 2, 1 + ascent, num, len); 
	    XDrawString(display, pm, gc, i_lon + 2, y_size - descent - 1,
			num, len);
	}
    }
}

/*
  Create/Initialize variables used to overlay a grid on the image. This
  function can be called a number of times - for example once when the
  program starts and then again when the user changes size of the image
  display area.
*/

void
init_image_grid()
{
    if (grid)
	XFreePixmap(display, grid);
    
    grid = XCreatePixmap(display, root_window, window_pixmap_x,
			 window_pixmap_y, 1);

    if (!grid_gc)
	grid_gc = XCreateGC(display, grid, 0, NULL);

}

/*
   Draw the grid. Currently assumes a rectangular image projection.

   If the lat/lon corner points are all 0.0, then do not draw the grid.
*/

void
show_image_grid(browse, force_redraw)
BROWSE *browse;
int force_redraw;
{
    double new_tllon, new_brlon, new_tllat, new_brlat;

    get_lat_lon(browse->pyramid, 0, 0, &new_tllat, &new_tllon);
    get_lat_lon(browse->pyramid, window_pixmap_x, window_pixmap_y,
		&new_brlat, &new_brlon);

    if (new_tllon == 0.0 && new_brlon == 0.0
	&& new_tllat == 0.0 && new_brlat == 0.0)
	return;
    
    if (force_redraw		/* eg when image_size changes */
	|| grid_tllat != new_tllat
	|| grid_tllon != new_tllon
	|| grid_brlat != new_brlat
	|| grid_brlon != new_brlon) {

	/* erase the current grid */
	XSetForeground(display, grid_gc, 0);
	XSetBackground(display, grid_gc, 0);
	XFillRectangle(display, grid, grid_gc, 0, 0, window_pixmap_x,
		       window_pixmap_y); 

	/* draw a new grid */
	XSetForeground(display, grid_gc, 1);
	
	grid_tllat = new_tllat;
	grid_tllon = new_tllon;
	grid_brlat = new_brlat;
	grid_brlon = new_brlon;
	
	grid_font = XLoadQueryFont(display, a_data.grid_font);
	XSetFont(display, grid_gc, grid_font->fid);
	
	if (!alloc_color(a_data.grid_color, &grid_color))
	    err_dialog("Could not allocate grid color");
	
	grid_line_width = a_data.grid_line_width;
	XSetLineAttributes(display, grid_gc, grid_line_width, LineSolid,
			   CapButt, JoinBevel);  
	
	draw_grid_lines(browse->pyramid, grid_tllat, grid_tllon, grid_brlat,
			grid_brlon, window_pixmap_x, window_pixmap_y,
			a_data.grid_spacing, grid, grid_gc, grid_font);  

	/* set image_gc's clipmask so that everything passes */
	XSetForeground(display, clip_gc, 1);
	XFillRectangle(display, clip_mask, clip_gc, 0, 0, window_pixmap_x,
		       window_pixmap_y); 
    }
    

    /* overlay grid on pixmap (using grid as clipmask) */
    XSetForeground(display, image_gc, grid_color);
    XSetClipMask(display, image_gc, grid);
    XCopyPlane(display, grid, window_pixmap, image_gc, 0, 0,
	       window_pixmap_x, window_pixmap_y, 0, 0, 1);
    
    /* write grid on clip mask  - prevent screen updates from erasing grid */
    XSetForeground(display, clip_gc, 0);
    XSetFunction(display, clip_gc, GXandInverted);
    XCopyArea(display, grid, clip_mask, clip_gc, 0, 0,
	      window_pixmap_x, window_pixmap_y, 0, 0);
    
    /* update clip mask */
    XSetClipMask(display, image_gc, clip_mask);
    XSetFunction(display, clip_gc, GXcopy);
}

void
hide_image_grid()
{
    XSetForeground(display, clip_gc, 1);

    XFillRectangle(display, clip_mask, clip_gc, 0, 0, window_pixmap_x,
		   window_pixmap_y); 
    
    XSetClipMask(display, image_gc, clip_mask);
    XSetClipOrigin(display, image_gc, 0, 0);
    XSetForeground(display, clip_gc, 0);
}
