/*
  Copyright 1992 University of Rhode Island
 
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose is hereby granted without fee, provided that
  the above copyright notice appear in all copies and that both that
  copyright notice and this permission notice appear in supporting
  documentation, and that the name of U.R.I. not be used in advertising or
  publicity pertaining to distribution of the software without specific,
  written prior permission.  U.R.I. makes no representations about the
  suitability of this software for any purpose.  It is provided "as is"
  without express or implied warranty.
 
  U.R.I. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.R.I.
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
/*
  Functions for PIXMAP_MARKs. These allow graphical markers to be overlaid on
  pixmaps. The marks are drawn using the globals window_pixmap, clip_mask,
  the_image and the associated GCs.
  
  2/5/92 jhrg
*/

/*
 * $Log: pixmap-mark.c,v $
 * Revision 3.3  1993/02/02  16:12:32  jimg
 * Added conditional include for debugging malloc header.
 *
 * Revision 3.2  1992/11/19  21:14:30  jimg
 * ?
 *
 * Revision 3.1  1992/10/19  14:57:32  jimg
 * Memory allocation bug fixed.
 *
 * Revision 3.0  1992/07/28  15:24:44  jimg
 * All sources for xbrowse modified to use the libcomm.a and liberrmsg.s
 *
 * Revision 1.1  1992/03/23  20:39:01  jimg
 * Initial revision
 */

static char rcsid[]="$Id: pixmap-mark.c,v 3.3 1993/02/02 16:12:32 jimg Exp jimg $";

#include <X11/Intrinsic.h>	/* toolkit */

#ifdef MALLOC_DEBUG
#include <malloc.h>
#endif 

#include "xbrowse.h"
#include "dialog.h"
#include "pixmap-mark.h"

extern Display *display;
extern Widget the_image;
extern Pixmap window_pixmap;	/* holds image offscreen - server side */
extern Pixmap clip_mask;	/* clip mask so that hash marks show */
extern GC image_gc;		/* used to transfer window_pixmap to screen */
extern GC clip_gc;		/* used to update clip mask */
extern Window root_window;

PIXMAP_MARK *
create_pixmap_mark(bits, width, height, x_hot, y_hot, color1, color2)
char *bits;
unsigned int height, width;
int x_hot, y_hot;
char *color1, *color2;
{
    PIXMAP_MARK *mark = XtNew(PIXMAP_MARK);
    
    mark->bitmap = XCreateBitmapFromData(display, root_window, bits, width,
					 height);
#ifdef NEVER
    mark->backing_store = XCreatePixmap(display, root_window, width, height,
					8);
#endif
    if (!mark->bitmap /* || !mark->backing_store*/ ) {
	err_dialog("Could not create pixmap mark");
	return NULL;
    }

    if (!alloc_color(color1, &mark->color))
	err_dialog("Could not allocate color for mark");
    if (!alloc_color(color2, &mark->color2))
	err_dialog("Could not allocate color for mark");

    mark->width = width;
    mark->height = height;
    mark->x_hot = x_hot;
    mark->y_hot = y_hot;

    return mark;
}

void
free_pixmap_mark(mark)
PIXMAP_MARK *mark;
{
    if (!mark) {
	XFreePixmap(display, mark->bitmap);
#ifdef NEVER
	XFreePixmap(display, mark->backing_store);
#endif
	XtFree((char *)mark);
    }
}

/*
  Draw a mark at x,y that is size pixels wide and size pixel high. This
  function, along with the other `hash mark' code could be more efficient,
  but it is rarely called.

  Backing store allows bitmaps to be used as stencils; the interior pixels
  don't appear white, but rather have the colors of the image before the mark
  was added.
*/

void 
draw_pixmap_mark(mark, x, y)
PIXMAP_MARK *mark;
int x;
int y;
{
    XExposeEvent e;

    XSetForeground(display, image_gc, mark->color);
    XSetForeground(display, clip_gc, 0);
    XSetFunction(display, clip_gc, GXandInverted);
    
#ifdef NEVER
    /* save backing store */
    XCopyArea(display, window_pixmap, mark->backing_store, image_gc,
	      x - mark->x_hot, y - mark->y_hot, mark->width, mark->height,
	      0, 0);
#endif
    
    /* write mark on pixmap */
    XSetClipMask(display, image_gc, mark->bitmap);
    XSetClipOrigin(display, image_gc, x - mark->x_hot, y - mark->y_hot);
    XCopyPlane(display, mark->bitmap, window_pixmap, image_gc, 0, 0,
	       mark->width, mark->height, x - mark->x_hot, y - mark->y_hot, 1);

    /* write mark on clip mask */
    XCopyArea(display, mark->bitmap, clip_mask, clip_gc, 0, 0, mark->width,
	      mark->height, x - mark->x_hot, y - mark->y_hot);

    /* update clip mask */
    XSetClipMask(display, image_gc, clip_mask);
    XSetClipOrigin(display, image_gc, 0, 0);
    XSetFunction(display, clip_gc, GXcopy);
    
#ifdef NEVER
    /* refresh backing store */
    XCopyArea(display, mark->backing_store, window_pixmap, image_gc, 0, 0,
	      mark->width, mark->height, x - mark->x_hot, y - mark->y_hot); 
#endif
    
    /* move the new contents of the pixmap to the window */
    e.x = x - mark->x_hot; e.width = mark->width;
    e.y = y - mark->y_hot; e.height = mark->height;

    redraw_picture(the_image, &e);
}

/*
  Chnage the color of an existing mark. Infrequently used.
*/

void 
draw_pixmap_mark2(mark, x, y)
PIXMAP_MARK *mark;
int x;
int y;
{
    XExposeEvent e;

    XSetForeground(display, clip_gc, 1);
    XFillRectangle(display, clip_mask, clip_gc, x - mark->x_hot,
		   y - mark->y_hot, mark->width, mark->height);

    XSetClipMask(display, image_gc, clip_mask);
    XSetClipOrigin(display, image_gc, 0, 0);
    XSetForeground(display, image_gc, mark->color2);

    /* write mark on pixmap */
    XSetClipMask(display, image_gc, mark->bitmap);
    XSetClipOrigin(display, image_gc, x - mark->x_hot, y - mark->y_hot);
    XCopyPlane(display, mark->bitmap, window_pixmap, image_gc, 0, 0,
	       mark->width, mark->height, x - mark->x_hot, y - mark->y_hot, 1);
    
    /* update clip mask */
    XSetClipMask(display, image_gc, clip_mask);
    XSetClipOrigin(display, image_gc, 0, 0);
    XSetFunction(display, clip_gc, GXcopy);

    /* move the new contents of the pixmap to the window */
    e.x = x - mark->x_hot; e.width = mark->width;
    e.y = y - mark->y_hot; e.height = mark->height;

    redraw_picture(the_image, &e);

    XSetForeground(display, clip_gc, 0);
}

void 
erase_pixmap_mark(mark, x, y)
PIXMAP_MARK *mark;
int x;
int y;
{
    XSetForeground(display, clip_gc, 1);
    XFillRectangle(display, clip_mask, clip_gc, x - mark->x_hot,
		   y - mark->y_hot, mark->width, mark->height);

    XSetClipMask(display, image_gc, clip_mask);
    XSetClipOrigin(display, image_gc, 0, 0);
    XSetForeground(display, clip_gc, 0);
}
