/*
  Copyright 1991 University of Rhode Island
 
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose is hereby granted without fee, provided that
  the above copyright notice appear in all copies and that both that
  copyright notice and this permission notice appear in supporting
  documentation, and that the name of U.R.I. not be used in advertising or
  publicity pertaining to distribution of the software without specific,
  written prior permission.  U.R.I. makes no representations about the
  suitability of this software for any purpose.  It is provided "as is"
  without express or implied warranty.
 
  U.R.I. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.R.I.
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
  PYRAMID is a type used to manage the image information that is
  to be displayed on the screen. The screen is assumed to be a N by N
  matrix where each element holds the value of one pixel. Each pixel is
  assumed to be one byte.
  
  Modifcations
  
  7/26/90
  Added clearDisplay function.

  12/11/90
  Ported to unix/X.

  4/8/91
  Replaced some function calls with macros.

  4/18/91
  Added code to handle second region. The second region mirrors the region
  used by the server. Having it simplifies the optimization scheme.

  6/?/91
  regions were changed almost completely - They are now implemented
  differently on the client and server. On the client (here) a region is a
  list of vertices and their bounding box.

  8/3/91
  Added `user_region'. This allows the user to specify a region either via
  the mouse or from a file once and hit `use region' when each image is
  displayed (w/o re reading the file, ...)

  8/9/91
  Aded macros for commonly called functions.
*/

/* $Log: pyramid.c,v $
 * Revision 3.3  1993/02/02  16:12:32  jimg
 * Added conditional include for debugging malloc header.
 *
 * Revision 3.2  1992/11/19  21:14:30  jimg
 * ?
 *
 * Revision 3.1  1992/10/19  14:57:32  jimg
 * Added code to handle `query-based' access to the server.
 *
 * Revision 3.0  1992/07/28  15:24:44  jimg
 * All sources for xbrowse modified to use the libcomm.a and liberrmsg.a
 *
 * Revision 1.1  1992/03/23  20:39:01  jimg
 * Initial revision
 *
 */

static char rcsid[]="$Id: pyramid.c,v 3.3 1993/02/02 16:12:32 jimg Exp jimg $";

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <sys/time.h>

#include <X11/X.h>
#include <X11/Intrinsic.h>

#ifdef MALLOC_DEBUG
#include <malloc.h>
#endif

#define TRUE 1
#define FALSE 0

#ifndef MAX
#define MAX(a, b) ((a > b) ? a : b)
#endif

#include "vertex.h"
#include "region.h"
#include "ximage.h"
#include "pyramid.h"
#include "xbt-const.h"
#include "xbt-types.h"
#include "pixmap-mark.h"
#include "browse.h"
#include "dialog.h"
#include "errmsg.h"

/* Globals */

/* xbrowse.c */
extern GC image_gc;
extern Pixmap window_pixmap;
extern long window_pixmap_x;
extern long window_pixmap_y;
extern Display *display;
extern unsigned char image_pal[];

/*
  Allocate a new instance of PYRAMID. The initial level to be displayed
  is set via the init_level parameter. The array palette (which is used to
  transform the 0-255 values in an 8 bit image to pixel values in the user's
  current palette) is initialized to pass the image data w/o modification. The
  region instance associated with the pyramid structure is initialized to be
  large enough to describe a region that is the same size as the user's
  current window.  However, the region is not initialized to this rectangle
  (that is done by reset_pyramid). The XImage structures (raw_image holds the
  image data transmitted by the server and image holds the last row transmitted
  after it has been transformed by the mapping stored in palette) are
  initialized so that they can hold an image that is the smae size as the
  user's window. This function returns the newly initialized instance of
  PYRAMID if successful, otherwise it returns NULL.
*/

PYRAMID *
init_pyramid(init_level)
long init_level;
{
    PYRAMID *pyramid = XtNew(PYRAMID);

    if (!pyramid) 
	sys_quit_dialog("Could not allocate memory for PYRAMID");

    pyramid->init_level = init_level;
    pyramid->max_pyramid_level = floor_level(window_pixmap_x, window_pixmap_y);

    pyramid->image_region = init_region(window_pixmap_x, window_pixmap_y);
    if (!pyramid->image_region)
	exit(1);
    pyramid->region = init_region(window_pixmap_x, window_pixmap_y);
    if (!pyramid->region)
	exit(1);
    pyramid->user_region = NULL;

    pyramid->image = init_ximage(window_pixmap_y, 1);
    if (!pyramid->image)
	exit(1);
    pyramid->raw_image = init_ximage(window_pixmap_y, window_pixmap_x);
    if (!pyramid->raw_image)
	exit(1);

    pyramid->header = XtNew(struct image_header);
    if (!pyramid->header)
	sys_err_dialog("Could not create pyramid header structure");

    null_image_header(pyramid);
    
    /* set the level field when the server opens an image - initialize here */

    pyramid->level = 0;
    
    /* these fields can't be set until the size of the image is known */

    pyramid->image_size_x = 0;
    pyramid->image_size_y = 0;
    pyramid->image_origin_x = 0;
    pyramid->image_origin_y = 0;

    pyramid->max_image_level = 0;
    pyramid->level_ratio = 1;
    
    return pyramid;
}

/*
  Deallocate an instance of PYRAMID. This function frees the region
  struct and the two ximage structures in addition to deallocating storage
  associated with 'pyramid'. There is no return value.
*/

void 
free_pyramid(pyramid)
PYRAMID *pyramid;
{
    if (pyramid) {
	free_region(pyramid->image_region);
	free_region(pyramid->user_region);
	free_region(pyramid->region);
	
	XDestroyImage(pyramid->image);
	XDestroyImage(pyramid->raw_image);

	XtFree((char *)pyramid->header);
	
	XtFree((char *)pyramid);
    }
}

/*
  Reset the fields in an instance of PYRAMID so that a new image
  can be displayed. The region is reset to be a rectangle that encloses the
  highest level that the user can display in their window. The XImage
  structures that hold the data from the server (raw holds all the data
  recieved from  the server; image just holds the currewnt row after it has
  been transformed by the palette) are erased. Return FALSE if an error is
  detected, TRUE otherwise.
*/

long
reset_pyramid(pyramid, image_x, image_y)
PYRAMID *pyramid;
long image_x;
long image_y;
{
    /* the max pyramid level must be reset since zooming an image may change
       it (e.g. the max_pyramid_level of a 512^2 window that displays a
       (zoomed) 2048^2 image is 11 NOT 9). */
    
    pyramid->max_pyramid_level = floor_level(window_pixmap_x, window_pixmap_y);

    if (image_x != pyramid->image_size_x || image_y != pyramid->image_size_y) {
	pyramid->max_image_level = ceil_level(image_x, image_y);
	pyramid->image_size_x = image_x;
	pyramid->image_size_y = image_y;
    }

    pyramid->current_row = 0;
    
    pyramid->level = pyramid->init_level;
    pyramid->scale_factor = 1 << (pyramid->max_pyramid_level
				  - pyramid->init_level);

    pyramid->level_ratio = MAX(1, 1 << (pyramid->max_image_level
					- pyramid->max_pyramid_level));
    
    /* set size of initial region to be rectangle enclosing higest level
       displayable in user's current window
       NB: This is NOT the same as making the entire window the current
       region since the image may be an odd size that will not square up
       inside the current window (or vice-versa). */

    if (!copy_region(pyramid->region, pyramid->image_region))
	return FALSE;

    make_rect_region(pyramid, 0, 0, image_y, image_x);

    /* new images are never zoomed, top,left is 0,0 */
    pyramid->image_origin_x = 0;
    pyramid->image_origin_y = 0;
    
    clear_ximage(pyramid->image);
    clear_ximage(pyramid->raw_image);

    return TRUE;
}

/*
  Write the current contents of 'raw_image' to the user's window. This
  function does not return a value.
*/

void 
display_pyramid(pyramid)
PYRAMID *pyramid;
{
    long i;

    for (i=0; i<window_pixmap_y; i++)
	display_ximage_row(pyramid->image, i);
}

/*
  Advance to the next level of resolution (i.e., the next level of the
  image pyramid). Return TRUE if successful, FALSE otherwise (i.e. if
  already at maximium resolution).
*/

long 
pyramid_next_level(pyramid)
PYRAMID *pyramid;
{
    /* if displaying as much as can fit in the user's window OR if the user
       has reached the upper limit on the level of data they can see, return
       FALSE */

    if (pyramid->level >= pyramid->max_pyramid_level) {
	print_level(pyramid, FALSE);
	return FALSE;
    }
    
    pyramid->level++;
    pyramid->scale_factor = 1 << (pyramid->max_pyramid_level - pyramid->level);
    
    pyramid->current_row = pyramid->region->top;
    
    return TRUE;
}

/*
  Initialize the current values in the image header structure with null values.
  NB: the image header struct is filled in when it is read from the socket
  using get_struct(). Thus, there is no code here to fill in the structure.
*/

void
null_image_header(pyramid)
PYRAMID *pyramid;
{
    struct image_header *h = pyramid->header;

    strcpy(h->image_name, "");
    strcpy(h->projection, "");
    strcpy(h->x_to_lon, "");
    strcpy(h->y_to_lat, "");
    strcpy(h->lon_to_x, "");
    strcpy(h->lat_to_y, "");
    h->top_left_lat = 0.0;
    h->top_left_lon = 0.0;
    h->bot_right_lat = 0.0;
    h->bot_right_lon = 0.0;
    strcpy(h->data_units, "");
    strcpy(h->data_eqn, ""),
    h->temp_slope = 0.0;
    h->temp_intcp = 0.0;
    strcpy(h->sat_type, "");
    h->resolution = 0.0;
}
    
/*
  Given a pointer to a PYRAMID and a pointer to byte data from the
  image server, expand that data so that it can be displayed in the user's
  current window. Bytes are read from the packet until a complete row of data
  has been extracted. Each byte that is read is `expanded' so that it fills the
  sequence of bytes necessary to represent the current level of resolution
  given the user's current window size and the image file size. The `exapaned'
  data is stored in the XImage struct raw_image. Once a complete row has been
  processed, it is then used to `fill down' in the same fashion as each pixel
  was used to `fill right'. The pixels in the current row are then mapped to
  the user's palette and transferred to the offscreen pixmap (via the XImage
  structure image). The function returns the number of bytes read from packet.

  Note: This function handles the special case where, at some levels,
  portions of a region are 0 blocks wide. 

  Note 2: The old image data must be stored since we allow the user to save
  an image file. Otherwise, we could get away with just using stuff from the
  offscreen pixmap. 
*/

long
add_sampled_row(pyramid, packet, len, start1, row1, opt)
PYRAMID *pyramid;
char *packet;
short len;
short start1;
short row1;
char opt;
#ifdef NEVER
char rll;			/* TRUE if packet contains rll encoded data */
char rll_byte;			/* value of rrl encoded data */
#endif
{
    register long i, j, k;
    register long scale_factor = pyramid->scale_factor;
    register long interval = opt ? scale_factor << 1 : scale_factor;
    XImage *image = pyramid->image;
    XImage *raw_image = pyramid->raw_image;
    register char *image_row = image->data;
    register char *raw_image_row;
    register short row, start;
    
    start = (start1 - pyramid->image_origin_x) / pyramid->level_ratio;
    row = (row1 - pyramid->image_origin_y) / pyramid->level_ratio;

    raw_image_row = get_ximage_row(raw_image, row);
   
#ifdef NEVER
    /* rll code - under development */
    for (i=start, k=0, n = 0; n < len; i+=interval, k++, n++) {
	if (rll && (rll_byte == *(packet + k++))) {
	    for (m = 0 ; m < *(packet + k); m++, n++, i += interval) {
		for (j=0; j<scale_factor; j++)
		    *(raw_image_row+i+j) = rll_byte;
	    }
	}
	else
	    for (j=0; j<scale_factor; j++)
		*(raw_image_row+i+j) = *(packet+k);
    }
#endif
    
    for (i=start, k=0; k < len; i+=interval, k++)
	for (j=0; j<scale_factor; j++)
	    *(raw_image_row+i+j) = *(packet+k);
    
    for (i = row + 1; i < row + scale_factor; i++)
	copy_ximage_row(raw_image, i, row);

    copy_using_palette(image_row, raw_image_row);

    for (i = row; i < row + scale_factor; i++)
	XPutImage(display, window_pixmap, image_gc, image,
		  0, 0, 0, i, image->width, 1);

    /* update this for state_machine */
    pyramid->current_row = row1 + scale_factor * pyramid->level_ratio;
    
    return k;			/* return nbytes used from packet */
}

/*
  Copy the bytes pointed to by raw_image to those pointed to by image using
  the palette field of display to map the values in raw_image to values in
  image. Optimize this code. 4/17/91.
*/

void
update_display(pyramid)
PYRAMID *pyramid;
{
    register long i, j;
    
    for (i = 0; i < window_pixmap_y; i += pyramid->scale_factor) {
	copy_using_palette(pyramid->image->data,
			   get_ximage_row(pyramid->raw_image, i));

	for (j = i; j < i + pyramid->scale_factor; j++)
	    XPutImage(display, window_pixmap, image_gc, pyramid->image,
		      0, 0, 0, j, pyramid->image->width, 1);
    }
}
	
void
copy_using_palette(image_row, raw_image_row)
unsigned char *image_row;
unsigned char *raw_image_row;
{
    register long i;

    for(i = 0; i < window_pixmap_x; i++)
	*(image_row + i) = *(image_pal + *(raw_image_row + i));
}

/*
  Given a pointer to a pyramid, return the maximum number of times that the
  current region's size can be increased. This function takes into account
  the size of the image relative to the current window as well as the size of
  the current region relative to the current window. It returns the number of
  times that region may be zoomed. For a small region in a 512^2 window
  displaying a 2048^2 image, pyramid_zoom_factor returns 4. This means that
  one could zoom from level 9 to level 11 in the current window. If the
  current region cannot be zoomed, then this function returns 0.

  7/29/91
  Added code that constrains the bounding box used to determine how much
  zooming can take place to multiples of the current levels resolution. This
  might be over kill (maybe it could be 2x finer resolution), but it is a
  hack anyway. This is necessary for zoomfactor to always be usable by
  calc_new_origin.
*/

long
pyramid_zoom_factor(pyramid)
PYRAMID *pyramid;
{
    REGION *region = pyramid->region;
    
    long bb_height = (((region->bottom - region->top) / pyramid->level)
		     + 1) * pyramid->level;
    long bb_width = (((region->right - region->left) / pyramid->level)
		    + 1) * pyramid->level;

    long region_zoom = 1 << (floor_level(region->max_size_x,
					region->max_size_y)
			    - ceil_level(bb_height, bb_width));

    long image_zoom = 1 << (floor_level(pyramid->image_size_x,
				      pyramid->image_size_y) -
			   pyramid->max_pyramid_level);

    long zoom = (image_zoom < region_zoom) ? image_zoom : region_zoom;

    zoom = 2 * (long) ceil(M_LOG2E * log((double) zoom));

    return zoom;
}

/*
  Once the region has been zoomed by zoom_factor, adjust the various fields
  of pyramid. The fields that are updated are: current_row, scale_factor (not
  the same as region's scale_factor!), resolution and max_pyramid_level. This
  function does not return a value.

  N.B. I can't think of a case where the pyramid's scale_factor is actually a
  different value then the region's. Rather, they are different fields for
  `historical' reasons.
*/

void
zoom_pyramid(pyramid, zoom_factor, y_offset)
PYRAMID *pyramid;
long zoom_factor;
long y_offset;
{
    pyramid->max_pyramid_level +=
	(long) ceil(M_LOG2E * log((double) zoom_factor));
    pyramid->scale_factor *= zoom_factor;
    pyramid->level_ratio = 1 << (pyramid->max_image_level
				 - pyramid->max_pyramid_level);
    
    region_bounding_box(pyramid);
}

long
can_zoom(pyramid)
PYRAMID *pyramid;
{
    if (window_pixmap_x >= pyramid->image_size_x
	&& window_pixmap_y >= pyramid->image_size_y)
	return FALSE;
    else
	return TRUE;
}

/*
  Given the SCREEN coordinates, return the temperature of a pixel. Return value
  as a double to avoid problems w/float->double promotion on older compilers.
  Computed using `y = mx + b'.
*/

double
get_temperature(pyramid, x, y)
PYRAMID *pyramid;
long x, y;
{
    return (double) (XGetPixel(pyramid->raw_image, x, y)
		     * pyramid->header->temp_slope + pyramid->header->temp_intcp);

}

/*
   These macros are used for mapping screen coordinates to locations on
   images which use a mercator projection.
*/

#define D_TO_R(d) ((d) * M_PI / 180.0)
#define R_TO_D(r) ((r) * 180.0 / M_PI)

#define MAP_X_SF(s_x, s_lon) ((s_x) / (s_lon))
#define MAP_Y_SF(s_y, s_lat) ((s_y) / log( tan( M_PI_4 + D_TO_R(s_lat)/2.0 )))

#define MER_MAP_X_TO_LON(x, s_x, s_lon, o_lon) \
    ( ((x) / MAP_X_SF(s_x, s_lon)) + o_lon )

#define MER_MAP_Y_TO_LAT(y, s_y, s_lat, o_lat) \
    (R_TO_D((2.0 * (atan(exp( (y) / MAP_Y_SF(s_y, s_lat) )) - M_PI_4) )) + o_lat)

#define MER_MAP_LON_TO_X(lon, s_x, s_lon, o_lon) \
    (int)((MAP_X_SF(s_x, s_lon) * (lon - o_lon)) )

#define MER_MAP_LAT_TO_Y(lat, s_y, s_lat, o_lat) \
    (int)(MAP_Y_SF(s_y, s_lat) * log(tan( M_PI_4 + D_TO_R((lat - o_lat))/2.0)))

/*
  Given x and y in SCREEN coordinates, return the corresponding lat and lon.
  This function takes into account the offset and scaling caused by zoomed
  images. Assume a linear projection.
*/

void
get_lat_lon(pyramid, x, y, lat, lon)
PYRAMID *pyramid;
long x, y;
double *lat, *lon;
{
    long x_loc = (x * pyramid->level_ratio + pyramid->image_origin_x);
    long y_loc = (y * pyramid->level_ratio + pyramid->image_origin_y);

    if (strcmp(pyramid->header->projection, "Mercator") == 0) {
	*lon = MER_MAP_X_TO_LON(x_loc, pyramid->image_size_x,
				(pyramid->header->bot_right_lon
				 - pyramid->header->top_left_lon),
				pyramid->header->top_left_lon);
	*lat = MER_MAP_Y_TO_LAT(y_loc, pyramid->image_size_y,
				(pyramid->header->bot_right_lat
				 - pyramid->header->top_left_lat),
				pyramid->header->top_left_lat);
    }
    else {			/* cylindrical-equirectangular is default */
	*lon = (double) (((((float)x_loc / (float)pyramid->image_size_x)
			   * (pyramid->header->bot_right_lon
			      - pyramid->header->top_left_lon))
			  + pyramid->header->top_left_lon));
    
	*lat = (double) (((((float)y_loc / (float)pyramid->image_size_y)
			   * (pyramid->header->bot_right_lat
			      - pyramid->header->top_left_lat))
			  + pyramid->header->top_left_lat));
    }
}

/*
  Given a (lon, lat) point on the current image, return the screen
  coordinates at that (lon, lat). Assume a linear projection.
*/

void
get_x_y(pyramid, lon, lat, x, y)
PYRAMID *pyramid;
double lat, lon;
long *x, *y;
{
    int image_x, image_y;
    double tll, brl, wrap_lon;	/* temps for shifted lon corner pts */

    lon += 180.0;
    tll = pyramid->header->top_left_lon + 180.0;
    brl = pyramid->header->bot_right_lon + 180.0;

    if (tll > brl) {		/* account for date line crossing */
	wrap_lon = 360.0;
	brl += 360.0;
    }
    else
	wrap_lon = 0;

    if (tll > lon)
	lon += wrap_lon;
    
    if (strcmp(pyramid->header->projection, "Mercator") == 0) {
#ifdef NEVER
	err_print("lon: %lf, lat:%lf", lon, lat);
	err_print("map y sf: %lf (s_x: %f, s_lon: %f)",
		  MAP_Y_SF((float)pyramid->image_size_y,
			   (float)(pyramid->header->bot_right_lat
				   - pyramid->header->top_left_lat)),
		  (float)pyramid->image_size_x,
		  (float)(pyramid->header->bot_right_lat
			  - pyramid->header->top_left_lat));
#endif
	
	image_x = MER_MAP_LON_TO_X(lon, (float)pyramid->image_size_x,
				   (float)(brl - tll), tll);
	image_y = MER_MAP_LAT_TO_Y(lat, (float)pyramid->image_size_y,
				   (float)(pyramid->header->bot_right_lat
					   - pyramid->header->top_left_lat),
				   (float)pyramid->header->top_left_lat);
    }
    else {			/* cylindrical-equirectangular is default */
	image_x = (int)(((lon - tll) / (brl - tll))
			* pyramid->image_size_x);
	
	image_y = (int)(((lat - pyramid->header->top_left_lat)
			 / (pyramid->header->bot_right_lat
			    - pyramid->header->top_left_lat))
			* pyramid->image_size_y);
    }

#ifdef NEVER
    err_print("lon: %lf, lat: %lf, image_x: %d, image_y:%d", lon, lat,
	      image_x, image_y);
#endif
    
    *x = image_to_screen_x(pyramid, image_x);
    *y = image_to_screen_y(pyramid, image_y);
}

long
image_to_screen_x(pyramid, image_x)
PYRAMID *pyramid;
long image_x;
{
    return (image_x - pyramid->image_origin_x) / pyramid->level_ratio;
}

long
image_to_screen_y(pyramid, image_y)
PYRAMID *pyramid;
long image_y;
{
    return (image_y - pyramid->image_origin_y) / pyramid->level_ratio;
}

long
screen_to_image_x(pyramid, screen_x)
PYRAMID *pyramid;
long screen_x;
{
    return screen_x * pyramid->level_ratio + pyramid->image_origin_x;
}

long
screen_to_image_y(pyramid, screen_y)
PYRAMID *pyramid;
long screen_y;
{
    return screen_y * pyramid->level_ratio + pyramid->image_origin_y;
}

#ifndef MACROS

/*
  Return `a' reduced to the largest number of `unit' it can hold.
*/

long
unitize(a, unit)
long a, unit;
{
    return (( (long) a / (long) unit) * (long) unit);
}

/*
  Given the X and Y dimensions of an image, return the smallest pyramid level
  that can hold the entire image (i.e., find the smallest 2^n by 2^n array
  that can hold the image and return n).
*/

long
ceil_level(x, y)
long x;
long y;
{
    return (long) ceil(M_LOG2E * log((double) ((x > y) ? x : y)));
}

long
floor_level(x, y)
long x;
long y;
{
    return (long) floor(M_LOG2E * log((double) ((x < y) ? x : y)));
}

#endif /* MACROS */


