/*
  Copyright 1991 University of Rhode Island
 
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose is hereby granted without fee, provided that
  the above copyright notice appear in all copies and that both that
  copyright notice and this permission notice appear in supporting
  documentation, and that the name of U.R.I. not be used in advertising or
  publicity pertaining to distribution of the software without specific,
  written prior permission.  U.R.I. makes no representations about the
  suitability of this software for any purpose.  It is provided "as is"
  without express or implied warranty.
 
  U.R.I. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.R.I.
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
  Regions are used to constrain the information transmitted by the server to a
  specific part of the image. Regions are defined by the user as a polygon.
  They are initially represented as a list of vertex points. However, the
  software that determines which portion of the image to sample and transmit
  uses an array of line starts/stops to represent regions. Regions are also
  used to implenment 'zooming' (i.e. displaying a portion of an image on the
  display)

  Modifications

  7/24/90 - Changed MarkLineEnds to compensate for different size
  display/images.  Previously the vertex points of the region were scaled when
  an image file was a different size than the display (i.e., larger). However,
  when a scaled set of vertices were used to build the line start/stop array on
  the server, that array did not match the corresponding array on the client.
  This was because of the aliasing of the lines that marked the polygon's
  boundry.  The regions are no longer scaled, rather MarkLineEnds 'magnifies'
  the 'line' of the smaller region (including aliasing).

  4/9/91 - remove the regionMap structure in an effort to clean up the
  'region mess' problem. Moved the hash mark stuff to the callbacks.c
  file. Remove unused regionFile field and functions.

  4/16/91 - hacked DrawLine so that it handled the zooming region case
  correctly. DrawLine is based on a line drawing algorithm. When a user decided
  that they want to zoom in on a region of the image, the region must be maped
  (via map_region_vertices). However, a region's lines (which are drawn between
  vertices and define the boundry of a region) are aliased on both the client
  and the server. The server is very careful to preserve this aliasing even
  though the actual size of the region (and resolution of the image data) may
  be different on the server side than on the client side.  When a region is
  zoomed it must also preserve this aliasing so that the new zoomed region
  matches the region defined on the server side of the system.  The function(s)
  involving region mapping (translating the region vertices to line start/stop
  entries in the lines array) use the region scaleFactor to maintain the
  screen's aliasing pattern.

  8/8/91
  Regions are fundementally different now; the row offsets, ... are computed
  solely by the server and transmitted along with the image data. Thus
  regions no longer compute the line start stop array or any of the stuff
  that has to do with the different levels.

  8/26/91
  Added code to copy regions.
*/

/* $Log: region.c,v $
 * Revision 3.3  1993/02/02  16:12:32  jimg
 * Added conditional include for debugging malloc header.
 *
 * Revision 3.2  1992/11/19  21:14:30  jimg
 * ?
 *
 * Revision 3.1  1992/10/19  14:57:32  jimg
 * Examined vertex ordering bug - not fixed.
 *
 * Revision 3.1  1992/10/19  14:57:32  jimg
 * Examined vertex ordering bug - not fixed.
 *
 * Revision 3.0  1992/07/28  15:24:44  jimg
 * All sources for xbrowse modified to use the libcomm.a and liberrmsg.a
 *
 * Revision 1.1  1992/03/23  20:39:01  jimg
 * Initial revision
 *
 */

static char rcsid[]="$Id: region.c,v 3.3 1993/02/02 16:12:32 jimg Exp jimg $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <sys/types.h>

#include <X11/X.h>
#include <X11/Intrinsic.h>

#ifdef MALLOC_DEBUG
#include <malloc.h>
#endif

#define TRUE 1
#define FALSE 0

#include "vertex.h"
#include "region.h"
#include "pyramid.h"
#include "dialog.h"
#include "errmsg.h"

/*
  Initialize a new instance of REGION. max_size_x and max_size_y are the
  maximum size of the region (or its bounding box) and scale_factor is the
  ratio between the resolution of the region's coordinate system and the
  resolution of the window used to display the image data. Normally,
  scale_factor is 1.0, however, once the user decides to zoom in on a region
  scale_factor takes on a value > 1.  Look at the function map_region_vertices
  to understand how scale_factor is used. This function returns the new REGION
  if successful, otherwise it returns NULL.

  8/8/91
  Removed scale_factor and scan_lines code.
*/

REGION *
init_region(max_size_x, max_size_y)
long max_size_x;
long max_size_y;
{
    REGION *region = XtNew(REGION);

    if (!region) {
	sys_err_dialog("Could not allocate structure REGION");
	return NULL;
    }
	
    region->vertices = init_vertices();
    if (!region->vertices)
	return NULL;		/* message printed by init_vertices */

    region->max_size_x = max_size_x;
    region->max_size_y = max_size_y;
    
    return region;
}

/*
  Free storage associated with an instance of REGION. This function
  does not return a value.
*/

void 
free_region(region)
REGION *region;
{
    if (region) {
	free_vertices(region->vertices);
	XtFree((char *)region);
    }
}

/*
  Given that file is a FILE * type pointer to a file opened for reading,
  read_region reads from that file a sequence of points and stores then in
  region's vertex list. The file is a plain text file and the points are
  stored as x, y pairs (two values/line). Any line starting with '\n',
  '\' or '#' is taken to be a comment. This function returns the number of
  points actually read
*/

long 
read_region(region, file)
REGION *region;
FILE *file;
{
    char line[256];
    char *c;
    long line_count = 0;
    long x, y;

    assert(region != 0);

    free_vertices(region->vertices);
    region->vertices = init_vertices();
    if (!region->vertices)
	return FALSE;
    
    while(!feof(file)) {
	c = fgets(line, 255, file);
	if (c && line[0]!='/' && line[0]!='\n' && line[0]!='#') {
	    sscanf(line, "%d %d %*[^\n]", &x, &y);
	    if (!add_vertex(region->vertices, x, y))
		return FALSE;
	    line_count++;
	}
    }

    return line_count;
}

/*
  Given that file is a FILE pointer to a file opened for writing,
  write_region writes the vertices that describe the region region to that
  file. The x an y coordinates of each point are written to a separate line
  in the file. The file also contains a comment noting that it is a browse
  region file. This function does not return a value.
*/

void 
write_region(region, file)
REGION *region;
FILE *file;
{
    VERTICES *v;

    assert(region > 0);

    v = region->vertices;

    fprintf(file, "\n");
    fprintf(file, "/browse region file\n");

    reset_vertices(v);

    do {
	fprintf(file, "%5d %5d\n", get_vertex_x(v), get_vertex_y(v));
    } while (next_vertex(v));
}


/*
  Print out the vertices of a region. This function is intened for debugging.
  it does not return a value.
*/

void 
print_region_vertices(region)
REGION *region;
{
  assert(region > 0);

    reset_vertices(region->vertices);

    do {
	err_print("print_region_verteces: %d, %d",
		  get_vertex_x(region->vertices),
		  get_vertex_y(region->vertices));
    } while(next_vertex(region->vertices));
    
}

/*
  Find the region bounding box. This is recorded in REGION's top, left,
  bottom and right fields.
*/

void 
region_bounding_box(pyramid)
PYRAMID *pyramid;
{
    REGION *region = pyramid->region;
    VERTICES *v = region->vertices;
    
    region->top = region->max_size_y - 1;
    region->bottom = 0;
    region->left = region->max_size_x - 1;
    region->right = 0;

    reset_vertices(v);

    do {
	long x, y;

	x = image_to_screen_x(pyramid, get_vertex_x(v));
	y = image_to_screen_y(pyramid, get_vertex_y(v));
	
	region->top = (y < region->top) ? y : region->top;
	region->bottom = (y > region->bottom) ? y : region->bottom;
	region->left = (x < region->left) ? x : region->left;
	region->right = (x > region->right) ? x : region->right;
    } while(next_vertex(v));
}


/*
  Given a pointer to the current region, determine the largest zoom factor
  for that region (i.e. how many times can that region's size be doubled and
  still fit inside the current window). This function either returns the max
  zoom factor or 1 if the current region cannot be zoomed.

  NB: This function does not take into account the size of the image, so it
  may not be possible to zoom a region by the factor returned from this
  routine and then fill that region up with data.

  7/5/91
  Added the `+ 1' since regions include their verteces. Hence a region
  with top = 128 and bottom = 384 is really 257 (!) pixels high.
    
*/

long 
max_zoom_factor(region)
REGION *region;
{
    register long height_factor, width_factor; /* long division truncates !! */

    height_factor = region->max_size_y / (region->bottom - region->top + 1);
    width_factor = region->max_size_x / (region->right - region->left + 1);

    return (height_factor < width_factor) ? height_factor: width_factor;
}

/*
  Given a pointer to the current region and the desired zoom factor for the
  region, figure out which pixel will be moved to the window origin when the
  contents of the current region (plus some surrounding pixels probably) are
  expanded to fill the current window. The x and y coordinates are returned
  in the vars x and y; the function itself does not return a value.

  NB: this function defines many variables that aren't really necessary and
  performs some assignments that are superflous, but they add much in the way
  of understanding. This code is only called at most twice per image and then
  only if the user zooms, so there is no real reason to optimize (i.e.
  obfuscate) it.

  7/29/91
  Added code that constrains the region bounding box to multiples of the
  current level. This will ensure that new rows of data match the old data
  when the screen is zoomed.

  9/26/91
  Rewrote. 
*/

void
calc_new_origin(region, zoom_factor, scale_factor, x, y)
REGION *region;
long zoom_factor;
long scale_factor;		/* current scale_factor */
long *x, *y;			/* the dreaded value-result parameters */
{
     {
	long i_r_top = region->top * zoom_factor;
	long i_r_bot = region->bottom * zoom_factor;
	long i_r_bb_height = i_r_bot - i_r_top;
	long d_h_1 = (region->max_size_y - i_r_bb_height) / 2;
	long d_h_2 = d_h_1 + (region->max_size_y - i_r_bb_height) % 2;

	long i_top;
	long i_bot;
	long s_top;
	long s_bottom;

	if (i_r_top - d_h_1 < 0) {
	    i_top = 0;
	    i_bot = i_r_bot + d_h_2 + (d_h_1 - i_r_top);
	}
	else if (i_r_bot + d_h_2 >= (region->max_size_y - 1) * zoom_factor) {
	    i_bot = region->max_size_y * zoom_factor;
	    i_top = i_r_top - d_h_1 - (i_r_bot + d_h_2 - i_bot);
	}
	else {
	    i_top = i_r_top - d_h_1;
	    i_bot = i_r_bot + d_h_2;
	}

	s_top = (i_top / zoom_factor) / scale_factor * scale_factor;
	s_bottom = (i_bot / zoom_factor) / scale_factor * scale_factor;

	assert((s_bottom - s_top) * zoom_factor == region->max_size_y);
	
	*y = s_top;
    }

    {
	long i_r_left = region->left * zoom_factor;
	long i_r_right = region->right * zoom_factor;
	long i_r_bb_width = i_r_right - i_r_left;
	long d_w_1 = (region->max_size_x - i_r_bb_width) / 2;
	long d_w_2 = d_w_1 + (region->max_size_x - i_r_bb_width) % 2;

	long i_left;
	long i_right;
	long s_left;
	long s_right;

	if (i_r_left - d_w_1 < 0) {
	    i_left = 0;
	    i_right = i_r_right + d_w_2 + (d_w_1 - i_r_left);
	}
	else if (i_r_right + d_w_2 > (region->max_size_x - 1) * zoom_factor) {
	    i_right = region->max_size_x * zoom_factor;
	    i_left = i_r_left - d_w_1 - (i_r_right + d_w_2 - i_right);
	}
	else {
	    i_left = i_r_left - d_w_1;
	    i_right = i_r_right + d_w_2;
	}

	s_left = (i_left / zoom_factor) / scale_factor * scale_factor;
	s_right = (i_right / zoom_factor) / scale_factor * scale_factor;

	assert((s_right - s_left) * zoom_factor == region->max_size_x);
	
	*x = s_left;
    }
}
     
/*
  make_rect_region makes the current region a rectangle with the given top left
  and bottom right coordinates. This functiion does not return a value.
*/

void 
make_rect_region(pyramid, top, left, bottom, right)
PYRAMID *pyramid;
long top;
long left;
long bottom;
long right;
{
    VERTICES *vertices = pyramid->region->vertices;
    
    free_vertices(vertices);
    vertices = init_vertices();

    /* After the free/init calls above, vertices is no longer the same */
    /* as pyramid->region->vertices. When the region vertices are */
    /* referenced in region_bounding_box() - KABOOM. The line of code */
    /* below fixes this. jhrg 11/21/94 */
    pyramid->region->vertices = vertices;

    add_vertex(vertices, left, top);
    add_vertex(vertices, left, bottom);
    add_vertex(vertices, right, bottom);
    add_vertex(vertices, right, top);
    
    region_bounding_box(pyramid);
}

/*
  Copy the REGION src to dest. This includes copying the VERTICES. Returns
  FALSE if unable to copy the src region's vertices, otherwise returns TRUE.
*/

long
copy_region(dest, src)
REGION *dest;
REGION *src;
{
    free_vertices(dest->vertices); /* MUST free before struct copy */

    *dest = *src;

    /* allocate a new set of vertices */

    dest->vertices = init_vertices();
    if (!dest->vertices)
	return FALSE;

    if (vertices_p(src->vertices)) {
	reset_vertices(src->vertices);

	do {
	    add_vertex(dest->vertices, get_vertex_x(src->vertices),
		       get_vertex_y(src->vertices));
	} while (next_vertex(src->vertices));
    }
    
    return TRUE;
}

/*
  Using new vertex routines, sort the vertices so that they are listed in
  clockwise order. This allows poeple to input the vertices in any order and
  still have the polygon `make sense'. This can be turned on or off using the
  `sort_vertices' files of the application resources structure a_data.
*/

void
sort_region_vertices(region)
REGION *region;
{
    VERTICES *v = region->vertices;
    C_POINT init, final;
    
    init.x = 0, init.y = 0;

    arith_mean(v, &final);
    
    translate_origin(v, &init, &final);

    cart_to_polar(v);

    /* compare_theta is defined in vertex.c and declared in vertex.h */
    sort_polar_vertices(v, compare_theta);
    
    polar_to_cart(v);
    
    translate_origin(v, &final, &init);

}
