/*
  Copyright 1991 University of Rhode Island
 
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose is hereby granted without fee, provided that
  the above copyright notice appear in all copies and that both that
  copyright notice and this permission notice appear in supporting
  documentation, and that the name of U.R.I. not be used in advertising or
  publicity pertaining to distribution of the software without specific,
  written prior permission.  U.R.I. makes no representations about the
  suitability of this software for any purpose.  It is provided "as is"
  without express or implied warranty.
 
  U.R.I. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.R.I.
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
  A linked list type for the vertices used to describe a region.

  If really long lists of points are used to describe regions, then the type
  VERTICES should be changed so that points are stored in a dynamically
  allocated array rather than a linked list.
  
  jhrg

  11/27/91
  Added num_vertices field.
  Added get_num_vertices function.

  12/4/91
  Added function for coordinate transformations.
  
*/

/* $Log: vertex.c,v $
 * Revision 3.3  1993/02/02  16:12:32  jimg
 * Added conditional include for debugging malloc header.
 *
 * Revision 3.2  1992/11/19  21:14:30  jimg
 * ?
 *
 * Revision 3.1  1992/10/19  14:57:32  jimg
 * Examined vertex ordering bug - not fixed.
 *
 * Revision 3.1  1992/10/19  14:57:32  jimg
 * Examined vertex ordering bug - not fixed.
 *
 * Revision 3.0  1992/07/28  15:24:44  jimg
 * All sources for xbrowse modified to use the libcomm.a and liberrmsg.a
 *
 * Revision 1.1  1992/03/23  20:39:01  jimg
 * Initial revision
 *
 */

static char rcsid[]="$Id: vertex.c,v 3.3 1993/02/02 16:12:32 jimg Exp jimg $";

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include <X11/Intrinsic.h>

#ifdef MALLOC_DEBUG
#include <malloc.h>
#endif

#define TRUE 1
#define FALSE 0

#include "errmsg.h"
#include "vertex.h"

/*
  Create and initialize a vertex list. Return a pointer to the new list or
  NULL if unable to create the list.
*/

VERTICES *
init_vertices()
{
    VERTICES *v = XtNew(VERTICES);

    v->num_vertices = 0;
    v->head = NULL;
    v->tail = NULL;
    v->current = NULL;
    
    return v;
}

void
free_vertices(v)
VERTICES *v;
{   
    struct vertex *temp_v;

    if (v) {
	temp_v = v->head;	/* store first element */
	
	while(temp_v) {
	    v->head = v->head->next;
	    XtFree((char *)temp_v);
	    temp_v = v->head;	/* store the next element */
	}
	
	XtFree((char *)v);
    }
}

long
add_vertex(v, x, y)
VERTICES *v;
long x, y;
{
    if(!v->tail) {
	v->head = v->current = v->tail = XtNew(VERTEX);
	v->num_vertices++;
	v->current->p.c_p.x = x;
	v->current->p.c_p.y = y;
	v->current->next = NULL;
    }
    else {
	v->tail->next = XtNew(VERTEX);
	v->num_vertices++;
	v->current = v->tail = v->tail->next;
	v->current->p.c_p.x = x;
	v->current->p.c_p.y = y;
	v->current->next = NULL;
    }

    return TRUE;
}

void
reset_vertices(v)
VERTICES *v;
{
    assert(v->current);
    
    v->current = v->head;
}

/*
  Return FALSE if at the end of the list of vertices, otherwise move to the
  next element in the list and return TRUE.
*/

long
next_vertex(v)
VERTICES *v;
{
    if(v->current->next) {
	v->current = v->current->next;
	return TRUE;
    }
    else
	return FALSE;
}

long
get_num_vertices(v)
VERTICES *v;
{
    return (v->num_vertices);
}

/*
  Return the x coordinate of the current vertex. If at the end of te list,
  return FALSE.
*/

long
get_vertex_x(v)
VERTICES *v;
{
    assert(v->current);

    return v->current->p.c_p.x;
}

long
get_vertex_y(v)
VERTICES *v;
{
    assert(v->current);

    return v->current->p.c_p.y;
}

double
get_vertex_r(v)
VERTICES *v;
{
    assert(v->current);

    return v->current->p.p_p.r;
}

double
get_vertex_t(v)
VERTICES *v;
{
    assert(v->current);

    return v->current->p.p_p.t;
}

void
set_vertex_x(v, x)
VERTICES *v;
long x;
{
    assert(v->current);

    v->current->p.c_p.x = x;
}


void
set_vertex_y(v, y)
VERTICES *v;
long y;
{
    assert(v->current);

    v->current->p.c_p.y = y;
}

void
set_vertex_r(v, r)
VERTICES *v;
double r;
{
    assert(v->current);

    v->current->p.p_p.r = r;
}

void
set_vertex_t(v, t)
VERTICES *v;
double t;
{
    assert(v->current);

    v->current->p.p_p.t = t;
}

int
vertices_p(v)
VERTICES *v;
{
    return v->head ? TRUE : FALSE;
}

/*
  Return the POINT that is at the arithmetic mean of the vertices.
*/

void
arith_mean(v, am)
VERTICES *v;
C_POINT *am;
{
    reset_vertices(v);

    do {
	am->x += get_vertex_x(v);
	am->y += get_vertex_y(v);
    } while (next_vertex(v));

    am->x /= get_num_vertices(v);
    am->y /= get_num_vertices(v);
}

/*
  Translate the origin of the vertex points to the point given. This code
  assumes that the points are all in quadrent I and the origin is initially
  at 0,0.
*/

void
translate_origin(v, init, final)
VERTICES *v;
C_POINT *init;
C_POINT *final;
{
    long dx = init->x - final->x;
    long dy = init->y - final->y;

    reset_vertices(v);

    do {
	set_vertex_x(v, get_vertex_x(v) + dx);
	set_vertex_y(v, get_vertex_y(v) + dy);	
    } while (next_vertex(v));
}

/*
  Convert the points in the vertex list from cartesian coordinates to polar
  coordinates. Store R and theta in the x,y fields of the POINTs in the
  vertex list.
*/

void
cart_to_polar(v)
VERTICES *v;
{
    double x, y;
    reset_vertices(v);

    do {
	x = (double)get_vertex_x(v);
	y = (double)get_vertex_y(v);

	set_vertex_r(v, sqrt(pow(x, 2.0) + pow(y, 2.0)));
	set_vertex_t(v, atan2(y, x));
    } while (next_vertex(v));
}

void
polar_to_cart(v)
VERTICES *v;
{
    double r, t;
    
    reset_vertices(v);

    do {
	r = get_vertex_r(v);
	t = get_vertex_t(v);
	set_vertex_x(v, (long)((r + 0.05) * cos(t))); /* see the hack... */
	set_vertex_y(v, (long)((r + 0.05) * sin(t))); /* see the hack run. */
    } while (next_vertex(v));
}

void
sort_polar_vertices(v, compare)
VERTICES *v;
int (*compare)();
{
    long i = 0;
    P_POINT *table = (P_POINT *)XtCalloc(get_num_vertices(v), sizeof(P_POINT));

    reset_vertices(v);

    do {
	table[i].r = get_vertex_r(v);
	table[i++].t = get_vertex_t(v);
    } while (next_vertex(v));

    qsort((char *)table, i, sizeof(P_POINT), compare);

    i = 0;
    reset_vertices(v);

    do {
	set_vertex_r(v, table[i].r);
	set_vertex_t(v, table[i++].t);
    } while (next_vertex(v));

    XtFree((char *)table);
}

int
compare_theta(p1, p2)
P_POINT *p1;
P_POINT *p2;
{
    if (p1->t > p2->t)
	return 1;
    else if (p1->t < p2->t)
	return -1;
    else
	return 0;
}

/*
  Print out the vertices of a region. This function is intened for debugging.
  it does not return a value.
*/

void 
print_cart_vertices(vertices)
VERTICES *vertices;
{
    reset_vertices(vertices);

    do {
	err_print("print_cart_verteces: %d, %d",
		  get_vertex_x(vertices),
		  get_vertex_y(vertices));
    } while(next_vertex(vertices));
}

void 
print_polar_vertices(vertices)
VERTICES *vertices;
{
    reset_vertices(vertices);

    do {
	err_print("print_polar_verteces: %lf, %lf",
		  get_vertex_r(vertices),
		  get_vertex_t(vertices));
    } while(next_vertex(vertices));
}

#ifdef NEVER

long
main(void)
{
    long i;
    long n;
    long x, y;
    VERTICES *v = init_vertices();
    C_POINT init, final;
    
    if (!v)
	return EXIT_FAILURE;

    printf("Number of vertices: ");
    scanf("%d", &n);

    for (i = 0; i < n; i++) {
	printf("Enter vertex[%d]: ", i);
	scanf("%d %d", &x, &y);

	add_vertex(v, x, y);
    }

    print_cart_vertices(v);

    init.x = 0, init.y = 0;

#ifdef NEVER
    printf("Enter new origin: ");
    scanf("%d %d", &x, &y);

    final.x = x, final.y = y;
#endif

    arith_mean(v, &final);
    
    translate_origin(v, &init, &final);

    print_cart_vertices(v);

    cart_to_polar(v);

    print_polar_vertices(v);

    sort_polar_vertices(v, compare_theta);

    print_polar_vertices(v);
    
    polar_to_cart(v);
    
    print_cart_vertices(v);
    
    translate_origin(v, &final, &init);

    print_cart_vertices(v);
}
#endif

