/*
  Copyright 1991, 1992 University of Rhode Island
 
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose is hereby granted without fee, provided that
  the above copyright notice appear in all copies and that both that
  copyright notice and this permission notice appear in supporting
  documentation, and that the name of U.R.I. not be used in advertising or
  publicity pertaining to distribution of the software without specific,
  written prior permission.  U.R.I. makes no representations about the
  suitability of this software for any purpose.  It is provided "as is"
  without express or implied warranty.
 
  U.R.I. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.R.I.
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
  Create the widgets used by xbrowse. xbrowse connects to an image server
  over a tcp/ip network. Based on user input (handled with Xt callback
  functions) xbrowse makes various requests of the server (e.g., `Send the
  next image in the current directory', `restrict image transmission to the
  following region', ...). Image data is recieved via an action procedure
  which reads the data from a tcp/ip socket. 
  
  Modifications

  1/1/91
  Pixmap and image_gc are globals now.

  1/21/91
  Started using bsd compatibility mode of A/UX.

  1/31/91
  Fixed print_location (by writting ScreenGetPixel).

  2/7/91
  Screen GetPixel is no longer used since I now store the raw image data in
  an XImage structure.

  2/13/91
  Added display of cost to window.

  3/6/91
  Added start and end switches. When used xbrowse connects to a different
  server and requests the jukebox to move images from start date to end date
  onto magnetic media. The user is told how long the process will take and is
  notified of its completion via email.

  7/24/91
  Added three new widgets in place of the `old' coordinates widget. xbrowse
  now displays lat, lon and temp instead of pixel x,y and value.

  8/9/91
  Fixed overlay graphics (hash marks) so that it worked reasonably (rather
  than using the pixel value 255 it now uses a color name).

  8/21/91
  Added use region, etc. This allows a more flexible region setup.

  10/2/91
  Added auto zoom button.

  10/15/91
  Added fallback resources.
*/

/*
 * $Log: xbrowse.c,v $
 * Revision 3.5  1993/02/02  16:12:32  jimg
 * Added conditional include for debugging malloc header.
 *
 * Revision 3.4  1992/11/19  21:14:30  jimg
 * New lat --> string and lon --> string converters.
 * New lat/lon display.
 *
 * Revision 3.3  1992/11/03  19:43:35  jimg
 * Removed -start and -end.
 * Added -xbt_server (in usage()).
 *
 * Revision 3.2  1992/10/21  17:17:03  jimg
 * Added include file for dialogs.
 *
 * Revision 3.1  1992/10/14  13:19:00  jimg
 * Minor bug fixes.
 *
 * Revision 3.1  1992/10/14  13:19:00  jimg
 * Minor bug fixes.
 *
 * Revision 3.0  1992/07/28  15:24:44  jimg
 * All sources for xbrowse modified to use the libcomm.a and liberrmsg.a
 *
 * Revision 1.1  1992/03/23  20:39:01  jimg
 * Initial revision
 */

static char rcsid[]="$Id: xbrowse.c,v 3.5 1993/02/02 16:12:32 jimg Exp jimg $";

#include <stdio.h>
#include <math.h>

#include <sys/time.h>
#include <rpc/rpc.h>

#ifdef AUX
#include <compat.h>
#endif 

#include <X11/Intrinsic.h>	/* toolkit */
#include <X11/StringDefs.h>

#include <X11/Xaw/Command.h>	/* widget public includes */
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Box.h>

#ifdef MALLOC_DEBUG
#include <malloc.h>
#endif 

#define HASH_MARK_SIZE 4	/* distance from center to edge */
#define canonical_size(x) ((long) pow((double)2, floor(M_LOG2E * log((double) x))))
#define min(x, y) ((x < y) ? x : y)

#include "errmsg.h"
#include "comm.h"
#include "browse-list.h"
#include "browse-log.h"
#include "vertex.h"
#include "region.h"
#include "ximage.h"
#include "pyramid.h"
#include "pixmap-mark.h"
#include "xbt-const.h"
#include "xbt-types.h"
#include "browse.h"
#include "dialog.h"
#include "callbacks.h"
#include "xbrowse.h"
#include "xbrowse-resources.h"
#include "xbrowse-options.h"
#include "xbrowse-actions.h"
#include "make-pathname.h"

/* Globals */

/* from callbacks.c */
extern BROWSE *browse;		/* maintained by the state machine */

/*
  Widget and non-widget resources if the application defaults file can't be
  found.  Generated automatically from XBrowse.ad by "ad2c".  Comment out the
  include line (but not the NULL) if you don't want any resources compiled
  in.

  From George Ferguson's xarchie program. jhrg 10/15/91
*/

static String fallbackResources[] = {
#include "XBrowse.ad.h"
    NULL
};

char *pname;			/* program name */
XtAppContext a_context;		/* for portability */

APPDATA a_data;			/* application spec resources */
Pixmap window_pixmap;		/* holds image offscreen - server side */
Pixmap clip_mask;		/* clip mask so that hash marks show */
long window_pixmap_x;		/* size of offscreen pixmap (horiz) */
long window_pixmap_y;		/* size of offscreen pixmap (vert) */
long screen_num;		/* number of default screen */
long display_w;			/* width in pixels of default display */
long display_h;			/* height in pixels of default display */
long default_depth;		/* default depth of screen_num */
long static_visual;		/* FALSE if the def vis suppoerts mutable
				   colormaps  */
unsigned char hash_mark_color;	/* pixel value assoc with hashmarks */
unsigned char old_hash_mark_color; /* pixel value assoc with old hashmark */
unsigned char image_pal[256];	/* maps pixels in images to pix on screen */
Colormap colormap;		/* color map of default visual */
GC image_gc;			/* used to transfer window_pixmap to screen */
GC clip_gc;			/* used to update clip mask */
GC update_gc;			/* used when drawing into the viewport */
Display *display;		/* user's display */
Visual *visual;			/* default visual of user's window */
Window root_window;		/* user's root window */

Widget toplevel;		/* ancestor of all widgets */
Widget the_image;		/* holds image on screen */
Widget image_name;		/* label widgets to display ... */
Widget projection, satellite_type;
Widget max_res, current_res;
#ifdef NEVER
Widget latitude, longitude, temperature;
#endif
Widget cursor;			/* display data under cursor */
Widget cost;			/* display cost in window */

Widget pause_tx, save, next;	/* button widgets */
Widget log_entry, zoom, region_use;
Widget region_define, region_write;
Widget region_read, palette, quit;
Widget auto_zoom;
Widget xbt_display;
Widget auto_region;
Widget image_info, display_info, session_info;

int
main(argc, argv)
int argc;
char *argv[];
{
    Widget form;		/* holds menu, labels and viewport */
    Widget image_label_form;	/* holds image data labels */
    Widget display_label_form;
    Widget session_label_form;
    Widget button_form;		/* holds buttons */
    Widget image_view;		/* viewport that holds the image */
    
#ifdef AUX
    if (setcompat(getcompat() | COMPAT_BSD) == -1)
	err_sys("main: could not set 4.2BSD comapibility");
#endif

    pname = argv[0];		/* used by liberrmsg functions */

    toplevel = XtVaAppInitialize
	(&a_context, "XBrowse", options, XtNumber(options),
	 (Cardinal *)&argc, argv, fallbackResources, NULL, (Cardinal)0);

    /* Start out using the default colormap. This may be changed inside
       add_palette_to_colormap since that code will switch to a local color
       map if the default does not have space for all the colors specified
       in the user's default palette. */
    
    display = XtDisplay(toplevel);
    screen_num = DefaultScreen(display);
    root_window = RootWindow(display, screen_num);
    display_w = DisplayWidth(display, screen_num);
    display_h = DisplayHeight(display, screen_num);
    visual = DefaultVisual(display, screen_num);
    colormap = DefaultColormap(display, screen_num);
    default_depth = DefaultDepth(display, screen_num);

    if (!visual_ok_p())
	err_quit("X default visual not supported.");

    XtGetApplicationResources(toplevel, &a_data, resources,
			      XtNumber(resources), NULL, 0);

    if ((argc > 1) && !usage(argc, argv)) /* check for extraneous switches */
	exit(1);

    /* initialize the default palette (one to one data --> pix mapping) */
    {
    	unsigned int i;
	for (i = 0; i < 256; i++)
	    image_pal[i] = (unsigned char)i;
    }
    
    /* call do palette to install user's default colormap if one has been
       specified. This call MUST go after XtGetAppRes and BEFORE the xbrowse
       widgets are created. */
	
    if (a_data.palette[0] && !static_visual) {
	char *pal_name = make_pathname(a_data.palette_path, a_data.palette);
	FILE *fp = fopen(pal_name, "r");
	if (fp) {
	    add_palette_to_colormap(fp, image_pal, TRUE);
	    fclose(fp);
	}
	else
	    err_ret("Could not open palette file `%s'", a_data.palette);

	XtFree(pal_name);
    }
    
    if (a_data.image_size)
	window_pixmap_x = window_pixmap_y = a_data.image_size;
    else
	window_pixmap_x = window_pixmap_y = canonical_size(min(display_h,
							       display_w));

    if (a_data.debug)
	(void) XSynchronize(display, True);
    
    window_pixmap = XCreatePixmap(display, root_window,
				  window_pixmap_x, window_pixmap_y,
				  default_depth);

    clip_mask = XCreatePixmap(display, root_window,
			      window_pixmap_x, window_pixmap_y, 1);

    update_gc = XCreateGC(display, window_pixmap, 0, NULL);
    
    image_gc = XCreateGC(display, window_pixmap, 0, NULL);

    clip_gc = XCreateGC(display, clip_mask, 0, NULL);
    
    XSetForeground(display, clip_gc, 1);
    XSetBackground(display, clip_gc, 1);

    /* This sets the clip_mask pixmap - drawing hashmarks clears rects on
       this so that new image data does not overwrite the marks */
    XFillRectangle(display, clip_mask, clip_gc,
		   0, 0, window_pixmap_x, window_pixmap_y);

    XSetForeground(display, clip_gc, 0);
    XSetForeground(display, image_gc, WhitePixel(display, screen_num));
    XSetBackground(display, image_gc, WhitePixel(display, screen_num));
    XSetClipMask(display, image_gc, clip_mask);
    XSetClipOrigin(display, image_gc, 0, 0);

    /* given the fore and background colors, this clears the window pixmap */
    XFillRectangle(display, window_pixmap, image_gc,
		   0, 0, window_pixmap_x, window_pixmap_y);
    XSetForeground(display, image_gc, BlackPixel(display, screen_num));

    /* form widgets */
       
    form = XtVaCreateManagedWidget
	("Form", formWidgetClass, toplevel, XtNcolormap, colormap, NULL);

    image_label_form = XtVaCreateManagedWidget
	("ImageLabel", formWidgetClass, form, NULL);
    XtVaSetValues(image_label_form, XtNleft, XtChainLeft, XtNright,
		  XtChainLeft, XtNtop, XtChainTop, XtNbottom, XtChainTop,
		  NULL); 

    display_label_form = XtVaCreateManagedWidget
	("DisplayLabel", formWidgetClass, form, NULL);
    XtVaSetValues(display_label_form, XtNleft, XtChainLeft, XtNright,
		  XtChainLeft, XtNtop, XtChainTop, XtNbottom, XtChainTop,
		  NULL); 

    session_label_form = XtVaCreateManagedWidget
	("SessionLabel", formWidgetClass, form, NULL);
    XtVaSetValues(session_label_form, XtNleft, XtChainLeft, XtNright,
		  XtChainLeft, XtNtop, XtChainTop, XtNbottom, XtChainTop,
		  NULL); 

    button_form = XtVaCreateManagedWidget
	("Button", formWidgetClass, form, NULL);
    XtVaSetValues(button_form, XtNleft, XtChainLeft, XtNright, XtChainLeft,
		  XtNtop, XtChainTop, XtNbottom, XtChainTop, NULL);

    /* label widgets */
    
    image_info = XtVaCreateManagedWidget
	("ImageInfo", labelWidgetClass, image_label_form, NULL); 

    image_name = XtVaCreateManagedWidget
	("ImageName", labelWidgetClass, image_label_form, NULL); 

    projection = XtVaCreateManagedWidget
	("Projection", labelWidgetClass, image_label_form, NULL); 

    satellite_type = XtVaCreateManagedWidget
	("SatelliteType", labelWidgetClass, image_label_form, NULL); 

    max_res = XtVaCreateManagedWidget
	("MaxRes", labelWidgetClass, image_label_form, NULL); 

    display_info = XtVaCreateManagedWidget
	("DisplayInfo", labelWidgetClass, display_label_form, NULL); 

    current_res = XtVaCreateManagedWidget
	("CurrentRes", labelWidgetClass, display_label_form, NULL);
    
    cursor = XtVaCreateManagedWidget
	("Cursor", labelWidgetClass, display_label_form, NULL);
    
    session_info = XtVaCreateManagedWidget
	("SessionInfo", labelWidgetClass, session_label_form, NULL); 

    cost = XtVaCreateManagedWidget
	("Cost", labelWidgetClass, session_label_form, NULL);

    /* button widgets */

    quit = XtVaCreateManagedWidget
	("Quit", commandWidgetClass, button_form, NULL);
    XtAddCallback(quit, XtNcallback, quit_button, 0);

    pause_tx = XtVaCreateManagedWidget
	("Pause", commandWidgetClass, button_form, NULL);
    XtAddCallback(pause_tx, XtNcallback, pause_button, 0);
    
    next = XtVaCreateManagedWidget
	("Next", commandWidgetClass, button_form, NULL);
    XtAddCallback(next, XtNcallback, next_button, 0);

    log_entry = XtVaCreateManagedWidget
	("Log", commandWidgetClass, button_form, NULL);
    XtAddCallback(log_entry, XtNcallback, log_button, 0);

    save = XtVaCreateManagedWidget
	("Save", commandWidgetClass, button_form, NULL);
    XtAddCallback(save, XtNcallback, save_button, 0);

    palette = XtVaCreateManagedWidget
	("Palette", commandWidgetClass, button_form, NULL);
    XtAddCallback(palette, XtNcallback, palette_button, 0);

    region_use = XtVaCreateManagedWidget
	("UseRegion", commandWidgetClass, button_form, NULL);
    XtAddCallback(region_use, XtNcallback, use_region_button, 0);
    XtSetSensitive(region_use, False);

    region_define = XtVaCreateManagedWidget
	("DefineRegion", commandWidgetClass, button_form, NULL);
    XtAddCallback(region_define, XtNcallback, def_region_button, 0);
    
#ifdef NEVER
    region_write = XtVaCreateManagedWidget
	("WriteRegion", commandWidgetClass, button_form, NULL);
    XtAddCallback(region_write, XtNcallback, write_region_button, 0);
    XtSetSensitive(region_write, False);

    region_read = XtVaCreateManagedWidget
	("ReadRegion", commandWidgetClass, button_form, NULL);
    XtAddCallback(region_read, XtNcallback, read_region_button, 0);
#endif

    xbt_display = XtVaCreateManagedWidget
	("XBT", commandWidgetClass, button_form, NULL);
    XtAddCallback(xbt_display, XtNcallback, get_xbt_button, 0);

    auto_region = XtVaCreateManagedWidget
	("AutoRegion", commandWidgetClass, button_form, NULL);
    XtAddCallback(auto_region, XtNcallback, auto_region_button, 0);
    
    zoom = XtVaCreateManagedWidget
	("Zoom", commandWidgetClass, button_form, NULL);
    XtAddCallback(zoom, XtNcallback, zoom_button, 0);

    auto_zoom = XtVaCreateManagedWidget
	("AutoZoom", commandWidgetClass, button_form, NULL);
    XtAddCallback(auto_zoom, XtNcallback, auto_zoom_button, 0);

    image_view = XtVaCreateManagedWidget
	("ImageView", viewportWidgetClass, form,
	 XtNheight, window_pixmap_y,
	 XtNwidth, window_pixmap_x,
	 XtNleft, XtChainLeft,
	 XtNtop, XtChainTop, NULL);
    
    the_image = XtVaCreateManagedWidget
	("TheImage", widgetClass, image_view,
	 XtNheight, window_pixmap_y,
	 XtNwidth, window_pixmap_x, NULL);

    XtAppAddActions(a_context, image_actions, XtNumber(image_actions));

    /* This allows the event loop to run before calling init_state_machine */
    XtAppAddWorkProc(a_context, init_state, toplevel);

    XtRealizeWidget(toplevel);

    XtAppMainLoop(a_context);

    return TRUE;
}

/*
  Scan the argv array for any options (preceded by '-'). X should have
  taken care of all the options, so any remaining ones must be
  extraneous. Note that words not prefixed by '-' won't signal an
  error. This will allow file names to be supplied on the command line
  without printing error messages.

  Return TRUE if no options remain in the array, FALSE (illegal
  options specified) otherwise. 
*/

long 
usage(argc, argv)
long argc;
char *argv[];
{
    long i;
    long ill_opt = FALSE;

    for (i = 1; i < argc; i++) { /* i=1 --> skip program name */
	if (argv[i][0] == '-') {
	    long j;
	    long opt_found = FALSE;
	    
	    for (j = 0; j < XtNumber(options); j++)
		if (strcmp(argv[i], options[j].option) == 0) {
		    fprintf(stderr, "Option: %s specified incorrectly\n",
			    argv[i]);
		    opt_found = TRUE;
		}

	    if (!opt_found)
		fprintf(stderr, "Option: %s unrecognized\n", argv[i]);

	    ill_opt = TRUE;
	}
    }

    if (ill_opt) {
	fprintf(stderr, "\n\
\n\
In addition to the standard X and X Toolkit options,\n\
xbrowse accepts the following:\n\
\n\
   -help   	    	Print this message and exit\n\
   -debug	    	Synchronize X client and server\n\
   -statistics          Print information about transmission times, ...\n\
   -no_question_exit     Do not ask `Really Quit?' when Quit is chosen\n\
   -auto_zoom           Always zoom in on the user defined region\n\
   -sort_vert           Order region vertices clockwise\n\
\n\
   -region_path `path'  Prepend `path' to relative filenames for regions\n\
   -log_path `path'     Prepend `path' to relative filenames for log files\n\
   -save_path `path'    Prepend `path' to relative filenames for images\n\
\n\
   -palette `file'      Read color information from `file'\n\
   -server `host'    	Connect to a browse image server on `host'\n\
   -xbt_server `host'   Connect to an xbt meta data server on `host'\n\
   -list `file'  	List of images to browse (full pathnames)\n\
   -path `path'	    	Path (on the archive host) of images to browse\n\
   -log `file' 	    	Name of logfile\n\
\n\
   -image_size `num'    Size (pixels) of image display widget\n\
   -level `num'	    	Initial level to display\n\
   -hashmark `color'    Color to use for region vertex hashmarks\n\
   -old_hashmark `color' Color to use for old region's hashmarks\n\
\n\
   -xrm `resource'     specify an arbitrary X resource/value\n\
\n\
Note: See the man page for information.\n\
\n\
");
	return(FALSE);
    }

    return(TRUE);
}
	    
/*
  Given that an expose event has been generated for the visible
  window, update (copy) the pixmap containing the image to the window

  NB: use update_gc to copy things like the grid, ... that are masked by
  image_gc's clipmask. 
*/

void
redraw_picture(w, e)
Widget w;
XExposeEvent *e;
{
    if (e) {
	register long x = e->x;
        register long y = e->y; 
	register long height = e->height;
	register long width = e->width;

	XCopyArea(display, window_pixmap, XtWindow(w), update_gc,
		  x, y, width, height, x, y);
    } 
    else {
	XCopyArea(display, window_pixmap, XtWindow(w), update_gc,
		  0, 0, window_pixmap_x, window_pixmap_y, 0, 0);
    }
}

/*
  When the mouse is inside the image display window and the mouse
  button is pressed, display the coordinates of the corresponding pixel.
  This must handle a zoomed image. When the image size and window size are
  equal (e.g. 512), then printing the lon, lat is simple. However, if the
  image is 2048 and the window is 512, then first map the x,y coordinates
  from the event record to image coordinates, then map image coordinates to
  lon, lat.
*/

void 
print_location(w, e)
Widget w;
XButtonEvent *e;
{
    static char coor_str[31];	/* text to be displayed */
    double temp;		/* value to display */
    double lon, lat;		/* locations to display */
    PYRAMID *pyramid = browse->pyramid;

    /* Don't try to map points within the image boundry */
    if (e->x < 0 || e->x > window_pixmap_x - 1
	|| e->y < 0 || e->y > window_pixmap_y - 1)
	return;

    if (pyramid->level == 0) {
	XBell(display, 50);
	return;
    }

    temp = get_temperature(pyramid, e->x, e->y);
    get_lat_lon(pyramid, e->x, e->y, &lat, &lon);
    
    sprintf(coor_str, "Cursor: %s, %s = %4.2f%s", lat2str(lat), lon2str(lon),
	    temp, pyramid->header->data_units); 

    XtVaSetValues(cursor, XtNlabel, coor_str, NULL);
}

/*
  Return TRUE if the X server can support the color facilities we need, FALSE
  otherwise.

  Note: This function isn't really necessary, since all we care about is a
  depth of 8 or more (color or grey - it doesn't matter). However, if at some
  time we want to add support for servers whose default visual is not
  sufficient for browse, but that does in fact have 8 bit visuals, then this
  function is a good starting point.
*/

int
visual_ok_p()
{
    int i;
    int visuals_matched;
    XVisualInfo v_template;	/* template for visual we want */
    XVisualInfo *visual_list;	/* list of visuals that match */

    v_template.screen = screen_num;

    visual_list = XGetVisualInfo(display, VisualScreenMask, &v_template,
				 &visuals_matched); 

    /* if no visuals matched, signal massive error */
    if (visuals_matched == 0) {
	XFree((char *)visual_list);	
	return FALSE;
    }

    /* get the XVisualInfo structure for the default visual */
    for (i = 0; i < visuals_matched; i++) 
	if (visual == visual_list[i].visual)
	    break;		/* i == default */

    /* id none of the visuals matched is the default visual, siganl error */
    if (i == visuals_matched) {
	XFree((char *)visual_list);
	return FALSE;
    }

    /* return TRUE if the visual is supported and the pixel depth is 8 bits
       or greater. Set the static_visual global */

    if (visual_list[i].depth < 8) {
	XFree((char *)visual_list);	
	return FALSE;
    }

    switch(visual_list[i].class) {
/*    case StaticGray: */
/*    case StaticColor: */
    case TrueColor:
	static_visual = TRUE;
	XFree((char *)visual_list);	
	return TRUE;

/*    case GrayScale: */
    case PseudoColor:
    case DirectColor:
	static_visual = FALSE;
	XFree((char *)visual_list);	
	return TRUE;

    default:
	XFree((char *)visual_list);	
	return FALSE;
    }
}

void
add_palette_to_colormap(palette_fp, pal, new_cm_p)
FILE *palette_fp;
unsigned char pal[];
long new_cm_p;
{
    long value = 0;
    long i = 0;
    char color_name[256];
    char line[256];
    
    while (fgets(line, 256, palette_fp)) {
	i++;
	if (sscanf(line, "%d %s\n", &value, color_name) != 2) {
	    err_print("Line %d of palette file is malformed\n\
 - format <value> <color name>", i);
	    continue;
	}
	if (value < 0 || value > PAL_MAX - 1) {
	    err_print("Palette error : Pixel `%d' on line %d is out of range\n\
 - maximum pixel value is %d", value, i, PAL_MAX - 1);
	    continue;
	}
	if (!alloc_color(color_name, &pal[value]))
	    if (new_cm_p) {
		err_print("Too many colors for default colormap - using local colormap");
		colormap = XCopyColormapAndFree(display, colormap);
		if (!alloc_color(color_name, &pal[value]))
		    err_print("Local colormap overflow - redesign palette");
	    }
	    else
		err_print("Could not allocate color `%s'", color_name);
    }
}

/*
  Given a color name (a String), return the pixel value associated with that
  color if a color cell was/is allocated to it. Otherwise, return FALSE.

  9/9/91
  Print an error message and return TRUE when an invalid color name
  (one not in the color database) is given. This allows calling programs to
  detect when the colormap is full and a new colormap needs to be allocated.

  Pass pixel value back by reference to avoid confusion over TRUE/FALSE and
  pixels 0 and 1.
*/

long
alloc_color(color_name, pixel)
char *color_name;
unsigned char *pixel;
{
    XColor exact_def;

    if (!XParseColor(display, colormap, color_name, &exact_def)) {
	err_print("Color `%s' not in database - using Black instead", color_name);
	*pixel = BlackPixel(display, screen_num);
	return TRUE;
    }

    if (!XAllocColor(display, colormap, &exact_def))
	return FALSE;

    *pixel = (unsigned char)exact_def.pixel;

    return TRUE;
}

/*
  Return a pointer to static memory which contains the string representation
  of the longitude vaule `lon'. '\260' is typically the degree symbol.
*/

char *
lon2str(lon)
double lon;
{
    static char lon_str[8];
    
    sprintf(lon_str, "%5.1f%c%c", fabs(lon), '\260', (lon > 0.0) ? 'E' : 'W');

    return lon_str;
}
    
char *
lat2str(lat)
double lat;
{
    static char lat_str[8];
    
    sprintf(lat_str, "%4.1f%c%c", fabs(lat), '\260', (lat > 0.0) ? 'N' : 'S');

    return lat_str;
}
