/*
  Connect with the xbt server and get locations of xbt drops. Since there is
  a fair amount of overhead associated with initiating a connection with a
  server, establishing a connection will be delayed until the user actually
  requests some xbt data. Once made the connection will be maintained for the
  duration of the user's session. At the end of the session, the connection
  will be closed.

  2/3/92 jhrg
*/

/*
 * $Log: xbt.c,v $
 * Revision 3.3  1993/02/02  16:12:32  jimg
 * Modified so that structures could easily be used - works with new xbt server.
 *
 * Revision 3.2  1992/10/21  17:17:03  jimg
 * No longer uses libcomm's record marking feature.
 *
 * Revision 3.2  1992/10/21  17:17:03  jimg
 * No longer uses libcomm's record marking feature.
 *
 * Revision 3.1  1992/10/14  13:19:00  jimg
 * Modified for use with the new xbt server.
 *
 * Revision 3.0  1992/07/28  15:24:44  jimg
 * All sources for xbrowse modified to use the libcomm.a and liberrmsg.s
 * a
 *
 * Revision 1.1  1992/03/23  20:39:01  jimg
 * Initial revision
 */

static char rcsid[]="$Id: xbt.c,v 3.3 1993/02/02 16:12:32 jimg Exp jimg $";

#include <stdio.h>

#include <rpc/rpc.h>

#include <X11/X.h>
#include <X11/Intrinsic.h>

#ifdef MALLOC_DEBUG
#include <malloc.h>
#endif 

#include "errmsg.h"
#include "comm.h"
#include "vertex.h"
#include "region.h"
#include "ximage.h"
#include "pyramid.h"
#include "pixmap-mark.h"
#include "xbt-const.h"
#include "xbt-types.h"
#include "xbt-funcs.h"
#include "xbt-codes.h"
#include "dialog.h"
#include "xbrowse.h"

#define XBT_SERVER_PORT 7229
#define XBT_SERVER_VERSION 2
#define CLIENT_VERSION 2

extern APPDATA a_data;

/*
  Establish a connection with an xbt data server. All xbt servers should use
  port 7229. Returns the socket descriptor if the connection was established
  OK, otherwise returns NULL.
*/

XDR_TCP *
xbt_connect(host)
char *host;
{
    int sd = client_connect(host, NULL, XBT_SERVER_PORT, sys_err_dialog);
    int status;
    long version;
    XDR_TCP *xbt_server;

    while (sd == -1) {
	char *temp_server = get_x11_string("Server machine name:",
					   "uri.gso.uri.edu");
	if (!temp_server)
	    return NULL;	/* user hits cancel */
	a_data.xbt_server = XtNewString(temp_server);
	sd = client_connect(a_data.xbt_server, NULL, XBT_SERVER_PORT,
			    sys_err_dialog); 
    }

    xbt_server = xdr_tcp_bind(sd, 4096, sys_err_dialog);

    status = send_long(xbt_server, CLIENT_VERSION, TRUE);
    if (!status) {
	err_dialog("Could not connect with xbt server on `%s'", host);
	return NULL;
    }
    
    status = get_long(xbt_server, &version, TRUE);
    if (!status || (version != XBT_SERVER_VERSION)) {
	err_dialog("Could not connect with xbt server on `%s'", host);
	return NULL;
    }

    /* register structs used for messages */
    
    xdr_tcp_set_struct(XBT_BOX, sizeof(xbt_box_t), xdr_xbt_box);
    xdr_tcp_set_struct(XBT_DROP, sizeof(xbt_drop_t), xdr_xbt_drop);
    xdr_tcp_set_array(XBT_DROP_ARRAY, MAX_XBT_DROPS, sizeof(xbt_drop_t),
		      xdr_xbt_drop);

    return xbt_server;
}

/*
  Send time and place info to the xbt server and get back an array of lon.lat
  location marking xbt drop points. Returns FALSE if an error was detected,
  oterwise returns the number of xbt drops.

  sd is the socket that connects this process to the xbt server.
  box contains the top-left and bottom-right corners of the area of interest,
  (lon,lat), (lon,lat).
  time is a long specifying yydddhhmm.
  time_window is a long specifying a window around time to search.
  xbt-drops is the array in which locations are returned to the caller. It
  must be large enough to hold the information returned by the server
*/

int
get_xbt_drops(xsd, box, time, time_window, xbt_drops)
XDR_TCP *xsd;
float box[4];
long time;
long time_window;
xbt_drop_t xbt_drops[];		/* return value */
{
    long num_hits;
    xbt_box_t xbt_box;
    
    send_action(xsd, XBT_REQUEST_SB, TRUE);
    
    xbt_box.box[0] = box[0];
    xbt_box.box[1] = box[1];
    xbt_box.box[2] = box[2];
    xbt_box.box[3] = box[3];

    xbt_box.imgdy = time / 10000;
    xbt_box.imgtm = time - xbt_box.imgdy * 10000;
    
    xbt_box.window = time_window;
    xbt_box.points_p = TRUE;

    send_struct(xsd, (caddr_t)&xbt_box, XBT_BOX, TRUE);

    get_long(xsd, &num_hits, TRUE);

    if (xbt_box.points_p && num_hits){
	u_int length, type;
	
	get_array(xsd, (caddr_t)&xbt_drops[0], &length, &type, TRUE);
    }

#ifdef NEVER
    xdr_tcp_flush_record(xsd);
#endif
    
    return num_hits;
}

/*
  Close a connection to an xbt server.
*/

void
xbt_close(xsd)
XDR_TCP *xsd;
{
    send_action(xsd, EXIT, TRUE);

    xdr_tcp_close(xsd);
}

/*
  Plot the xbt data on window_pixmap. This function belongs in with the as
  yet unrealized in_situ data type.
*/

void
draw_xbt_drops(num_drops, xbt_drops, pyramid, xbt_mark)
int num_drops;
xbt_drop_t xbt_drops[];
PYRAMID *pyramid;
PIXMAP_MARK *xbt_mark;
{
    int i;
    
    for (i = 0; i < num_drops; i++) {
	int x, y;
	
	get_x_y(pyramid, xbt_drops[i].lon, xbt_drops[i].lat, &x, &y);
#ifdef NEVER
	fprintf(stderr, "Drop site #%d (%f, %f) (%d, %d)\n", i + 1,
		xbt_drops[i].lon, xbt_drops[i].lat, x, y);
#endif
	draw_pixmap_mark(xbt_mark, x, y);
    }
}

/*
  Erase the xbt drops. Really belongs in the in_situ type.
*/

void
erase_xbt_drops(num_drops, xbt_drops, pyramid, xbt_mark)
int num_drops;
xbt_drop_t xbt_drops[];
PYRAMID *pyramid;
PIXMAP_MARK *xbt_mark;
{
    int i;
    
    for (i = 0; i < num_drops; i++) {
	int x, y;
	
	get_x_y(pyramid, xbt_drops[i].lon, xbt_drops[i].lat, &x, &y);
#ifdef NEVER
	fprintf(stderr, "Drop site #%d (%f, %f) (%d, %d)\n", i + 1,
		xbt_drops[i].lon, xbt_drops[i],lat, x, y);
#endif
	erase_pixmap_mark(xbt_mark, x, y);
    }
}
