/*
  Copyright 1991 University of Rhode Island
 
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose is hereby granted without fee, provided that
  the above copyright notice appear in all copies and that both that
  copyright notice and this permission notice appear in supporting
  documentation, and that the name of U.R.I. not be used in advertising or
  publicity pertaining to distribution of the software without specific,
  written prior permission.  U.R.I. makes no representations about the
  suitability of this software for any purpose.  It is provided "as is"
  without express or implied warranty.
 
  U.R.I. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.R.I.
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
  XImage structs are used to store image information as it is received from
  the image server. Once in the XImage struct, data is transfered to the
  screen using the XPutImage function. The functions in this file were ported
  from the VMS version of browse which used a struct called ScreenStruct to
  store image data.

  Modifications

  12/11/90 Ported to UNIX/X Window System. Images are now displayed in
  a window rather than the Dsp display.  Instead of writing image
  information to the Dsp display, that information will be written to an
  X Pixmap which in turn will be displayed in a window. ScreenStruct
  will read an store information in an XImage structure and use
  XPutImage to transfer the contents of that structure to the screen.

  A single XImage is allocated with `data' pointing to a size^2
  char*'s. As each row of the image is transmitted from the image
  server it is added to the `matrix' using memcpy. When it is
  received, the data for each row is expanded so that it is size bytes
  in length. Rather than copy each row, XPutImage is called with a range of
  values for dest_y. Thus, the Pixmap gets the multiple lines w/o the overhead
  of 512 memcpy calls. Note that this means that there are more XPutImage
  calls. Part of the reason this `optimization' is used is to ease the
  transition from dsp browse to xbrowse.

  1/21/91
  Added a function to return the value at (x,y). The point comes from a
  display of the current level of the pryamid on the user's screen and is
  then mapped to the (row,col) that contain the actual data being displayed.
  Called ScreenGetPixel.

  2/14/91
  Changed these functions to operate on XImages rather and the now defunct
  ScreenStruct. ScreenStruct was a holdover from the days before X.

  4/17/91
  Added zoom_ximage which copies a square within one XImage to another
  XImage. It is used by the code that zooms in on a region.

  9/3/91
  Modified error conditions - use dialog boxes and ensure that erros are only
  reported once.
*/

/*
 *$Log: ximage.c,v $
 * Revision 3.3  1993/02/02  16:12:32  jimg
 * Added conditional include for debugging malloc header.
 *
 * Revision 3.2  1992/11/19  21:14:30  jimg
 * ?
 *
 * Revision 3.1  1992/10/19  14:57:32  jimg
 * Minor fixes.
 *
 * Revision 3.1  1992/10/19  14:57:32  jimg
 * Minor fixes.
 *
 * Revision 3.0  1992/07/28  15:24:44  jimg
 * All sources for xbrowse modified to use the libcomm.a and liberrmsg.a
 *
 * Revision 1.1  1992/03/23  20:39:01  jimg
 * Initial revision
 */

static char rcsid[]="$Id: ximage.c,v 3.3 1993/02/02 16:12:32 jimg Exp jimg $";

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include <X11/X.h>
#include <X11/Intrinsic.h>

#ifdef MALLOC_DEBUG
#include <malloc.h>
#endif 

#include "dialog.h"
#include "errmsg.h"

#define MAX_IMAGE_DIM 1024	/* used by assert */

/* Globals */

/* xbrowse.c */
extern Pixmap window_pixmap;
extern GC image_gc;
extern Display *display;
extern Visual *visual;

/*
  Create an XImage height by width pixels in size. Each pixel is stored in
  a byte, and there is no padding between scan lines. The XImage format is
  ZPixmap. Use XDestroyImage to release the memory used by this structure.
*/

XImage *
init_ximage(height, width)
long height;
long width;
{
    XImage *image;
    char *data;

    assert(height < (MAX_IMAGE_DIM + 1));
    assert(width < (MAX_IMAGE_DIM + 1));

    data = (char *)XtMalloc(height * width * sizeof(char));

#ifdef NEVER
    assert(data != NULL);
    if (!data) {
	sys_err_dialog("Could not allocate memory for\n\
new (or temporary) XImage");
	return NULL;
    }
#endif
    
    image = XCreateImage
	(display,		/* the display */
	 visual,		/* default visual */
	 8,			/* bit depth */
	 ZPixmap,		/* image format */
	 0,			/* bit offset */
	 data,			/* image data */
	 height,		/* image height */
	 width,			/* image width */
	 8,			/* scan line quantum */
	 width);		/* bytes per scan line */

    if (!image) {
	sys_err_dialog("Could not create new (or temporary) XImage");
	return NULL;
    }
    
    return image;
}

/*
  Given an image and zoom_factor, create a second image that contains a
  magnified portion of IMAGE. The part of IMAGE to be magnified is the square
  that starts at x_origin, y_origin and ends at (image->width / zoom_factor +
  x_origin), (image->height / zoom_factor + y_origin). The resultin XImage is
  returned.
*/

XImage *
zoom_ximage(image, zoom_factor, x_origin, y_origin)
XImage *image;
long zoom_factor;
long x_origin, y_origin;
{
    register long i, l, x_br = image->width / zoom_factor + x_origin;
    register long j, m, y_br = image->height / zoom_factor + y_origin;
    register long k;
    XImage *dest_image;
    register char *src;
    register char *dest;
    register char *dest_row;
    register char *src_row;

    dest_image = init_ximage(image->height, image->width);
    if (!dest_image)
	return NULL;

    src = image->data;
    dest = dest_image->data;
    
    /* optimize this loop */
    
    for (j = y_origin, m = 0; j < y_br; j++, m += zoom_factor) {
	
#if (defined(ultrix) || defined(__ultrix)) && (defined(mips) || defined(__mips))
	assert(((dest_image->width * m) + dest_image->height) \
	       <= (dest_image->height * dest_image->width));
	assert(((image->width * j) + image->height) \
	       <= (image->width * image->height));
#else
	assert(((dest_image->width * m) + dest_image->height)
	       <= (dest_image->height * dest_image->width));
	assert(((image->width * j) + image->height)
	       <= (image->width * image->height));
#endif
	dest_row = dest + dest_image->width * m;
	src_row = src + image->width * j;
	
	/* copy first row in block */
	
	for (i = x_origin, l = 0; i < x_br; i++, l += zoom_factor) {
	    for (k = 0; k < zoom_factor; k++) {
		assert((l + k) < dest_image->width);
		assert(i < image->width);
		*(dest_row + l + k) = *(src_row + i);
	    }
	}
	/* copy first row to remaining rows in block */
	
	for (k = 1; k < zoom_factor; k++) {
#if (defined(ultrix) || defined(__ultrix)) && (defined(mips) || defined(__mips))
	    assert(((dest_image->width * k) + dest_image->width) \
		   <= (dest_image->width * dest_image->height));
#else
	    assert(((dest_image->width * k) + dest_image->width)
		   <= (dest_image->width * dest_image->height));
#endif	    

	    memcpy((dest_row + dest_image->width * k), (dest_row),
		   dest_image->width);
	}
    }

    return dest_image;
}

#if !(defined(MACROS) && defined(NDEBUG))

/*
  Return a pointer to the start of row in image.
*/

char *
get_ximage_row(image, row)
XImage *image;
long row;
{
    assert(row < image->height);
    return (char *)image->data + (row * image->width);
}

/*
  Copy image->width pixels (bytes) starting at row src to row dest. This may
  not work if src and dest are the same (since memcpy may not work when
  copying overlapping memory).
*/

void
copy_ximage_row(image, dest, src)
XImage *image;
long dest;
long src;
{
    register long size = image->width;
    char *data = image->data;

    assert(src * size + size <= image->height * image->width);
    assert(dest * size + size <= image->height * image->width);
    assert(src != dest);
    (void) memcpy(data + (dest * size), data + (src * size), size);
}

/*
  Transfers the named row of the ximage to the pixmap which is displayed in
  the user's window.
*/

void
display_ximage_row(image, row)
XImage *image;
long row;
{
    assert(row < image->height);
    XPutImage(display, window_pixmap, image_gc, image,
	      0, row, 0, row, image->width, 1);
}

/*
  Clear the XImage data memory.
*/

void 
clear_ximage(image)
XImage *image;
{
    (void) memset(image->data, 0, image->height * image->width);
}

#endif /* ! (MACROS && NDEBUG) */
