/*
  Copyright 1990, 1991, 1992 University of Rhode Island
 
  Permission to use, copy, modify and distribute this software and its
  documentation for any purpose is hereby granted without fee, provided that
  the above copyright notice appear in all copies and that both that
  copyright notice and this permission notice appear in supporting
  documentation, and that the name of U.R.I. not be used in advertising or
  publicity pertaining to distribution of the software without specific,
  written prior permission.  U.R.I. makes no representations about the
  suitability of this software for any purpose.  It is provided "as is"
  without express or implied warranty.
 
  U.R.I. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.R.I.
  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
  OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS ACTION, ARISING OUT OF OR IN 
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
  Communications functions. These functions implement a simple communications
  system that can be used to transmit data between processes. Once two
  processes are connected with one or more sockets, they can use these
  functions to exchange data. The functions provide a mechanism to transmit:

  1) actions (single byte flags),

  2) integers (four bytes),

  3) doubles (eight bytes),

  4) strings (as in C),

  5) structs (length determined at compile time)

  6) arrays (variable length)

  7) opaque blocks of memory (variable length)

  There are seven functions that handle sending these types of data (e.g.,
  send_action, send_float, ...) and seven functions that handle reading the
  information from the socket (e.g., get_float, ...).

  When the send functions are used to transmit data, they prefix that actual
  data to be sent with a `type code' that the receiving process can use to
  determine what type of data is contained in the pipe. A second set of read
  functions exist to exploit this feature. Using these functions
  (read_type_code, read_integer, ...) the process can examine the type code
  of the next block of data and then call the appropriate function to extract
  that type data.

  In order to send structs and arrays, several additional functions must be
  defined by the user's application. For structs, a function returning the
  length of the structure and a function that transforms the struct's local
  representation to the XDR representation must be defined. For an array, a
  length function, an element length function and an element encoding
  function must all be provided. These functions are simple to write - see
  the XDR documentation.

  In addition to the encoding and length functions, each struct and array
  must be identified with a unique integer. The struct length and encoding
  functions can then be accessed by calling struct_length(<struct type code>)
  and the struct encoding function can be assessed by calling
  struct_encoder(<struct type code>).

  Since this library uses xdr, programs must link so that those functions are
  available. On unix this is generally no problem - all unix kernals that
  support NFS must have xdr. On VMS, the xdr library must be explicitly built
  (I have the sources). For this file to compile, the constant `NO_XDR' must
  be #define'd so that the #includes will all work.

  Lastly, this library uses XDR's record marking features. Each call to a
  send_* function can cause all the data transmitted as a result of that call
  to be marked as a separate record or to be buffered and grouped with other
  data. In the later case, once the buffer fills its contents are written to
  the socket. The function xdr_tcp_flush_record() can be used to remove the
  remainder of the current record from the socket and set the mark at the
  begining to the next record.
*/

/*
 * Change log:
 * $Log: comm.c,v $
 * Revision 2.3  1992/11/10  17:44:22  jimg
 * Fixed prototype bugs.
 * Fixed compiler symbols (unix, ...) so that they work with gcc -ansi
 *
 * Revision 2.2  1992/11/02  17:49:16  jimg
 * Fixed problems with error_handler prototypes.
 *
 * Revision 2.1  1992/10/30  18:35:34  jimg
 * Removed arg list style formal parameter declarations since they broke
 * old compilers.
 *
 * Revision 2.0  1992/09/28  14:02:33  jimg
 * Release version.
 *
 * Revision 1.7  1992/09/23  19:42:10  jimg
 * Added {send,get,read}_float() functions.
 *
 * Revision 1.6  1992/09/23  19:33:59  jimg
 * Fixed bug on VMS: Client disconnects whould cause the server to go into an
 * infinite loop.
 *
 * Revision 1.5  1992/08/31  18:50:24  jimg
 * Fixed bug in xdr_tcp_close where xdr_free() was erroroniously called -
 * now XDR_DESTROY() is called.
 *
 * Revision 1.4  1992/07/30  21:14:39  jimg
 * Fixed bug where VMS Channel descriptor was stored in an
 * int rather than a short.
 *
 * Revision 1.3  1992/05/11  19:33:59  jimg
 * Merged patches from vms port.
 * On vms xdr_tcp_read_poll (and ...write_poll probably) does not work.
 *
 * Revision 1.2  1992/04/16  19:05:45  jimg
 * Switched from direct interface with sockets to XDR --> sockets.
 * Added facilities for the transmission of arrays and structs.
 * Added record-based transmission capabilities.
 *
 * Revision 1.1  1992/03/10  16:12:04  jimg
 * Initial revision
 */

static char rcsid[]="$Id: comm.c,v 2.3 1992/11/10 17:44:22 jimg Exp jimg $";

#if !(defined(unix) || defined(__unix)) && !defined(vms)
error "Either vms or unix must be defined"
#elif (defined(unix) || defined(__unix)) && defined(vms)
error "Only one of vms or unix can be defined"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#if defined(unix) || defined(__unix)
#include <unistd.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <netinet/in.h>
#ifdef AIX
#include  <sys/select.h>
#endif /* AIX */
#ifdef NO_XDR
#include "xdr_types.h"
#include "xdr.h"
#else
#include <rpc/rpc.h>
#endif /* NO_XDR */
#else
#ifdef vms
#include <descrip>
#include <ssdef>
#include <iodef>
#include "multinet_root:[multinet.include.sys]types.h"
#include "multinet_root:[multinet.include.sys]socket.h"
#include "multinet_root:[multinet.include.sys]time.h"
#include "multinet_root:[multinet.include.vms]inetiodef.h"
#include "multinet_root:[multinet.include.netinet]in.h"
#include "multinet_root:[multinet.include]netdb.h"
#include "xdr_types.h"
#include "xdr.h"
#else
error "Must define either vms *or* unix"
#endif /* vms */
#endif /* unix */

#include "comm.h"
#include "errmsg.h"

#ifndef TRUE
#define FALSE (0)
#define TRUE (1)
#endif

#define SET_XDR_ENCODE(a) a->x_op = XDR_ENCODE
#define SET_XDR_DECODE(a) a->x_op = XDR_DECODE
#define XDR_TCP_ERROR (*xdr_tcp->err_handler)

#define MAX_STRUCT_TYPES 256
#define MAX_ARRAY_TYPES 256

/*
  These macros might be defined as functions to aid in debugging.
*/

#define STRUCT_SIZE(t) (s_sizes[t])
#define STRUCT_CODER(t) (s_coders[t])

#define ARRAY_MAXSIZE(t) (a_sizes[t])
#define ARRAY_ELEM_SIZE(t) (a_elem_sizes[t])
#define ARRAY_CODER(t) (a_coders[t])
   
typedef int (*FUNC_RET_INT) ();

/*
  These static arrays hold the size constants and the encoding/decoding
  functions for the arrays and structs that are known to the system. The
  arrays are initialized at runtime via the functions xdr_tcp_set_array and
  xdr_tcp_set_struct and depend on the user defining the coders and
  identifying constants for the structs and arrays. These are should not be
  accessed by any other functions
*/

static int (*s_coders[MAX_STRUCT_TYPES])();
static u_int s_sizes[MAX_STRUCT_TYPES];

static int (*a_coders[MAX_ARRAY_TYPES])();
static u_int a_sizes[MAX_ARRAY_TYPES];
static u_int a_elem_sizes[MAX_ARRAY_TYPES];

/*
  This static global is used (on vms only) to hold the number of VMS
  channels. This value is determined from getdtablesize() and used in the
  select calls in xdr_tcp_{read,write}_poll(). Hack.
*/

#ifdef vms
static int vms_channels;
#undef FD_SETSIZE
#define FD_SETSIZE vms_channels
#endif

/*
  tcp_write and tcp_read handle TCP/IP socket i/o for the record-based XDR.
  They either write/read nbytes from/to the buffer `data' and return then
  number of bytes written/read or -1 if an error is detected. In this case,
  the error code will be left in errno.
*/

static int
tcp_write(iohandle, data, nbytes)
caddr_t iohandle;		/* for unix, ptr to socket descriptor */
char *data;
int nbytes;
{
    register int status;
    
#if defined(unix) || defined(__unix)
    
 write_data:
    status = write(*(int *)iohandle, data, nbytes);

    if (status > 0 && status < nbytes) {
	data += status;
	nbytes -= status;
	goto write_data;
    }

    if (status < 0)
	err_ret("status < 0");
    
    return status;

#elif defined(vms)

    return socket_write(*(short *)iohandle, data, nbytes);

#endif    
}

static int
tcp_read(iohandle, data, nbytes)
caddr_t iohandle;
char *data;
int nbytes;
{
    register int status;

#if defined(unix) || defined(__unix)

    status = read(*(int *)iohandle, data, nbytes);
    return (status == 0) ? -1 : status;
    
#elif defined(vms)

    status = socket_read(*(short *)iohandle, data, nbytes);
    return (status == 0) ? -1 : status;

#endif    
}

/*
  xdr_tcp_bind binds a TCP/IP socket descriptor to an instance of the
  record-based XDR stream. xdr_tcp_flush_record causes the XDR buffer to be
  written to the socket.
*/

#if defined(unix) || defined(__unix)
XDR_TCP *
xdr_tcp_bind(sd, buff_size, error_handler)
int sd;
int buff_size;
#ifdef __STDC__
void (*error_handler)(const char *fmt, ...);
#else
void (*error_handler)();
#endif /* __STDC__ */
#elif defined(vms)
XDR_TCP *
xdr_tcp_bind(sd, buff_size, error_handler)
short sd;
int buff_size;
void (*error_handler)(const char *fmt, ...);
#endif
{
    XDR_TCP *xdr_tcp = (XDR_TCP *)malloc(sizeof(XDR_TCP));
    XDR *xdrs = (XDR *)malloc(sizeof(XDR));		
    
    if (!xdr_tcp)
	/* memory allocation errors at this level are fatal */
	err_sys("Could not allocate XDR_TCP object");

    xdr_tcp->err_handler = error_handler;

    xdr_tcp->socket = sd;
    
    xdrs->x_ops = NULL;		/* test to det success of xdrrec_create */
    xdrrec_create(xdrs, buff_size, buff_size, (caddr_t)&xdr_tcp->socket,
		  (FUNC_RET_INT)tcp_read, (FUNC_RET_INT)tcp_write); 
    if (!xdrs->x_ops) {
	err_print("Could not create buffered TCP/IP stream");
	return NULL;
    }
    
    xdr_tcp->xdrs = xdrs;

    /* This kludge needed by vms */
#ifdef vms
    vms_channels = getdtablesize();
#endif
    
    return xdr_tcp;
}

/*
  Deallocate memory associated with a pointer to an instance of XDR_TCP and
  close its stream socket.
*/

void
xdr_tcp_close(xsd)
     XDR_TCP *xsd;
{
    XDR_DESTROY(xsd->xdrs);
    close(xsd->socket);
    free(xsd);
}

/*
  If the xdr_tcp connection is marked as XDR_ENCODE, place an end of record
  marker in the stream and write it to the socket. If the xdr_tcp connection
  is marked as XDR_DECODE, skip to the begining of the next record. Return
  TRUE if successful, FALSE otherwise.
*/

int
xdr_tcp_flush_record(xdr_tcp)
XDR_TCP *xdr_tcp;
{
    switch (xdr_tcp->xdrs->x_op) {
    case XDR_ENCODE:
	return xdrrec_endofrecord(xdr_tcp->xdrs, TRUE);

    case XDR_DECODE:
	return xdrrec_skiprecord(xdr_tcp->xdrs);

    default:		/* should never get here */
	return FALSE;
    }
}

/*
  Return TRUE if data exists on the TCP/IP connection that can be read,
  return FALSE Otherwise.
*/

int
xdr_tcp_read_poll(xdr_tcp)
XDR_TCP *xdr_tcp;
{
    fd_set rfd;
    struct timeval tm;

    timerclear(&tm);
    
    FD_ZERO(&rfd);
    FD_SET(xdr_tcp->socket, &rfd);

    if (select(FD_SETSIZE, &rfd, NULL, NULL, &tm) < 0)
	err_sys("xdr_tcp_read_poll: select failed");
    
    return (FD_ISSET(xdr_tcp->socket, &rfd));
}

/*
  Return TRUE if it is possible to write information to the TCP/IP
  connection, return FALSE otherwise.
*/

int
xdr_tcp_write_poll(xdr_tcp)
XDR_TCP *xdr_tcp;
{
    fd_set wfd;
    struct timeval tm;

    timerclear(&tm);
    
    FD_ZERO(&wfd);
    FD_SET(xdr_tcp->socket, &wfd);
    
    if (select(FD_SETSIZE, NULL, &wfd, NULL, &tm) < 0)
	err_sys("xdr_tcp_write_poll: select failed");

    return (FD_ISSET(xdr_tcp->socket, &wfd));
}

/*
  Register the struct type s_type (which must be a predefined constant) as
  having size and being encoded/decoded using coding_func. When sending this
  struct, it will be identified by the constant s_type. That value will be
  used to select parameters sent to the underlying XDR layer which is the
  interface with TCP/IP.
*/

int
xdr_tcp_set_struct(s_type, size, coding_func)
u_int s_type;
u_int size;
int (*coding_func)();
{
    if (s_type > MAX_STRUCT_TYPES) {
	err_print("Struct type code (%d) too large");
	return FALSE;
    }
	
    s_sizes[s_type] = size;
    s_coders[s_type] = coding_func;

    return TRUE;
}

int
xdr_tcp_set_array(a_type, size, elem_size, coding_func)
u_int a_type;
u_int size;
u_int elem_size;
int (*coding_func)();
{
    if (a_type > MAX_ARRAY_TYPES) {
	err_print("Array type code (%d) too large");
	return FALSE;
    }
	
    a_sizes[a_type] = size;
    a_elem_sizes[a_type] = elem_size;
    a_coders[a_type] = coding_func;

    return TRUE;
}

/*
  Functions to transmit data: send_action, send_long, send_double,
  send_string, send_struct, send_array and send_opaque.
*/

int
send_action(xdr_tcp, action, flush_now)
XDR_TCP *xdr_tcp;
u_int action;
int flush_now;
{
    enum type_code type = action_tc;
    
    assert(xdr_tcp);

    SET_XDR_ENCODE(xdr_tcp->xdrs);

    if (!xdr_u_int(xdr_tcp->xdrs, &type)) {
	XDR_TCP_ERROR("Could not send action type code");
	return FALSE;
    }
    
    if (!xdr_u_int(xdr_tcp->xdrs, &action)) {
	XDR_TCP_ERROR("Could not send action");
	return FALSE;
    }

    if (flush_now && !xdrrec_endofrecord(xdr_tcp->xdrs, flush_now)) {
	XDR_TCP_ERROR("Could not flush xdr tcp stream");
	return FALSE;
    }

    return TRUE;
}

int
send_long(xdr_tcp, data, flush_now)
XDR_TCP *xdr_tcp;
long data;
int flush_now;
{
    enum type_code type = long_tc;

    assert(xdr_tcp);

    SET_XDR_ENCODE(xdr_tcp->xdrs);
    
    if (!xdr_u_int(xdr_tcp->xdrs, &type)) {
	XDR_TCP_ERROR("Could not send integer type code");
	return FALSE;
    }

    if (!xdr_long(xdr_tcp->xdrs, &data)) {
	XDR_TCP_ERROR("Could not send integer");
	return FALSE;
    }

    if (flush_now && !xdrrec_endofrecord(xdr_tcp->xdrs, TRUE)) {
	XDR_TCP_ERROR("Could not flush xdr tcp stream");
	return FALSE;
    }

    return TRUE;
}

/*
int send_float(XDR_TCP *xdr_tcp, float data, int flush_now);
*/

int
send_float(xdr_tcp, data, flush_now)
XDR_TCP *xdr_tcp;
float data;
int flush_now;
{
    enum type_code type = float_tc;

    assert(xdr_tcp);

    SET_XDR_ENCODE(xdr_tcp->xdrs);
    
    if (!xdr_u_int(xdr_tcp->xdrs, &type)) {
	XDR_TCP_ERROR("Could not send float type code");
	return FALSE;
    }

    if (!xdr_float(xdr_tcp->xdrs, &data)) {
	XDR_TCP_ERROR("Could not send float");
	return FALSE;
    }

    if (flush_now && !xdrrec_endofrecord(xdr_tcp->xdrs, TRUE)) {
	XDR_TCP_ERROR("Could not flush xdr tcp stream");
	return FALSE;
    }

    return TRUE;
}

int
send_double(xdr_tcp, data, flush_now)
XDR_TCP *xdr_tcp;
double data;
int flush_now;
{
    enum type_code type = double_tc;

    assert(xdr_tcp);

    SET_XDR_ENCODE(xdr_tcp->xdrs);
    
    if (!xdr_u_int(xdr_tcp->xdrs, &type)) {
	XDR_TCP_ERROR("Could not send doauble type code");
	return FALSE;
    }
    
    if (!xdr_double(xdr_tcp->xdrs, &data)) {
	XDR_TCP_ERROR("Could not send double");
	return FALSE;
    }

    if (flush_now && !xdrrec_endofrecord(xdr_tcp->xdrs, TRUE)) {
	XDR_TCP_ERROR("Could not flush xdr tcp stream");
	return FALSE;
    }

    return TRUE;
}

int
send_string(xdr_tcp, str, maxlen, flush_now)
XDR_TCP *xdr_tcp;
char *str;
u_int maxlen;
int flush_now;
{
    enum type_code type = string_tc;

    assert(xdr_tcp);
    SET_XDR_ENCODE(xdr_tcp->xdrs);

    if (!xdr_u_int(xdr_tcp->xdrs, &type)) {
	XDR_TCP_ERROR("Could not send integer type code");
	return FALSE;
    }

    if (!xdr_string(xdr_tcp->xdrs, &str, maxlen)) {
	XDR_TCP_ERROR("Could not send string");
	return FALSE;
    }

    if (flush_now && !xdrrec_endofrecord(xdr_tcp->xdrs, TRUE)) {
	XDR_TCP_ERROR("Could not flush xdr tcp stream");
	return FALSE;
    }

    return TRUE;
}

int
send_struct(xdr_tcp, struct_data, s_type, flush_now)
XDR_TCP *xdr_tcp;
caddr_t struct_data;
u_int s_type;			/* type of structure - application spec */
int flush_now;
{
    enum type_code type = struct_tc;

    assert(xdr_tcp);

    SET_XDR_ENCODE(xdr_tcp->xdrs);
    
    if (!xdr_u_int(xdr_tcp->xdrs, &type)) {
	XDR_TCP_ERROR("Could not send struct type code");
	return FALSE;
    }

    if (!xdr_u_int(xdr_tcp->xdrs, &s_type)) {
	XDR_TCP_ERROR("Could not send application struct type code");
	return FALSE;
    }

    if (!xdr_reference(xdr_tcp->xdrs, &struct_data, STRUCT_SIZE(s_type),
		       STRUCT_CODER(s_type))) {
	XDR_TCP_ERROR("Could not send struct");
	return FALSE;
    }

    if (flush_now && !xdrrec_endofrecord(xdr_tcp->xdrs, TRUE)) {
	XDR_TCP_ERROR("Could not flush xdr tcp stream");
	return FALSE;
    }

    return TRUE;
}

/*
  NB: The number of elements in a multidimensional array is the product of
  the size of the dimensions.
*/

int
send_array(xdr_tcp, array, size, a_type, flush_now)
XDR_TCP *xdr_tcp;
caddr_t array;
u_int size;			/* number of elements in array */
u_int a_type;			/* type of structure - application spec */
int flush_now;
{
    enum type_code type = array_tc;

    assert(xdr_tcp);

    SET_XDR_ENCODE(xdr_tcp->xdrs);
    
    if (!xdr_u_int(xdr_tcp->xdrs, &type)) {
	XDR_TCP_ERROR("Could not send array type code");
	return FALSE;
    }

    if (!xdr_u_int(xdr_tcp->xdrs, &a_type)) {
	XDR_TCP_ERROR("Could not send application array type code");
	return FALSE;
    }

    if (!xdr_array(xdr_tcp->xdrs, &array, &size, ARRAY_MAXSIZE(a_type),
		   ARRAY_ELEM_SIZE(a_type), ARRAY_CODER(a_type))) {
	XDR_TCP_ERROR("Could not send array");
	return FALSE;
    }

    if (flush_now && !xdrrec_endofrecord(xdr_tcp->xdrs, TRUE)) {
	XDR_TCP_ERROR("Could not flush xdr tcp stream");
	return FALSE;
    }

    return TRUE;
}

int
send_opaque(xdr_tcp, opaque, size, flush_now)
XDR_TCP *xdr_tcp;
caddr_t opaque;
long size;
int flush_now;
{
    enum type_code type = opaque_tc;

    assert(xdr_tcp);

    SET_XDR_ENCODE(xdr_tcp->xdrs);
    
    if (!xdr_u_int(xdr_tcp->xdrs, &type)) {
	XDR_TCP_ERROR("Could not send opaque type code");
	return FALSE;
    }

    if (!xdr_long(xdr_tcp->xdrs, &size)) {
	XDR_TCP_ERROR("Could not send opaque data");
	return FALSE;
    }
    
    if (!xdr_opaque(xdr_tcp->xdrs, opaque, size)) {
	XDR_TCP_ERROR("Could not send opaque data");
	return FALSE;
    }

    if (flush_now && !xdrrec_endofrecord(xdr_tcp->xdrs, TRUE)) {
	XDR_TCP_ERROR("Could not flush xdr tcp stream");
	return FALSE;
    }

    return TRUE;
}

/*
  Functions to read data: get_action, get_long, get_double, get_string,
  get_struct, get_array, get_opaque. 
*/

u_int
get_action(xdr_tcp, next_rec)
XDR_TCP *xdr_tcp;
int next_rec;
{
    register enum type_code status;

    status = read_type_code(xdr_tcp, next_rec);
    if (status != action_tc)
	return (u_int)FALSE;
    
    return read_action(xdr_tcp);
}

int
get_long(xdr_tcp, datum, next_rec)
XDR_TCP *xdr_tcp;
long *datum;
int next_rec;
{
    register enum type_code status;

    status = read_type_code(xdr_tcp, next_rec);
    if (status != long_tc)
	return (int)FALSE;
    
    return read_long(xdr_tcp, datum);
}

int
get_float(xdr_tcp, datum, next_rec)
XDR_TCP *xdr_tcp;
float *datum;
int next_rec;
{
    register enum type_code status;

    status = read_type_code(xdr_tcp, next_rec);
    if (status != float_tc)
	return (int)FALSE;
    
    return read_float(xdr_tcp, datum);
}

int
get_double(xdr_tcp, datum, next_rec)
XDR_TCP *xdr_tcp;
double *datum;
int next_rec;
{
    register enum type_code status;

    status = read_type_code(xdr_tcp, next_rec);
    if (status != double_tc)
	return (int)FALSE;
    
    return read_double(xdr_tcp, datum);
}

/*
  NB: returns a pointer to the string. returns NULL if an error is detected.
  If `str' is NULL, read_string() will allocate space for the data read from
  the socket.
*/

char *
get_string(xdr_tcp, str, maxlen, next_rec)
XDR_TCP *xdr_tcp;
char *str;
int maxlen;
int next_rec;
{
    register enum type_code status;

    status = read_type_code(xdr_tcp, next_rec);
    if (status != string_tc)
	return NULL;
    
    return read_string(xdr_tcp, str, maxlen);
}

/*
  NB: the function returns a pointer to the struct. s_type is a value-result
  parameter. returns NULL if an error is detected.
*/

caddr_t
get_struct(xdr_tcp, struct_data, s_type, next_rec)
XDR_TCP *xdr_tcp;
caddr_t struct_data;
u_int *s_type;
int next_rec;
{
    register enum type_code status;

    status = read_type_code(xdr_tcp, next_rec);
    if (status != struct_tc)
	return NULL;
    
    return read_struct(xdr_tcp, struct_data, s_type);
}

/*
  NB: returns a pointer to the first element of the array. length and a_type
  are value result parameters. returns NULL if an error is detected.
*/

caddr_t
get_array(xdr_tcp, array, length, a_type, next_rec)
XDR_TCP *xdr_tcp;
caddr_t array;
u_int *length;
u_int *a_type;
int next_rec;
{
    register enum type_code status;

    status = read_type_code(xdr_tcp, next_rec);
    if (status != array_tc)
	return NULL;
    
    return read_array(xdr_tcp, array, length, a_type);
}

/*
  NB: returns TRUE/FALSE on success/failure. length is a value-result
  parameter.
*/

int
get_opaque(xdr_tcp, opaque, length, next_rec)
XDR_TCP *xdr_tcp;
caddr_t opaque;
long *length;
int next_rec;
{
    register enum type_code status;
    
    status = read_type_code(xdr_tcp, next_rec);
    if (status != opaque_tc)
	return FALSE;
    
    return read_opaque(xdr_tcp, opaque, length);
}

/*
  These I/O functions read part of a compete message. They allow a program to
  use the type code (that is transmitted at the beginning of each action) to
  determine the contents of the stream and act accordingly.
*/

/*
  Read a single char. If an error is detected, print a message and (if the
  error message reporter returns) return FALSE. If no error is detected,
  return the `type' (enum type_code). The declaration of `enum type_code'
  appears in comm.h.
*/

enum type_code
read_type_code(xdr_tcp, next_rec)
XDR_TCP *xdr_tcp;
int next_rec;
{
    enum type_code type;

    assert(xdr_tcp);

    SET_XDR_DECODE(xdr_tcp->xdrs);
    
    if (next_rec && !xdrrec_skiprecord(xdr_tcp->xdrs)) {
	XDR_TCP_ERROR("Could not move to next record");
	return FALSE;
    }
    
    if (!xdr_u_int(xdr_tcp->xdrs, &type)) {
	XDR_TCP_ERROR("Could not read type code");
	return FALSE;
    }

    return type;
}

u_int
read_action(xdr_tcp)
XDR_TCP *xdr_tcp;
{
    u_int action;

    assert(xdr_tcp);

    SET_XDR_DECODE(xdr_tcp->xdrs);
    
    if (!xdr_u_int(xdr_tcp->xdrs, &action)) {
	XDR_TCP_ERROR("Could not read action");
	return FALSE;
    }

    return action;
}

/*
  Read a single 4-byte integer. If an error is detected, print a message and
  (if the error message reporter returns) return FALSE.  If no error is
  detected, return TRUE.
*/

int
read_long(xdr_tcp, datum)
XDR_TCP *xdr_tcp;
long *datum;
{
    assert(xdr_tcp);

    SET_XDR_DECODE(xdr_tcp->xdrs);
    
    if (!xdr_long(xdr_tcp->xdrs, datum)) {
	XDR_TCP_ERROR("Could not read long");
	return FALSE;
    }

    return TRUE;
}

int
read_float(xdr_tcp, datum)
XDR_TCP *xdr_tcp;
float *datum;
{
    assert(xdr_tcp);
    
    SET_XDR_DECODE(xdr_tcp->xdrs);
    
    if (!xdr_float(xdr_tcp->xdrs, datum)) {
	XDR_TCP_ERROR("Could not read float");
	return FALSE;
    }

    return TRUE;
}

int
read_double(xdr_tcp, datum)
XDR_TCP *xdr_tcp;
double *datum;
{
    assert(xdr_tcp);
    
    SET_XDR_DECODE(xdr_tcp->xdrs);
    
    if (!xdr_double(xdr_tcp->xdrs, datum)) {
	XDR_TCP_ERROR("Could not read double");
	return FALSE;
    }

    return TRUE;
}

char *
read_string(xdr_tcp, str, maxlen)
XDR_TCP *xdr_tcp;
char *str;
int maxlen;
{
    assert(xdr_tcp);
    
    SET_XDR_DECODE(xdr_tcp->xdrs);
    
    if (!xdr_string(xdr_tcp->xdrs, &str, maxlen)) {
	XDR_TCP_ERROR("Could not read string");
	return NULL;
    }

    return str;
}

caddr_t
read_struct(xdr_tcp, struct_data, s_type)
XDR_TCP *xdr_tcp;
caddr_t struct_data;
u_int *s_type;
{
    assert(xdr_tcp);

    SET_XDR_DECODE(xdr_tcp->xdrs);
    
    if (!xdr_u_int(xdr_tcp->xdrs, s_type)) {
	XDR_TCP_ERROR("Could not read struct type code");
	return FALSE;
    }

    if (!xdr_reference(xdr_tcp->xdrs, (char **)&struct_data,
		       STRUCT_SIZE(*s_type), STRUCT_CODER(*s_type))) {
	XDR_TCP_ERROR("Could not read struct");
	return NULL;
    }

    return struct_data;
}

caddr_t
read_array(xdr_tcp, array, size, a_type)
XDR_TCP *xdr_tcp;
caddr_t array;
u_int *size;
u_int *a_type;
{
    assert(xdr_tcp);
    
    if (!xdr_u_int(xdr_tcp->xdrs, a_type)) {
	XDR_TCP_ERROR("Could not read array type code");
	return FALSE;
    }

    if (!xdr_array(xdr_tcp->xdrs, (char **)&array, size,
		   ARRAY_MAXSIZE(*a_type), ARRAY_ELEM_SIZE(*a_type),
		   ARRAY_CODER(*a_type))) {
	XDR_TCP_ERROR("Could not read array");
	return NULL;
    }

    return array;
}

int
read_opaque(xdr_tcp, opaque, length)
XDR_TCP *xdr_tcp;
caddr_t opaque;
long *length;
{
    assert(xdr_tcp);
    
    if (!xdr_long(xdr_tcp->xdrs, length)) {
	XDR_TCP_ERROR("Could not read array size");
	return FALSE;
    }
    
    if (!xdr_opaque(xdr_tcp->xdrs, opaque, *length)) {
	XDR_TCP_ERROR("Could not send opaque data");
	return FALSE;
    }

    return TRUE;
}
